(* ale_lib.v *)
(* Defines the stub library that is used in the ALE machine *)
(* The stub library includes the ALE specifications of mem_alloc and mem_free *)
(* TODO: TLB stubs are currently not used *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import label.
Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import c.
Require Import ale.

Module ALE_Lib.

Import C_ALE.

Definition FUN (ret : Z) : SAct (State := State) := 
  StackAct (pop +++ setRet ret).

Definition allocfail : SAct (State := State) :=
  StackAct (pop +++ setRet 0%Z).

Definition ALE_alloc (pg : Z) : ALE_State.t -> ALE_State.t -> Prop :=
  fun s s' => 
	 (forall l, s.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\ 
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then true else s.(allocdata) x)) /\
	(s'.(tlb) = s.(tlb)).

Definition allocpg (pg : Z) : State -> State -> Prop :=
  fun s s' => ALE_alloc pg s.(memory) s'.(memory) /\ StackRets pg s.(stack) s'.(stack).


Definition mem_alloc_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 0) s = true /\ StackCheck isFrame s = true /\ 
                         ALE_State.DM s.(memory) /\ MemGet getTLB s = true
     );

rel := (fun s s' =>
    (exists pg, (1 <= pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = false /\ allocpg pg s s')
    \/
    s' = allocfail s)
|}.


Definition ALE_free (pg : Z) : ALE_State.t -> ALE_State.t -> Prop :=
  fun s s' => 
	 (forall l, s'.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\ 
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then false else s.(allocdata) x)) /\
	(s'.(tlb) = s.(tlb)).


Definition freepg (pg : Z) : State -> State -> Prop :=
   fun s s' => ALE_free pg s.(memory) s'.(memory) /\ StackRets 0%Z s.(stack) s'.(stack).


Lemma ALE_free_data_unch : forall l x0 s s',
ALE_State.allocdata s' (Pg l) = true ->
ALE_Lib.ALE_free x0 s s' ->
ALE_State.data s' l = ALE_State.data s l.
intros.
red in H0; decompose record H0; clear H0.
rewrite H1; autosolve.
Qed.


Lemma ALE_free_allocdata_pullback : forall pg x0 s s',
ALE_Lib.ALE_free x0 s s' ->
ALE_State.allocdata s' pg = true ->
ALE_State.allocdata s pg = true.
intros.
red in H; decompose record H; clear H.
rewrite H3 in *; clear H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.




Definition mem_free_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                 let pg := nth 0 (StackGet getArgs s) 0%Z in
                                 (0 < pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = true /\ 
	                         PtDom(pg * PGSIZE + Off 0) = false /\
                                 ALE_State.DM s.(memory) /\
                                  MemGet getTLB s = true);

rel := (fun s s' => freepg (nth 0 (StackGet getArgs s) 0%Z) s s')
|}.

Definition tlb_reset_spec : Action State :=
SAct_Action (StackCheck (getArgsCheck 0))
	           (StackAct (pop +++ setRet 0%Z) +++ 
	             (MemAct (setTLB true))).

Definition ale_lib_spec : SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free_spec
  | mem_alloc_label => Some mem_alloc_spec
  | tlb_reset_label => Some tlb_reset_spec
  | _ => None
  end.

End ALE_Lib.
