(* ale_pe.v *)
(* Defines the refinement from the ALE memory model to the PE memory model *)
(* Also includes proofs that ALE's stub specifications match up with specifications of the allocator (mem_sig.v) *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.

Require Import mem_repr.
Require Import ale.
Require Import pe.

Module ALE_PE <: Mem_Repr_T ALE_State PE_State.

Opaque PE_State.PTROOT ALE_State.PTROOT PTROOT PE_State.PMM_data_loc.

Definition Repr_Mem (ale : ALE_State.t) (pe : PE_State.t) :=
   (forall l, ValidVA l = true -> ALE_State.allocdata ale (Pg l) = true -> (ALE_State.data ale) l = PE_State.data pe l) /\ 
   (forall pg, ValidPg pg = true -> ALE_State.allocdata ale pg = true -> 
                    PE_State.data pe (PE_State.PMM_data_loc + pg * 8) = 1%Z) /\
   (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> ALE_State.allocdata ale (Pg l) = false) /\
   (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> PE_State.data pe (PE_State.PMM_data_loc + (Pg l) * 8) = 1%Z) /\
   (PE_State.PTROOT = ALE_State.PTROOT) /\
   (ALE_State.tlb ale = PE_State.tlb pe) /\ 
   (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> ALE_State.allocdata ale (Pg l) = true).

Lemma Repr_vaCheck : forall (m1 : ALE_State.t) (m2 : PE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, ALE_State.vaCheck l m1 = true -> PE_State.ValidVA l m2 = true.
intros.
unfold ALE_State.vaCheck, ALE_State.transCheck, ALE_State.paCheck, PE_State.ValidVA in *; simpl; repeat autoclean.
red in H; repeat autoclean.
rewrite <- H9; rewrite H4; simpl.
unfold isPhys in *.
rewrite H8.
rewrite <- H; autosolve.
apply ValidPA_ValidVA.
apply ALE_State.PtDom_Valid.
apply ValidVA_ValidVPg; auto.
apply H10.
apply ALE_State.PtDom_Valid; auto.
apply ValidVA_ValidVPg; auto.
apply ALE_State.PtDom_PTROOT.
apply ValidVA_ValidVPg; auto.
Qed.

Lemma Repr_getMemCheck : forall (m1 : ALE_State.t) (m2 : PE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, ALE_State.getMemCheck l m1 = true -> PE_State.getMemCheck l m2 = true.
intros.
eapply Repr_vaCheck; eauto.
Qed.

Lemma Repr_getMem : forall (m1 : ALE_State.t) (m2 : PE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v, ALE_State.getMemCheck l m1 = true -> ALE_State.getMem l m1 = v -> PE_State.getMem l m2 = v.
intros.
red in H; repeat autoclean.
unfold ALE_State.getMem, ALE_State.trans, PE_State.getMem in *.
unfold ALE_State.getMemCheck in H0.
unfold ALE_State.vaCheck in H0; repeat autoclean.
unfold ALE_State.transCheck, ALE_State.paCheck, ALE_State.trans in *; repeat autoclean.
assert (ValidVA (ALE_State.PTROOT + Pg l * 8) = true).
apply ValidPA_ValidVA.
apply ALE_State.PtDom_Valid.
apply ValidVA_ValidVPg; auto.
assert (ALE_State.allocdata m1 (Pg (ALE_State.PTROOT + Pg l * 8)) = true).
apply H7.
apply ALE_State.PtDom_Valid.
apply ValidVA_ValidVPg; auto.
apply ALE_State.PtDom_PTROOT.
apply ValidVA_ValidVPg; auto.
rewrite H5; simpl.
repeat rewrite <- H; autosolve.
apply ValidPA_ValidVA; auto.
Qed.

Lemma Repr_setMemCheck : forall (m1 : ALE_State.t) (m2 : PE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, ALE_State.setMemCheck l m1 = true -> PE_State.setMemCheck l m2 = true.
intros.
unfold ALE_State.setMemCheck, PE_State.setMemCheck in *; repeat autoclean.
eapply Repr_vaCheck; eauto.
Qed.

Lemma Repr_setMem : forall (m1 : ALE_State.t) (m2 : PE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v , ALE_State.setMemCheck l m1 = true -> Repr_Mem (ALE_State.setMem l v m1) (PE_State.setMem l v m2).
intros.
red in H; repeat autoclean.
unfold ALE_State.setMemCheck, ALE_State.vaCheck, ALE_State.transCheck, ALE_State.paCheck in *; repeat autoclean.
red.
unfold ALE_State.setMem, ALE_State.trans, PE_State.setMem, Repr_Mem in *; simpl.
assert (ValidVA (ALE_State.PTROOT + Pg l * 8) = true).
apply ValidPA_ValidVA.
apply ALE_State.PtDom_Valid.
apply ValidVA_ValidVPg; auto.
assert (ALE_State.allocdata m1 (Pg (ALE_State.PTROOT + Pg l * 8)) = true).
apply H6.
apply ALE_State.PtDom_Valid.
apply ValidVA_ValidVPg; auto.
apply ALE_State.PtDom_PTROOT.
apply ValidVA_ValidVPg; auto.
repeat split; autosolve.

intros.
rewrite H4.
repeat rewrite <- H; auto.

intros.
rewrite H1; autosolve.
rewrite H4.
rewrite <- H; auto.
unfold isPhys in *.
makenew' (ValidPg (Pg l)); autosolve.
rewrite H9 in *; clear H9.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H15 in H7.
rewrite H2 in H7; autosolve.
apply PE_State.PMMDom_Valid; auto.
apply PE_State.PMMDom_Loc; auto.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H17 in *; clear H17.
rewrite H2 in H7; autosolve.
apply PE_State.PMMDom_Loc; auto.

intros.
rewrite H3; auto.
rewrite H4.
rewrite <- H; auto.
unfold isPhys in *.
makenew' (ValidPg (Pg l)); autosolve.
rewrite H9 in *; clear H9.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H15 in H7.
rewrite H2 in H7; autosolve.
apply PE_State.PMMDom_Valid; auto.
apply ValidPA_ValidPg; auto.
apply PE_State.PMMDom_Loc; auto.
apply ValidPA_ValidPg; auto.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H17 in *; clear H17.
rewrite H2 in H7; autosolve.
apply PE_State.PMMDom_Loc; auto.
apply ValidPA_ValidPg; auto.
Qed.

End ALE_PE.

Require Import c.

Module C_ALE_PE := C_Refine ALE_State PE_State ALE_PE C_ALE C_PE.

Import ALE_PE.
Require Import ale_lib.
Require Import mem_sig.
Require Import pe_lib.
Import mem_sig.

Opaque ALE_State.PTROOT PE_State.PTROOT PTROOT PE_State.PMM_data_loc.

Lemma exists_alloc : forall pg newpagesrc s, 
   s.(C_ALE.State.memory).(ALE_State.allocdata) pg = false ->
   exists s', 
   ALE_Lib.allocpg pg s s' /\  (forall l, Pg l = pg ->(C_ALE.State.memory s').(ALE_State.data) l = newpagesrc l).
intros.
unfold ALE_Lib.allocpg.
unfold ALE_Lib.ALE_alloc.
unfold C_ALE.State.Stack.StackRets.
destruct s; simpl in *.
destruct memory; simpl in *.
exists 
{| C_ALE.State.stack := (C_ALE.State.Stack.pop +++ C_ALE.State.Stack.setRet pg) stack;
    C_ALE.State.memory := 
         {| ALE_State.data := (fun l => if Zeq_bool (Pg l) pg then newpagesrc l else data l);
             ALE_State.allocdata := (fun x : Z => if Zeq_bool x pg then true else allocdata x);
             ALE_State.tlb := tlb
          |}
|}; simpl.
repeat split; auto.
intros; auto.
Zeq_pull'; Zeq_norm; repeat autoclean; simpl in *; autosolve.
rewrite H0 in H; autosolve.
intros.
rewrite H0; rewrite Zeq_bool_refl; auto.
Qed.

Lemma alloc_repr : forall x0 bs bs' s s',
ValidPg x0 = true -> 
PE_State.getPageInfo x0 s = false ->
(forall l : Z,  Pg l = x0 -> ALE_State.data bs' l = PE_State.data s' l) ->
Repr_Mem bs s ->
ALE_Lib.ALE_alloc x0 bs bs' ->
s' = PE_State.setPageInfo x0 true s ->
Repr_Mem bs' s'.
intros.
red in H2; repeat autoclean.
red; repeat split; autosolve.

intros.
red in H3; decompose record H3; clear H3.
rewrite H14 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H1; autosolve.
rewrite H12; autosolve.
rewrite H2; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H16 in *.
rewrite H6 in H11; autosolve.
apply PE_State.PMMDom_Valid; auto.
apply PE_State.PMMDom_Loc; autosolve.

intros.
unfold PE_State.setPageInfo; simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red in H3; decompose record H3; clear H3.
apply H5; autosolve.
rewrite H15 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros.
red in H3; decompose record H3; clear H3.
rewrite H14.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H13 in *; clear x0 H13.
unfold PE_State.getPageInfo in H0.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H7 in H13; autosolve.
apply H6; autosolve.

intros.
unfold PE_State.setPageInfo; simpl.
red in H3; decompose record H3; clear H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H7; auto.

unfold PE_State.setPageInfo; simpl; auto.
red in H3; decompose record H3; clear H3.
rewrite H13; auto.

intros.
red in H3; decompose record H3; clear H3.
rewrite H14.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H10; autosolve.
Qed.

Lemma allocdata_repr : forall x0 bs s,
ValidPg x0 = true ->
PE_State.getPageInfo x0 s = false ->
Repr_Mem bs s -> 
PE_State.getPageInfo x0 s = false -> 
ALE_State.allocdata bs x0 = false.
intros.
red in H1; repeat autoclean.
unfold PE_State.getPageInfo in H2.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (ALE_State.allocdata bs x0); autosolve.
rewrite H3 in H10; autosolve.
Qed.

Lemma DM_repr : forall x s,
Repr_Mem x s -> ALE_State.DM x -> PE_State.DM s.
intros.
red in H; repeat autoclean.
red in H0.
unfold ALE_State.trans in *.
red; intros.
rewrite H4.
rewrite <- H.
destruct (ValidVPg_existsVA pg).
apply ValidPg_ValidVPg; auto.
destruct H8.
rewrite <- H9 in *; clear H9 pg.
assert (ValidPA x0 = true).
apply ValidVA_Phys; auto.
specialize (H0 x0 H9).
rewrite Addr_parts in H0.
destruct (Zeq_dec (ALE_State.data x (ALE_State.PTROOT + Pg x0 * 8)) (Pg x0)); autosolve.
apply ValidPA_ValidVA.
apply ALE_State.PtDom_Valid.
apply ValidPg_ValidVPg; auto.
apply H6.
apply ALE_State.PtDom_Valid.
apply ValidPg_ValidVPg; auto.
apply ALE_State.PtDom_PTROOT.
apply ValidPg_ValidVPg; auto.
Qed.

Lemma alloc_weak1 : ActionGTE (C_ALE_PE.RF.reprAction C_ALE_PE.CRepr ALE_Lib.mem_alloc_spec)
  mem_sig.mem_alloc_spec.
split.

intros.
unfold ALE_Lib.mem_alloc_spec, mem_alloc_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_ALE.State.StackCheck, C_PE.State.StackCheck in *; simpl in *.
unfold C_ALE.State.MemGet, C_PE.State.MemGet in *; simpl in *.
unfold ALE_State.getTLB, PE_State.getTLB in *; simpl in *.
rewrite H in *.
red in H4; repeat autoclean.
repeat split; autosolve.
rewrite H9 in *; auto.
apply DM_repr with (C_ALE.State.memory x); autosolve.
red; repeat split; autosolve.

(* rel *)
intros.
unfold ALE_Lib.mem_alloc_spec, mem_alloc_spec in *; simpl in *; repeat autoclean.
destruct H0; repeat autoclean.
(* page x0 allocated *)
destruct (exists_alloc x0 (s'.(C_PE.State.memory).(PE_State.data)) bs) as [bs'].
(* exists_alloc: need to show that x0 is currently unallocated *)
unfold C_PE.State.MemGet, PE_State.getPageInfo in H10.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (ALE_State.allocdata (C_ALE.State.memory bs) x0); autosolve.
destruct H5.
red in H13; decompose record H13; clear H13.
rewrite H18 in H14; autosolve.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(*****)
exists bs'; repeat autoclean.
split.
red in H5; repeat autoclean.
red in H11; repeat autoclean.
red in H13; repeat autoclean.
red; split.
rewrite H17.
rewrite H16.
rewrite H5; auto.
apply alloc_repr with x0 (C_ALE.State.memory bs) (C_PE.State.memory s); auto.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
left.
exists x0.
repeat (split; auto).
red in H5; repeat autoclean.
unfold C_PE.State.MemGet in H8.
apply allocdata_repr with (C_PE.State.memory s); auto.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* page not allocated *)
eexists; split; eauto.
unfold ALE_Lib.allocfail.
red in H5; repeat autoclean.
red; simpl.
rewrite H0; auto.
Qed.


Lemma exists_free : forall pg bs, exists bs', ALE_Lib.freepg pg bs bs'.
intros.
unfold ALE_Lib.freepg.
unfold ALE_Lib.ALE_free; simpl.
unfold C_ALE.State.Stack.StackRets; simpl.
destruct bs; simpl.
destruct memory; simpl.
exists 
{| C_ALE.State.stack := (C_ALE.State.Stack.pop +++ C_ALE.State.Stack.setRet 0%Z) stack;
    C_ALE.State.memory := 
         {| ALE_State.data := data;
             ALE_State.allocdata := (fun x : Z => if Zeq_bool x pg then false else allocdata x);
             ALE_State.tlb := tlb
          |}
|}; simpl.
repeat split; auto.
Qed.

Lemma free_repr : forall pg bs bs' s,
ValidPg pg = true -> 
ALE_State.PtDom (pg * PGSIZE + Off 0) = false ->
Repr_Mem bs s ->
ALE_State.allocdata bs pg = true ->
PE_State.getPageInfo pg s = true ->
ALE_Lib.ALE_free pg bs bs' ->
Repr_Mem bs' (PE_State.setPageInfo pg false s).

intros.
red in H1; repeat autoclean.
unfold PE_State.getPageInfo, PE_State.setPageInfo in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
repeat autoclean.
red; repeat split; simpl in *; autosolve.

intros.
red in H4; decompose record H4; clear H4.
rewrite H13; auto.
rewrite H15 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H17 in H11.
rewrite H6 in H11; autosolve.
apply PE_State.PMMDom_Valid; auto.
rewrite PE_State.PMMDom_Loc; auto.
apply H1; auto.

intros.
red in H4; decompose record H4; clear H4.
rewrite H15 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H5; auto.

intros.
red in H4; decompose record H4; clear H4.
rewrite H15.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H6; auto.

intros.
rewrite H7; auto.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
assert (pg = Pg l); autosolve.
rewrite H13 in *; clear pg H13; repeat autoclean.
rewrite H6 in H2; autosolve.

red in H4; decompose record H4; clear H4.
rewrite H14; auto.

intros.
red in H4; decompose record H4; clear H4.
rewrite H15.
rewrite H10; auto.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H14 in *; clear H14 pg.
rewrite <- ALE_State.PtDom_zero in H0; autosolve.
rewrite H0 in H11; autosolve.
Qed.

Lemma free_weak1 : ActionGTE (C_ALE_PE.RF.reprAction C_ALE_PE.CRepr ALE_Lib.mem_free_spec)
  mem_sig.mem_free_spec.
split.

intros.
unfold ALE_Lib.mem_free_spec, mem_sig.mem_free_spec; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_ALE.State.StackCheck, C_PE.State.StackCheck in *; simpl in *.
unfold C_ALE.State.StackGet, C_PE.State.StackGet in *; simpl in *.
red in H7; repeat autoclean.
rewrite H in *; repeat split; autosolve.
unfold C_PE.State.MemGet, PE_State.getPageInfo.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew  (nth 0 (C_ALE.State.Stack.getArgs (C_PE.State.stack s)) 0%Z).

destruct (ValidVPg_existsVA H14).
apply ValidPg_ValidVPg.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct H17.
rewrite <- H18 in *; clear H18 H14.
rewrite H8 in H15; autosolve.
apply ValidPA_ValidPg.
apply ValidVA_Phys; auto.
unfold isPhys.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

unfold C_ALE.State.MemGet, C_PE.State.MemGet, ALE_State.getTLB, PE_State.getTLB in *.
rewrite H12 in *; auto.

apply DM_repr with (C_ALE.State.memory x); auto.
red; repeat split; auto.

(* rel *)
intros.
unfold ALE_Lib.mem_free_spec, mem_sig.mem_free_spec; simpl in *; repeat autoclean.
makenew' (C_PE.State.MemGet
          (PE_State.getPageInfo
             (nth 0 (C_PE.State.StackGet C_PE.State.Stack.getArgs s) 0%Z)) s); autosolve.
subst.
destruct (exists_free (nth 0 (C_ALE.State.StackGet C_ALE.State.Stack.getArgs bs) 0%Z) bs) as [bs'].
exists bs'; split; auto.
red in H; repeat autoclean.
unfold C_ALE.State.StackCheck, C_ALE.State.StackGet, C_PE.State.StackGet, C_PE.State.StackAct in *.
rewrite H in *.
red in H8; repeat autoclean.
rewrite H8 in *.
red; split; simpl.
red in H0; decompose record H0; clear H0.
rewrite H20.
rewrite H8; auto.

makenew (nth 0 (C_PE.State.Stack.getArgs (C_PE.State.stack s)) 0%Z).
red in H0; destruct H0.
apply free_repr with (C_ALE.State.memory bs); auto.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma tlb_reset_weak1 : ActionGTE (C_ALE_PE.RF.reprAction C_ALE_PE.CRepr ALE_Lib.tlb_reset_spec)
  PE_Lib.tlb_reset_spec.
split.

intros.
unfold ALE_Lib.tlb_reset_spec, PE_Lib.tlb_reset_spec in *; simpl in *.
repeat autoclean.
unfold C_ALE.State.StackCheck, C_PE.State.StackCheck in *.
red in H; repeat autoclean.
rewrite H in *; auto.

(* rel *)
intros.
unfold ALE_Lib.tlb_reset_spec, PE_Lib.tlb_reset_spec in *; simpl in *.
repeat autoclean.
eexists; split; eauto.
red in H2; repeat autoclean.
red; split; simpl.
rewrite H0; auto.
unfold ALE_State.setTLB, PE_State.setTLB.
red in H2; repeat autoclean.
red; simpl; repeat split; autosolve.
Qed.


(*********************************************************)
Lemma PMM_PT_separate' : forall l, ALE_State.PtDom l = true -> PE_State.PMMDom l = true -> False.
intros.
unfold ALE_State.PtDom, PE_State.PMMDom in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Transparent ALE_State.PTROOT PE_State.PMM_data_loc PTROOT.
unfold ALE_State.PTROOT, PE_State.PMM_data_loc, common.PTROOT in *.
autosolve.
Qed.
Opaque ALE_State.PTROOT PE_State.PMM_data_loc PTROOT.

Lemma ALE_PE_mem_exists : forall (s : PE_State.t), 
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                      s.(PE_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) ->
     (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> 
                      s.(PE_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) ->
     (forall pg, ValidPg pg = true -> s.(PE_State.data) (ALE_State.PTROOT + pg * 8) = pg) ->
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s.(PE_State.data) (ALE_State.PTROOT + pg * 8)= 0%Z) ->
      s.(PE_State.tlb) = true ->

(exists bs, ALE_PE.Repr_Mem bs s /\ 
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                       bs.(ALE_State.allocdata) (Pg l) = true) /\
     (forall pg, ValidPg pg = true -> bs.(ALE_State.data) (ALE_State.PTROOT + pg * 8) = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            bs.(ALE_State.data) (ALE_State.PTROOT + pg * 8)= 0%Z) /\
      bs.(ALE_State.tlb) = true
).
intros.
exists {| 
   ALE_State.data := (fun l => s.(PE_State.data) l);
   ALE_State.allocdata := (fun pg => 
              if PE_State.PMMDom (pg * PGSIZE + 0%Z) then false else
              if Zeq_bool (s.(PE_State.data) (PE_State.PMM_data_loc + pg * 8)%Z) 1%Z then true else false);
   ALE_State.tlb := true
|}; repeat split; simpl; autosolve.

intros.
makenew' (PE_State.PMMDom (pg * 4096 + 0)); autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros.
rewrite H0; auto.
rewrite <- PE_State.PMMDom_zero.
rewrite H5; auto.

intros.
rewrite <- PE_State.PMMDom_zero.
rewrite H; auto.
rewrite Zeq_bool_refl; auto.
makenew' (PE_State.PMMDom l); auto.
elimtype False; apply PMM_PT_separate' with l; auto.

intros.
rewrite <- PE_State.PMMDom_zero.
rewrite H; auto.
rewrite Zeq_bool_refl; auto.
makenew' (PE_State.PMMDom l); auto.
elimtype False; apply PMM_PT_separate' with l; auto.
Qed.

Require Import mem_sig.
Require Import pmap_sig.
Lemma kernel_init_weak : ActionGTE mem_sig.PE_kernel_init_spec
            (C_ALE_PE.RF.reprAction C_ALE_PE.CRepr pmap_sig.ALE_kernel_init_spec).
split.

intros.
unfold pmap_sig.ALE_kernel_init_spec, mem_sig.PE_kernel_init_spec in *; simpl in *; repeat autoclean.
unfold C_ALE.State.StackCheck, C_PE.State.StackCheck in *.
destruct s; simpl in *.
destruct stack; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean.
destruct (ALE_PE_mem_exists memory) as [bmem]; auto.
repeat autoclean.
exists {| C_ALE.State.stack := C_Stack.CallFrame nil :: stack; 
              C_ALE.State.memory := bmem |}; simpl.
split; auto.
red; simpl.
split; auto.

intros.
unfold pmap_sig.ALE_kernel_init_spec, mem_sig.PE_kernel_init_spec in *; simpl in *; repeat autoclean.
destruct s; simpl in *.
destruct (ALE_PE_mem_exists memory) as [bmem]; auto.
repeat autoclean.
destruct H0 with {| C_ALE.State.stack := stack; 
              C_ALE.State.memory := bmem |}; simpl in *; autosolve.
red; simpl.
split; auto.
split; auto.
Qed.


(******************************************************************)
Require Import label.

Definition ref := C_ALE_PE.EqRef.

Lemma WF_kernel_ale : forall kernel_code kernel_spec as_lib_code as_lib_spec
         (disj'' : C_PE.SCAP.SpecHeap.SpecHeap_disjoint PE_Lib.pe_lib_spec as_lib_spec),
         C_PE.SCAP.SpecHeap.SpecHeap_bigger as_lib_spec mem_sig.mem_spec ->
         C_PE.SCAP.SpecHeap.SpecHeap_disjoint
                 (C_PE.SCAP.SpecHeap.SpecHeap_join PE_Lib.pe_lib_spec as_lib_spec disj'')
                    (C_ALE_PE.PsiMake kernel_spec) ->
         C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine ALE_Lib.ale_lib_spec kernel_code kernel_spec ->
         
         C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine PE_Lib.pe_lib_spec as_lib_code as_lib_spec ->
         C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine 
                 (C_PE.SCAP.SpecHeap.SpecHeap_join PE_Lib.pe_lib_spec as_lib_spec disj'')
                                        (C_ALE_PE.CMake kernel_code) 
                                        (C_ALE_PE.PsiMake kernel_spec).
intros.
set (x := ref ALE_Lib.ale_lib_spec kernel_code kernel_spec).
eapply C_PE.SCAP.WFCode_stronghypo; eauto.
red; intros.
destruct l; autosolve; simpl.
apply tlb_reset_weak1.
red in H.
specialize (H mem_alloc_label); simpl in H.
unfold C_PE.SCAP.SpecHeap.SpecHeap_join; simpl.
makenew' (as_lib_spec mem_alloc_label); subst; auto.
apply alloc_weak1.

red in H.
specialize (H mem_free_label); simpl in H.
unfold C_PE.SCAP.SpecHeap.SpecHeap_join; simpl.
makenew' (as_lib_spec mem_free_label); subst; auto.
apply free_weak1.
Qed.

Lemma WF_kernel_ale' : forall kernel_code kernel_spec as_lib_code as_lib_spec
         (disj : C_PE.SCAP.SpecHeap.SpecHeap_disjoint (C_ALE_PE.PsiMake kernel_spec) as_lib_spec)
         (disj' : C_PE.SCAP.SpecHeap.SpecHeap_disjoint (C_ALE_PE.PsiMake kernel_spec) PE_Lib.pe_lib_spec)
         (disj'' : C_PE.SCAP.SpecHeap.SpecHeap_disjoint PE_Lib.pe_lib_spec as_lib_spec),
         C_PE.SCAP.SpecHeap.SpecHeap_bigger as_lib_spec mem_sig.mem_spec ->
         C_PE.SCAP.CH.CodeHeap_disjoint as_lib_code (C_ALE_PE.CMake kernel_code) ->
         C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine ALE_Lib.ale_lib_spec kernel_code kernel_spec ->
         C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine PE_Lib.pe_lib_spec as_lib_code as_lib_spec ->

      C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine 
                PE_Lib.pe_lib_spec 
                (C_PE.SCAP.CH.CodeHeap_join (C_ALE_PE.CMake kernel_code) as_lib_code) 
                (C_PE.SCAP.SpecHeap.SpecHeap_join (C_ALE_PE.PsiMake kernel_spec) as_lib_spec disj).
intros.
apply C_PE.SCAP.WFCode_linking_smp with disj''; auto.
eapply WF_kernel_ale; eauto.
apply C_PE.SCAP.SpecHeap.SpecHeap_disjoint_joinleft; auto.
apply C_PE.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
apply C_PE.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
Qed.
