(* as_lib.v *)
(* The stub library for the AS memory model *)
(* Includes mem_alloc, mem_free, as_request (reserve) , as_release *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import label.
Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import c.
Require Import "as".

Module AS_Lib.

Import C_AS.

Definition allocfail : SAct (State := State) :=
  StackAct (pop +++ setRet 0%Z).


Definition AS_alloc (pg : Z) : AS_State.t -> AS_State.t -> Prop :=
  fun s s' => 
	 (forall l, s.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\ 
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then true else s.(allocdata) x)).

Definition allocpg (pg : Z) : State -> State -> Prop :=
  fun s s' => AS_alloc pg s.(memory) s'.(memory) /\ StackRets pg s.(stack) s'.(stack).


Lemma AS_alloc_allocdata_unch : forall pg x0 bs bs',
pg <> x0 ->
AS_Lib.AS_alloc x0 bs bs' ->
AS_State.allocdata bs' pg = AS_State.allocdata bs pg.
intros.
red in H0; decompose record H0; clear H0.
rewrite H2; simpl.
Zeq_pull'; Zeq_norm; simpl; autosolve.
Qed.

Lemma AS_alloc_data_unch : forall l x0 bs bs',
AS_State.allocdata bs (Pg l) = true ->
AS_Lib.AS_alloc x0 bs bs' ->
AS_State.data bs' l = AS_State.data bs l.
intros.
red in H0; decompose record H0; clear H0.
rewrite H1; autosolve.
Qed.

Lemma AS_alloc_newpage_physical : forall x0 l bs bs',
AS_Lib.AS_alloc x0 bs bs' -> 
AS_State.allocdata bs (Pg l) <> true ->
AS_State.allocdata bs' (Pg l) = true ->
x0 = Pg l.

intros.
red in H; decompose record H; clear H.
rewrite H3 in H1; clear H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.



Definition mem_alloc_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 0) s = true /\ StackCheck isFrame s = true
     );

rel := (fun s s' =>
    (exists pg, (1 <= pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = false /\ allocpg pg s s')
    \/
    s' = allocfail s)
|}.



Definition AS_free (pg : Z) : AS_State.t -> AS_State.t -> Prop :=
  fun s s' => 
	 (forall l, s'.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\ 
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then false else s.(allocdata) x)).

Definition freepg (pg : Z) : State -> State -> Prop :=
   fun s s' => AS_free pg s.(memory) s'.(memory) /\ StackRets 0%Z s.(stack) s'.(stack).


Lemma AS_free_data_unch : forall l x0 s s',
AS_State.allocdata s' (Pg l) = true ->
AS_Lib.AS_free x0 s s' ->
AS_State.data s' l = AS_State.data s l.
intros.
red in H0; decompose record H0; clear H0.
rewrite H1; autosolve.
Qed.

Lemma AS_free_allocdata_unch : forall pg x0 s s',
pg <> x0 ->
AS_Lib.AS_free x0 s s' ->
AS_State.allocdata s' pg = AS_State.allocdata s pg.
intros.
red in H0; decompose record H0; clear H0.
rewrite H2.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma AS_free_allocdata_pullback : forall pg x0 s s',
AS_Lib.AS_free x0 s s' ->
AS_State.allocdata s' pg = true ->
AS_State.allocdata s pg = true.
intros.
red in H; decompose record H; clear H.
rewrite H2 in *; clear H2.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Definition mem_free_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                 let pg := nth 0 (StackGet getArgs s) 0%Z in
                                 (0 < pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = true);

rel := (fun s s' => freepg (nth 0 (StackGet getArgs s) 0%Z) s s')
|}.

Definition as_reserve_spec : Action State := {|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                  let vpage := nth 0 (StackGet getArgs s) 0%Z in
                                  ValidVPg vpage = true /\ ValidPg vpage = false /\ s.(memory).(allocdata) vpage = false
                                  );
rel := (fun s s' =>
       let vpage := nth 0 (StackGet getArgs s) 0%Z in
       (allocpg vpage s s')  \/ (s' = StackAct (pop +++ setRet 0%Z) s))
|}.

Definition as_release_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                 let pg := nth 0 (StackGet getArgs s) 0%Z in
                                 ValidVPg pg = true /\ ValidPg pg = false /\ s.(memory).(allocdata) pg = true);

rel := (fun s s' => freepg (nth 0 (StackGet getArgs s) 0%Z) s s')
|}.

Definition as_lib_spec : SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free_spec
  | mem_alloc_label => Some mem_alloc_spec
  | as_reserve_label => Some as_reserve_spec
  | as_release_label => Some as_release_spec
  | _ => None
  end.

End AS_Lib.
