(* as_pmap.v *)
(* The relation between the AS and PMAP memory models *)
(* Also includes the proofs of compatibility between AS and PMAP for alloc and free *)
(* For as_request and as_release compatibility is between AS library and implemntation of as functions
    in  as_sig.v *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.
Require Import IfProp.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.

Require Import mem_repr.
Require Import "as".
Require Import pmap.

Module AS_PMAP <: Mem_Repr_T AS_State PMAP_State.

Definition Repr_Mem (asm : AS_State.t) (pmap : PMAP_State.t) :=
   (forall l, ValidPA l = true -> asm.(AS_State.allocdata) (Pg l) = true -> asm.(AS_State.data) l = pmap.(PMAP_State.data) l) /\
   (forall l, ValidVA l = true -> ValidPA l = false -> asm.(AS_State.allocdata) (Pg l) = true ->
                           asm.(AS_State.data) l = pmap.(PMAP_State.data) (PMAP_State.trans l pmap)) /\
   (forall pg, ValidPg pg = true -> asm.(AS_State.allocdata) pg = true -> pmap.(PMAP_State.allocdata) pg = true) /\
   (forall pg, ValidVPg pg = true -> ValidPg pg = false -> asm.(AS_State.allocdata) pg = true ->
                         match (pmap.(PMAP_State.pagemap) pg) with
                          | Some ppg => ValidPg ppg = true /\ ppg <> 0%Z /\ pmap.(PMAP_State.allocdata) ppg = true 
                          | None => False
                          end) /\
   (forall pg, ValidVPg pg = true -> ValidPg pg = false -> asm.(AS_State.allocdata) pg = false ->
                         pmap.(PMAP_State.pagemap) pg = None) /\
   (forall l l', l <> l' -> ValidVA l = true -> ValidVA l' = true -> 
                        asm.(AS_State.allocdata) (Pg l) = true -> asm.(AS_State.allocdata) (Pg l') = true ->
                        PMAP_State.trans' l pmap <> PMAP_State.trans' l' pmap)
.

Lemma Repr_addrCheck : forall (m1 : AS_State.t) (m2 : PMAP_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, AS_State.addrCheck l m1 = true -> PMAP_State.vaCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold AS_State.addrCheck, PMAP_State.vaCheck in *; repeat autoclean.
unfold isPhys.
unfold PMAP_State.paCheck, PMAP_State.transCheck; simpl.
makenew' (ValidPg (Pg l)); autosolve.
rewrite ValidVA_Phys; autosolve; simpl.
apply H2; autosolve.
(* virt *)
rewrite H0; simpl; autosolve.
specialize (H3 (Pg l)).
unfold PMAP_State.trans in *.
assert (ValidVPg (Pg l) = true).
apply ValidVA_ValidVPg; autosolve.
specialize (H3 H7 H8 H6).
makenew' (PMAP_State.pagemap m2 (Pg l)); autosolve.
repeat autoclean.
rewrite ValidPA_parts; autosolve; simpl.
rewrite Pg_simpl; auto.
Qed.

Lemma Repr_getMemCheck : forall (m1 : AS_State.t) (m2 : PMAP_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, AS_State.getMemCheck l m1 = true -> PMAP_State.getMemCheck l m2 = true.
apply Repr_addrCheck.
Qed.

Lemma Repr_getMem : forall (m1 : AS_State.t) (m2 : PMAP_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v, AS_State.getMemCheck l m1 = true -> AS_State.getMem l m1 = v -> PMAP_State.getMem l m2 = v.
intros.
unfold AS_State.getMemCheck, AS_State.getMem, PMAP_State.getMem in *.
unfold isPhys.
red in H; repeat autoclean.
unfold AS_State.addrCheck in H0; repeat autoclean.
makenew' (ValidPg (Pg l)); autosolve.
(* phys *)
rewrite H; autosolve.
apply ValidVA_Phys; autosolve.
(* virt *)
rewrite H2; autosolve.
makenew' (ValidPA l); autosolve.
rewrite ValidPA_ValidPg in H8; autosolve.
Qed.

Lemma Repr_setMemCheck : forall (m1 : AS_State.t) (m2 : PMAP_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, AS_State.setMemCheck l m1 = true -> PMAP_State.setMemCheck l m2 = true.
apply Repr_addrCheck.
Qed.

Lemma PMAP_getMemCheck_setMem_unch : forall l0 l v m2, 
PMAP_State.getMemCheck l0 (PMAP_State.setMem l v m2) = PMAP_State.getMemCheck l0 m2.
intros.
unfold PMAP_State.setMem, PMAP_State.updateMem; simpl.
unfold PMAP_State.getMemCheck; simpl.
unfold PMAP_State.vaCheck; simpl.
makenew' (isPhys l0); makenew' (isPhys l); simpl; autosolve.
Qed.

Lemma AS_getMemCheck_setMem_unch : forall l0 l v m2, 
AS_State.getMemCheck l0 (AS_State.setMem l v m2) = AS_State.getMemCheck l0 m2.
intros.
unfold AS_State.getMemCheck; simpl.
unfold AS_State.addrCheck; simpl; autosolve.
Qed.

Lemma ValidPg_virt : forall l, ValidPA l = false -> ValidVA l = true -> ValidPg (Pg l) = false.
intros.
unfold ValidPA, ValidVA, ValidPg in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H in H0; autosolve.
Qed.

Lemma ValidPg_separate : forall pg pg', ValidPg pg = true -> ValidPg pg' = false -> pg <> pg'.
intros.
unfold ValidPg in *.
red; intros; subst.
rewrite H0 in H; autosolve.
Qed.

Lemma Repr_setMem : forall (m1 : AS_State.t) (m2 : PMAP_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v , AS_State.setMemCheck l m1 = true -> 
         Repr_Mem (AS_State.setMem l v m1) (PMAP_State.setMem l v m2).
intros.
unfold AS_State.setMemCheck, AS_State.setMem, PMAP_State.setMem in *.
red in H; repeat autoclean.
unfold AS_State.addrCheck in *; repeat autoclean.
red; simpl.
repeat split; autosolve.

intros.
unfold isPhys.
makenew' (ValidPg (Pg l)).
unfold PMAP_State.updateMem; simpl.
rewrite H; autosolve.
(* virt *)
simpl.
assert (ValidVPg (Pg l) = true).
apply ValidVA_ValidVPg; autosolve.
specialize (H3 (Pg l) H9 H10 H6).
unfold PMAP_State.trans.
makenew' (PMAP_State.pagemap m2 (Pg l)); autosolve.
rewrite H; autosolve.
assert  (ValidPA l = false).
makenew' (ValidPA l); autosolve.
rewrite ValidPA_ValidPg in H10; autosolve.
set (x := ValidPA_separate l0 l H7 H11).
Zeq_pull'; Zeq_norm; simpl; autosolve.
elimtype False.
apply (H5 l l0); autosolve.
apply ValidPA_ValidVA; auto.
unfold PMAP_State.trans', isPhys.
rewrite H10.
assert (ValidPg (Pg l0) = true).
apply ValidPA_ValidPg; auto.
rewrite H13.
unfold PMAP_State.trans.
rewrite H12; auto.

intros.
unfold isPhys.
destruct (Zeq_dec l l0).
subst.
rewrite ValidPg_virt; autosolve.
rewrite PMAP_State.trans_update_unch.
unfold PMAP_State.updateMem; simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* l <> l0 *)
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (ValidPg (Pg l)); autosolve.
(* l is physical *)
rewrite PMAP_State.trans_update_unch.
unfold PMAP_State.updateMem; simpl.
rewrite H1; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
specialize (H5 l l0 n H0 H7 H6 H9).
unfold PMAP_State.trans' in *; simpl.
unfold isPhys in H5.
rewrite H12 in H5.
rewrite ValidPg_virt in H5; autosolve.
(* l is virtual *)
rewrite PMAP_State.trans_update_unch.
unfold PMAP_State.updateMem; simpl.
rewrite H1; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
specialize (H5 l l0 n H0 H7 H6 H9).
unfold PMAP_State.trans' in *; simpl.
unfold isPhys in H5.
rewrite H12 in H5.
rewrite ValidPg_virt in H5; autosolve.

intros.
unfold isPhys.
makenew' (ValidPg (Pg l)); autosolve.
(* l is Phys *)
unfold PMAP_State.updateMem; simpl.
apply H2; autosolve.
(* l is virt *)
unfold PMAP_State.updateMem; simpl.
apply H2; autosolve.

intros.
unfold isPhys.
makenew' (ValidPg (Pg l)); autosolve.
(* l is phys *)
unfold PMAP_State.updateMem; simpl.
apply H3; autosolve.
(* l is virt *)
unfold PMAP_State.updateMem; simpl.
apply H3; autosolve.

intros.
unfold isPhys.
makenew' (ValidPg (Pg l)); autosolve.
(* l is phys *)
unfold PMAP_State.updateMem; simpl.
apply H4; autosolve.
(* l is virt *)
unfold PMAP_State.updateMem; simpl.
apply H4; autosolve.


intros.
unfold isPhys.
makenew' (ValidPg (Pg l)); autosolve.
rewrite PMAP_State.trans'_update_unch; autosolve.
apply H5; autosolve.
rewrite PMAP_State.trans'_update_unch; autosolve.
apply H5; autosolve.
Qed.

End AS_PMAP.


Require Import c.

Module C_AS_PMAP := C_Refine AS_State PMAP_State AS_PMAP C_AS C_PMAP.

Require Import as_lib.
Require Import as_sig.
Require Import pmap_lib.

Lemma exists_alloc : forall pg newpagesrc s, 
   s.(C_AS.State.memory).(AS_State.allocdata) pg = false ->
   exists s', 
   AS_Lib.allocpg pg s s' /\  (forall l, Pg l = pg -> (C_AS.State.memory s').(AS_State.data) l = newpagesrc l).
intros.
unfold AS_Lib.allocpg.
unfold AS_Lib.AS_alloc.
unfold C_AS.State.Stack.StackRets.
destruct s; simpl in *.
destruct memory; simpl in *.
exists 
{| C_AS.State.stack := (C_AS.State.Stack.pop +++ C_AS.State.Stack.setRet pg) stack;
    C_AS.State.memory := 
         {| AS_State.data := (fun l => if Zeq_bool (Pg l) pg then newpagesrc l else data l);
             AS_State.allocdata := (fun x : Z => if Zeq_bool x pg then true else allocdata x)
          |}
|}; simpl.
repeat split; auto.
intros; auto.
Zeq_pull'; Zeq_norm; repeat autoclean; simpl in *; autosolve.
rewrite H0 in H; autosolve.
intros.
rewrite H0; rewrite Zeq_bool_refl; auto.
Qed.

Lemma vvv : forall x0 bs s bs' s',
  (1 <= x0 < NPAGES)%Z ->
  s.(PMAP_State.allocdata) x0 = false ->
  (forall l, (Pg l = x0) -> bs'.(AS_State.data) l = 
                                s'.(PMAP_State.data) l ) ->
  AS_PMAP.Repr_Mem bs s ->
  AS_Lib.AS_alloc x0 bs bs' ->
  PMAP_Lib.PMAP_alloc x0 s s' ->
  AS_PMAP.Repr_Mem bs' s'.
intros.
assert (ValidPg x0 = true).
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.
red in H2; repeat autoclean.
red; repeat split.

intros.
destruct (Zeq_dec (Pg l) x0).
(* we are in the newly alloced page *)
apply H1; autosolve.
(* other pages *)
rewrite AS_Lib.AS_alloc_allocdata_unch with _ x0 bs bs' in H13; autosolve.
rewrite AS_Lib.AS_alloc_data_unch with _ x0 bs _; autosolve.
rewrite PMAP_Lib.PMAP_alloc_data_unch with _ x0 s _; autosolve.
apply H2; autosolve.
apply H7; autosolve.
apply ValidPA_ValidPg; autosolve.

intros.
rewrite <- PMAP_Lib.PMAP_alloc_trans_unch with x0 _ s _; autosolve.
destruct (Zeq_dec (Pg l) x0).
subst.
rewrite AS_PMAP.ValidPg_virt in H5; autosolve.
rewrite AS_Lib.AS_alloc_allocdata_unch with _ x0 bs _ in *; autosolve.
rewrite AS_Lib.AS_alloc_data_unch with _ x0 bs _; autosolve.
rewrite PMAP_Lib.PMAP_alloc_data_unch with _ x0 s _; autosolve.
apply H6; autosolve.
(* the l can not be a part of an newly allocated page, because
    then it would be unallocated in PMAP *)
unfold PMAP_State.trans.
assert (ValidVPg (Pg l) = true); try apply ValidVA_ValidVPg; autosolve.
assert (ValidPg (Pg l) = false); try apply AS_PMAP.ValidPg_virt; autosolve.
specialize (H8 (Pg l) H15 H16 H14).
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
rewrite Pg_simpl; autosolve.

intros.
destruct (Zeq_dec pg x0); autosolve.
(* new page *)
subst.
red in H4; decompose record H4; clear H4.
rewrite H16; simpl.
Zeq_pull'; Zeq_norm; autosolve.
(* old pages *)
rewrite PMAP_Lib.PMAP_alloc_allocdata_unch with _ x0 s _ in *; autosolve.
apply H7; autosolve.
rewrite AS_Lib.AS_alloc_allocdata_unch with _ x0 bs _ in *; autosolve.

intros.
destruct (Zeq_dec pg x0); autosolve.
(* new page *)
subst.
rewrite H5 in H13; autosolve.
(* old pages *)
rewrite AS_Lib.AS_alloc_allocdata_unch with _ x0 bs _ in *; autosolve.
specialize (H8 pg H12 H13 H14).
rewrite PMAP_Lib.PMAP_alloc_pagemap_unch with x0 s _; auto.
makenew' (PMAP_State.pagemap s pg); auto.
rewrite PMAP_Lib.PMAP_alloc_allocdata_unch with _ x0 s _ in *; autosolve.
red; intros; subst.
rewrite H0 in H8; autosolve.

intros.
destruct (Zeq_dec pg x0); autosolve.
(* new page *)
subst.
rewrite H5 in H13; autosolve.
(* old pages *)
rewrite PMAP_Lib.PMAP_alloc_pagemap_unch with x0 s _; auto.
apply H9; autosolve.
rewrite AS_Lib.AS_alloc_allocdata_unch with _ x0 bs _ in *; autosolve.

intros.
destruct (bool_dec (AS_State.allocdata bs (Pg l)) true).
destruct (bool_dec (AS_State.allocdata bs (Pg l')) true).
(* both l and l' were previously allocated *)
rewrite <- PMAP_Lib.PMAP_alloc_trans'_unch with x0 l s s'; autosolve.
rewrite <- PMAP_Lib.PMAP_alloc_trans'_unch with x0 l' s s'; autosolve.
apply H10; autosolve.
(* l allocated before alloc, l' was not previously allocated *)
assert (Pg l' = x0).
rewrite AS_Lib.AS_alloc_newpage_physical with x0 l' bs bs'; auto.
subst.
unfold PMAP_State.trans'.
unfold isPhys.
rewrite H5.
makenew' (ValidPg (Pg l)); autosolve.
rewrite <- PMAP_Lib.PMAP_alloc_trans_unch with (Pg l') l s _; autosolve.
red; intros.
unfold PMAP_State.trans in *.
assert (ValidVPg (Pg l) = true).
apply ValidVA_ValidVPg; autosolve.
specialize (H8 (Pg l) H19 H18 e).
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
rewrite <- H17 in *.
rewrite Pg_simpl in *.
repeat autoclean.
rewrite H22 in H0; autosolve.
(* l not previously allocated *)
assert (Pg l = x0).
rewrite AS_Lib.AS_alloc_newpage_physical with x0 l bs bs'; auto.
subst.
unfold PMAP_State.trans'.
unfold isPhys.
rewrite H5.
destruct (bool_dec (AS_State.allocdata bs (Pg l')) true).
(* l' previously allocated *)
makenew' (ValidPg (Pg l')); autosolve.
rewrite <- PMAP_Lib.PMAP_alloc_trans_unch with (Pg l) l' s _; autosolve.
red; intros.
unfold PMAP_State.trans in *.
assert (ValidVPg (Pg l') = true).
apply ValidVA_ValidVPg; autosolve.
specialize (H8 (Pg l') H19 H18 e).
makenew' (PMAP_State.pagemap s (Pg l')); autosolve.
rewrite H17 in *.
rewrite Pg_simpl in *.
repeat autoclean.
rewrite H22 in H0; autosolve.
(* both l and l' are not previously allocated (thus their pages are same *)
assert (Pg l' = Pg l).
rewrite AS_Lib.AS_alloc_newpage_physical with (Pg l) l' bs bs'; auto.
rewrite H17 in *.
rewrite H5; autosolve.
Qed.

Lemma ttt : forall x0 bs s bs' s',
  (1 <= x0 < NPAGES)%Z ->
  s.(C_PMAP.State.memory).(PMAP_State.allocdata) x0 = false ->
  (forall l, (Pg l = x0) -> bs'.(C_AS.State.memory).(AS_State.data) l = 
                                s'.(C_PMAP.State.memory).(PMAP_State.data) l ) ->
  C_AS_PMAP.CRepr bs s ->
  AS_Lib.allocpg x0 bs bs' ->
  PMAP_Lib.allocpg x0 s s' ->
  C_AS_PMAP.CRepr bs' s'.
intros.
red in H2; decompose record H2; clear H2.
red in H3; decompose record H3; clear H3.
red in H4; decompose record H4; clear H4.
red; split.
rewrite H5 in H7.
red in H7, H8.
rewrite H7; rewrite H8; autosolve.

(****)
clear H7 H8 H5.
eapply vvv; eauto.
Qed.

Lemma AS_PMAP_allocdata : forall x0 bs s,
ValidPg x0 = true ->
AS_PMAP.Repr_Mem bs s ->
PMAP_State.allocdata s x0 = false ->
AS_State.allocdata bs x0 = false.
intros.
red in H0; repeat autoclean.
makenew' (AS_State.allocdata bs x0); autosolve.
rewrite H3 in H1; autosolve.
Qed.

Lemma alloc_weak1 : ActionGTE (C_AS_PMAP.RF.reprAction C_AS_PMAP.CRepr AS_Lib.mem_alloc_spec)
                                  PMAP_Lib.mem_alloc_spec.
split.
(* pre *)
intros.
unfold AS_Lib.mem_alloc_spec, PMAP_Lib.mem_alloc_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_AS.State.StackCheck, C_PMAP.State.StackCheck in *; simpl in *.
rewrite H in *.
split; autosolve.
(* rel *)
intros.
unfold AS_Lib.mem_alloc_spec, PMAP_Lib.mem_alloc_spec in *; simpl in *; repeat autoclean.
inversion_clear H0; repeat autoclean. 

(* positive case - allocated x0 *)
destruct (exists_alloc x0 s'.(C_PMAP.State.memory).(PMAP_State.data) bs) as [bs'].
(* exists_alloc: need to show that x0 is currently unallocated *)
destruct H3.
red in H9; decompose record H9; clear H9.
makenew' (AS_State.allocdata (C_AS.State.memory bs) x0); autosolve.
rewrite H11 in H6; autosolve.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(*****)

exists bs'; repeat autoclean.
split.
eapply ttt; eauto.
left.
exists x0.
repeat (split; autosolve).
red in H3; repeat autoclean.
apply AS_PMAP_allocdata with (C_PMAP.State.memory s); autosolve.
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.

(* negative case *)
exists (AS_Lib.allocfail bs).
split.
red in H3; repeat autoclean.
unfold AS_Lib.allocfail, PMAP_Lib.allocfail; simpl.
red; split; simpl; autosolve.
rewrite <- H0; autosolve.
right; autosolve.
Qed.

Lemma exists_free : forall pg s, exists s', AS_Lib.freepg pg s s'.
intros.
unfold AS_Lib.freepg.
unfold AS_Lib.AS_free; simpl.
unfold C_AS.State.Stack.StackRets; simpl.
destruct s; simpl.
destruct memory; simpl.
exists 
{| C_AS.State.stack := (C_AS.State.Stack.pop +++ C_AS.State.Stack.setRet 0%Z) stack;
    C_AS.State.memory := 
         {| AS_State.data := data;
             AS_State.allocdata := (fun x : Z => if Zeq_bool x pg then false else allocdata x)
          |}
|}; simpl.
repeat split; auto.
Qed.

Lemma ReprMem_free_preserve : forall pg bs bs' s s',
ValidPg pg = true ->
AS_State.allocdata bs pg = true ->
AS_PMAP.Repr_Mem bs s -> 
PMAP_Lib.PMAP_free pg s s' ->
AS_Lib.AS_free pg bs bs' ->
AS_PMAP.Repr_Mem bs' s'.
intros.
red in H1; repeat autoclean.
red; repeat split.

intros.
rewrite AS_Lib.AS_free_data_unch with l pg bs bs'; autosolve.
rewrite PMAP_Lib.PMAP_free_data_unch with l pg s s'; autosolve.
apply H1; auto.
red in H3; decompose record H3; clear H3.
rewrite H12 in H10.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red in H2; decompose record H2; clear H2.
rewrite H13.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
red in H3; decompose record H3; clear H3.
rewrite H12 in H10.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H5; autosolve.
apply ValidPA_ValidPg; autosolve.
red in H3; decompose record H3; clear H3.
rewrite H15 in H10.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros.
destruct (Zeq_dec (Pg l) pg).
(* Case : page we are destroying is the freed one *)
rewrite e in *.
red in H3; decompose record H3; clear H3.
rewrite H13 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* The page is not the freed one *)
rewrite AS_Lib.AS_free_data_unch with l pg bs bs'; autosolve.
red in H3; decompose record H3; clear H3.
rewrite H13 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
specialize (H4 l H9 H10 H11).
rewrite H4.
unfold PMAP_State.trans.
red in H2; decompose record H2; clear H2.
rewrite H17.
assert (ValidVPg (Pg l) = true).
apply ValidVA_ValidVPg; autosolve.
assert (ValidPg (Pg l) = false).
apply AS_PMAP.ValidPg_virt; auto.
specialize (H6 (Pg l) H2 H15 H11).
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
rewrite <- H3; autosolve.
rewrite Pg_simpl.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* Case: Pg l maps to pg  - the page we are freeing *)
(* can not happen, since then two valid pages would map to same page *)
subst.
specialize (H8 l (pg * PGSIZE + Off l)%Z).
assert (ValidVA (pg * PGSIZE + Off l) = true).
apply ValidVAComp; auto.
apply ValidPg_ValidVPg; auto.
assert (l <> (pg * PGSIZE + Off l)%Z).
red; intros.
rewrite H20 in H14; rewrite Pg_simpl in *; autosolve.
specialize (H8 H20 H9 H18).
elimtype False; apply H8; autosolve.
rewrite Pg_simpl; auto.
unfold PMAP_State.trans'.
unfold isPhys.
rewrite Pg_simpl; auto.
rewrite H15.
rewrite H.
unfold PMAP_State.trans.
rewrite H19; auto.

intros.
red in H2; decompose record H2; clear H2.
rewrite H13.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
red in H3; decompose record H3; clear H3.
rewrite H12 in H10.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H5; autosolve.
red in H3; decompose record H3; clear H3.
rewrite H15 in H10.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

intros.
red in H3; decompose record H3; clear H3.
rewrite H13 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
specialize (H6 pg0 H9 H10 H11).
red in H2; decompose record H2; clear H2.
rewrite H17.
makenew' (PMAP_State.pagemap s pg0); autosolve.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
destruct (ValidVPg_existsVA pg0); auto.
destruct H2; subst.
assert (ValidVA (pg * PGSIZE + Off x) = true).
apply ValidVAComp; auto.
apply ValidPg_ValidVPg; autosolve.
elimtype False.
apply H8 with x (pg * PGSIZE + Off x)%Z; try rewrite Pg_simpl in *; autosolve.
red; intros.
rewrite H19 in H14; rewrite Pg_simpl in H14; autosolve.
unfold PMAP_State.trans', isPhys.
rewrite H10.
rewrite Pg_simpl.
rewrite H.
unfold PMAP_State.trans.
rewrite H15; auto.

intros.
red in H2; decompose record H2; clear H2.
rewrite H15.
apply H7; auto.
red in H3; decompose record H3; clear H3.
rewrite H13 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H in H10; autosolve.

intros.
red in H2; decompose record H2; clear H2.
red in H3; decompose record H3; clear H3.
rewrite H15 in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
specialize (H8 l l' H9 H10 H11 H12 H13).
unfold PMAP_State.trans', isPhys in *.
makenew' (ValidPg (Pg l)); makenew' (ValidPg (Pg l')); autosolve.
unfold PMAP_State.trans in *.
rewrite H17; autosolve.
unfold PMAP_State.trans in *.
rewrite H17; autosolve.
unfold PMAP_State.trans in *.
rewrite H17; autosolve.
Qed.

Lemma freepg_repr_preserve : forall pg bs bs' s s',
ValidPg pg = true ->
AS_State.allocdata (C_AS.State.memory bs) pg = true ->
C_AS_PMAP.CRepr bs s -> 
PMAP_Lib.freepg pg s s' ->
AS_Lib.freepg pg bs bs' ->
C_AS_PMAP.CRepr bs' s'.
intros.
red in H1; repeat autoclean.
red in H2; repeat autoclean.
red in H3; repeat autoclean.
red; split.
rewrite H6.
rewrite H5.
rewrite H1; autosolve.
eapply ReprMem_free_preserve; eauto.
Qed.


Lemma free_weak1 : ActionGTE (C_AS_PMAP.RF.reprAction C_AS_PMAP.CRepr AS_Lib.mem_free_spec)
                                  PMAP_Lib.mem_free_spec.
split.
(* pre *)
intros.
unfold AS_Lib.mem_free_spec, PMAP_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_AS.State.StackCheck, C_PMAP.State.StackCheck in *; simpl in *.
unfold C_AS.State.StackGet, C_PMAP.State.StackGet in *; simpl in *.
rewrite H in *.
repeat split; autosolve.
red in H4; repeat autoclean.
assert (ValidPg (nth 0 (C_AS.State.Stack.getArgs (C_PMAP.State.stack s)) 0%Z) = true).
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H6; autosolve.

(* rel *)
intros.
unfold AS_Lib.mem_free_spec, PMAP_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
destruct (exists_free (nth 0 (C_AS.State.StackGet C_AS.State.Stack.getArgs bs) 0%Z) bs) as [bs'].
exists bs'; split; auto.

eapply freepg_repr_preserve; eauto.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold C_AS.State.StackGet, C_PMAP.State.StackGet in *; simpl in *.
red in H5; repeat autoclean.
rewrite H5; autosolve.
Qed.


(* virtual allocation *)

Lemma vvvv : forall vpg ppg bs bs' s s',
ValidVPg vpg = true ->
ValidPg vpg = false ->
ValidPg ppg = true ->
ppg <> 0%Z ->
AS_State.allocdata bs vpg = false ->
PMAP_State.allocdata s ppg = false ->
(forall l, (Pg l = vpg) -> bs'.(AS_State.data) l = 
                                s'.(PMAP_State.data) (ppg * PGSIZE + Off l) ) ->
AS_PMAP.Repr_Mem bs s ->
AS_Lib.AS_alloc vpg bs bs' ->
as_sig.PMAP_valloc vpg ppg s s' ->
AS_PMAP.Repr_Mem bs' s'.
intros.
assert (vpg <> ppg).
red; intros; subst.
rewrite H1 in H0; autosolve.

red in H6; repeat autoclean.
red; repeat split; autosolve.

(* equality of physical ranges *)
intros.
destruct (Zeq_dec (Pg l) vpg).
assert (ValidPg vpg = true).
rewrite <- e.
apply ValidPA_ValidPg; autosolve.
rewrite H17 in H0; autosolve.
(* Pg l <> vpg  - since l is physical *)
rewrite AS_Lib.AS_alloc_allocdata_unch with (Pg l) vpg bs bs' in H16; autosolve.
rewrite AS_Lib.AS_alloc_data_unch with l vpg bs bs'; autosolve.
rewrite as_sig.PMAP_valloc_data_unch with l vpg ppg s s'; autosolve.
apply H6; autosolve.
apply H11; autosolve.
apply ValidPA_ValidPg; autosolve.

(* equality of virtual ranges *)
intros.
destruct (Zeq_dec (Pg l) vpg).
(* Pg l = vpg - use our premise that the new state is equal on that page *)
rewrite H5; autosolve.
unfold PMAP_State.trans.
red in H8; repeat autoclean.
rewrite H19.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* Pg l <> vpg - use the repr to guarantee equality *)
rewrite AS_Lib.AS_alloc_allocdata_unch with (Pg l) vpg bs bs' in H17; autosolve.
rewrite AS_Lib.AS_alloc_data_unch with l vpg bs bs'; autosolve.
rewrite <- as_sig.PMAP_valloc_trans_unch with vpg ppg l s s'; autosolve.
rewrite as_sig.PMAP_valloc_data_unch with (PMAP_State.trans l s) vpg ppg s s'; autosolve.
apply H10; autosolve.
assert (ValidVPg (Pg l) = true); try apply ValidVA_ValidVPg; autosolve.
assert (ValidPg (Pg l) = false); try apply InvalidPA_InvalidPg; autosolve.
specialize (H12 (Pg l) H18 H19 H17).
unfold PMAP_State.trans.
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
repeat autoclean.
rewrite Pg_simpl; autosolve.

(* allocdata for physical pages *)
intros.
assert (pg <> vpg).
red; intros; subst.
rewrite H15 in H0; autosolve.
destruct (Zeq_dec pg ppg).
(* pg = ppg *)
subst.
red in H8; repeat autoclean.
rewrite H18; simpl.
Zeq_pull'; Zeq_norm; autosolve.
(* pg <> ppg *)
rewrite as_sig.PMAP_valloc_allocdata_unch with vpg ppg pg s s'; autosolve.
apply H11; autosolve.
rewrite <- AS_Lib.AS_alloc_allocdata_unch with pg vpg bs bs'; autosolve.

(* page mappings for the valid virt pages *)
intros.
destruct (Zeq_dec pg vpg).
(* pg = vpg *)
subst.
red in H8; repeat autoclean.
rewrite H19.
rewrite H18.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
repeat split; auto.
(* pg <> vpg *)
rewrite AS_Lib.AS_alloc_allocdata_unch with pg vpg bs bs' in H17; autosolve.
specialize (H12 pg H15 H16 H17).
rewrite as_sig.PMAP_valloc_pagemap_unch with vpg ppg pg s s'; autosolve.
makenew' (PMAP_State.pagemap s pg); autosolve.
red in H8; repeat autoclean.
rewrite H18.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst; repeat split; auto.
repeat split; auto.

intros.
destruct (Zeq_dec pg vpg).
(* pg = vpg *)
subst.
red in H7; repeat autoclean.
rewrite H18 in H17.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* pg <> vpg *)
rewrite as_sig.PMAP_valloc_pagemap_unch with vpg ppg pg s s'; autosolve.
apply H13; autosolve.
red in H7; repeat autoclean.
rewrite H18 in H17.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* the non mapping case *)
intros.
destruct (bool_dec (AS_State.allocdata bs (Pg l)) true);
destruct (bool_dec (AS_State.allocdata bs (Pg l')) true).
(* both l and l' were previously allocated *)
specialize (H14 l l' H15 H16 H17).
assert (Pg l <> vpg).
red; intros; subst.
rewrite e in H3; autosolve.
rewrite <- as_sig.PMAP_valloc_trans'_unch with vpg ppg l s s'; autosolve.
rewrite <- as_sig.PMAP_valloc_trans'_unch with vpg ppg l' s s'; autosolve.
apply H14; autosolve.
red; intros; subst.
rewrite e0 in H3; autosolve.
(* l previously allocated, l' is not *)
assert (vpg = Pg l').
apply AS_Lib.AS_alloc_newpage_physical with bs bs'; autosolve.
subst. (*rewrite H20 in *; clear vpg H20. *)
unfold PMAP_State.trans', isPhys.
rewrite H0; simpl.
makenew' (ValidPg (Pg l)); autosolve.

assert (ValidVPg (Pg l) = true).
apply ValidVA_ValidVPg; autosolve.
unfold PMAP_State.trans.
red in H8; repeat autoclean.
rewrite H23.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red; intros; subst; rewrite Pg_simpl in *.
rewrite H11 in H4; autosolve.

rewrite <- as_sig.PMAP_valloc_trans_unch with (Pg l') ppg l s _; autosolve.
unfold PMAP_State.trans.
red in H8; repeat autoclean.
rewrite H22.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
assert (ValidVPg (Pg l) = true).
apply ValidVA_ValidVPg; autosolve.
specialize (H12 (Pg l) H23 H21 e).
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
red; intros.

assert ((Pg (z * PGSIZE + Off l)%Z) = Pg (ppg * PGSIZE + Off l')%Z).
f_equal; auto.
repeat rewrite Pg_simpl in H27.
subst.
decompose record H12; clear H12.
rewrite H30 in H4; autosolve.

red; intros.
rewrite <- H20 in *.
rewrite e in n; autosolve.
apply n; auto.

(* l is not previously allocated, l' is *)
assert (vpg = Pg l).
apply AS_Lib.AS_alloc_newpage_physical with bs bs'; autosolve.
subst. (*rewrite H20 in *; clear vpg H20. *)
unfold PMAP_State.trans', isPhys.
rewrite H0; simpl.
makenew' (ValidPg (Pg l')); autosolve.

assert (ValidVPg (Pg l') = true).
apply ValidVA_ValidVPg; autosolve.
unfold PMAP_State.trans.
red in H8; repeat autoclean.
rewrite H23.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red; intros; subst; rewrite Pg_simpl in *.
rewrite H11 in H4; autosolve.

rewrite <- as_sig.PMAP_valloc_trans_unch with (Pg l) ppg l' s _; autosolve.
unfold PMAP_State.trans.
red in H8; repeat autoclean.
rewrite H22.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
assert (ValidVPg (Pg l') = true).
apply ValidVA_ValidVPg; autosolve.
specialize (H12 (Pg l') H23 H21 e).
makenew' (PMAP_State.pagemap s (Pg l')); autosolve.
red; intros.

assert ((Pg (ppg * PGSIZE + Off l)%Z) = Pg (z * PGSIZE + Off l')%Z).
f_equal; auto.
repeat rewrite Pg_simpl in H27.
subst.
decompose record H12; clear H12.
rewrite H30 in H4; autosolve.

red; intros.
rewrite <- H20 in *.
rewrite e in n; autosolve.
apply n; auto.

(* neither l or l' are previously allocated *)
assert (Pg l = vpg).
rewrite AS_Lib.AS_alloc_newpage_physical with vpg l bs bs'; auto.
assert (Pg l' = vpg).
rewrite AS_Lib.AS_alloc_newpage_physical with vpg l' bs bs'; auto.
assert (Pg l = Pg l'); autosolve.
rewrite <- H20 in *; rewrite <- H22 in *.
clear H20 H21 vpg.
unfold PMAP_State.trans', isPhys.
rewrite <- H22.
makenew' (ValidPg (Pg l)); autosolve.
unfold PMAP_State.trans.
rewrite <- H22.
red in H8; decompose record H8; clear H8.
rewrite H25.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red; intros.
apply H15.
apply Pg_Off_eq; auto.
rewrite <- Off_simpl with ppg l.
rewrite <- Off_simpl with ppg l'.
rewrite H8; auto.
Qed.


Lemma tttt : forall vpg ppg bs s bs' s',
ValidVPg vpg = true ->
ValidPg vpg = false ->
ValidPg ppg = true ->
ppg <> 0%Z ->
AS_State.allocdata (bs.(C_AS.State.memory)) vpg = false ->
PMAP_State.allocdata (s.(C_PMAP.State.memory)) ppg = false ->
(forall l, (Pg l = vpg) -> bs'.(C_AS.State.memory).(AS_State.data) l = 
                                s'.(C_PMAP.State.memory).(PMAP_State.data) (ppg * PGSIZE + Off l) ) ->
  C_AS_PMAP.CRepr bs s ->
  AS_Lib.allocpg vpg bs bs' ->
  as_sig.vallocpg vpg ppg s s' ->
  C_AS_PMAP.CRepr bs' s'.
intros.
red in H6; decompose record H6; clear H6.
red in H7; decompose record H7; clear H7.
red in H8; decompose record H8; clear H8.
red; split.
rewrite H11.
rewrite H12.
rewrite H9; autosolve.
apply vvvv with vpg ppg (bs.(C_AS.State.memory)) (s.(C_PMAP.State.memory)); autosolve.
Qed.


Lemma reserve_weak1 : ActionGTE (C_AS_PMAP.RF.reprAction C_AS_PMAP.CRepr AS_Lib.as_reserve_spec)
                                  as_sig.as_reserve_spec.
split.

intros.
unfold AS_Lib.as_reserve_spec, as_sig.as_reserve_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_AS.State.StackCheck, C_PMAP.State.StackCheck in *; simpl in *.
unfold C_AS.State.StackGet, C_PMAP.State.StackGet in *; simpl in *.
rewrite H in *.
repeat split; autosolve.

(* rel *)
intros.
unfold AS_Lib.as_reserve_spec, as_sig.as_reserve_spec in *; simpl in *; repeat autoclean.
inversion_clear H0; repeat autoclean.

(* positive case *)
destruct (exists_alloc (nth 0 (C_AS.State.StackGet C_AS.State.Stack.getArgs bs) 0%Z)
                    (fun l => s'.(C_PMAP.State.memory).(PMAP_State.data) (x0 * PGSIZE + Off l)) 
                  bs) as [bs'].
(* exists_alloc: need to show that x0 is currently unallocated *)
destruct H5.
red in H13; decompose record H13; clear H13.
makenew' (AS_State.allocdata (C_AS.State.memory bs) 
            (nth 0 (C_AS.State.StackGet C_AS.State.Stack.getArgs bs) 0%Z) ); autosolve.
(*****)
exists bs'; repeat autoclean.
split.
red in H; repeat autoclean.
unfold C_AS.State.StackCheck, C_PMAP.State.StackCheck in *; simpl in *.
unfold C_AS.State.StackGet, C_PMAP.State.StackGet in *; simpl in *.
rewrite H in *.
apply tttt with 
     (nth 0 (C_AS.State.Stack.getArgs (C_AS.State.stack bs)) 0%Z)
     x0 bs s; autosolve.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red in H5; repeat autoclean.
rewrite H5; autosolve.
left; autosolve.

(* negative case *)
eexists; split; [|right; autosolve].
split; simpl.
red in H5; repeat autoclean.
rewrite H0; autosolve.
red in H5; repeat autoclean; autosolve.
Qed.


Lemma ReprMem_vfree_preserve : forall pg bs bs' s s',
ValidVPg pg = true ->
ValidPg pg = false ->
AS_State.allocdata bs pg = true ->
AS_PMAP.Repr_Mem bs s -> 
as_sig.PMAP_vfree pg s s' ->
AS_Lib.AS_free pg bs bs' ->
AS_PMAP.Repr_Mem bs' s'.
intros.
red in H2; repeat autoclean.
red; repeat split.

intros.
red in H4; decompose record H4; clear H4.
red in H3; decompose record H3; clear H3.
set (H11_copy := H11); generalize H11_copy; clear H11_copy; intros H11_copy.
rewrite H13 in H11.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H4.
rewrite H12; autosolve.
apply H2; autosolve.
makenew' (PMAP_State.pagemap s pg); autosolve.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(* case: we are comparing pages, one of which is the page we freed *)
(* this can not happen - the phys page could not have been allocated *)
elimtype False.
apply H9 with l (pg * PGSIZE + Off l)%Z; autosolve.
red; intros.
rewrite H3 in H14; rewrite Pg_simpl in *; autosolve.
apply ValidPA_ValidVA; auto.
apply ValidVAComp; auto.
apply ValidPA_ValidVA; auto.
rewrite Pg_simpl in *; auto.
unfold PMAP_State.trans', isPhys.
rewrite Pg_simpl.
rewrite H0; simpl.
replace (ValidPg (Pg l)) with true; simpl.
unfold PMAP_State.trans.
rewrite Pg_simpl.
rewrite H17.
rewrite Off_simpl.
rewrite <- H18.
apply Addr_parts.
symmetry; auto.
apply ValidPA_ValidPg; auto.
apply H6; autosolve.
apply ValidPA_ValidPg; auto.

intros.
red in H4; decompose record H4; clear H4.
red in H3; decompose record H3; clear H3.
set (H12_copy := H12); generalize H12_copy; clear H12_copy; intros H12_copy.
rewrite H14 in H12.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H13; autosolve.
rewrite H4; autosolve.
rewrite H5; autosolve.
f_equal.
unfold PMAP_State.trans.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold PMAP_State.trans.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (PMAP_State.pagemap s pg); autosolve.
assert (ValidVPg (Pg l) = true); try apply ValidVA_ValidVPg; autosolve.
assert (ValidPg (Pg l) = false).
apply AS_PMAP.ValidPg_virt; autosolve.
specialize (H7 (Pg l) H3 H20 H12).
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
rewrite Pg_simpl.
rewrite H17.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
(***)
rewrite H23 in *; clear H23 z0.
elimtype False.
apply H9 with l (pg * PGSIZE + Off l)%Z; autosolve.
red; intros.
rewrite H21 in *; rewrite Pg_simpl in *; autosolve.
apply ValidVAComp; autosolve.
rewrite Pg_simpl; auto.
unfold PMAP_State.trans', isPhys.
rewrite H20.
rewrite Pg_simpl.
rewrite H0.
unfold PMAP_State.trans.
rewrite Pg_simpl.
rewrite H19; rewrite H22.
rewrite Off_simpl; auto.

intros.
red in H4; decompose record H4; clear H4.
red in H3; decompose record H3; clear H3.
rewrite H13 in H11; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (PMAP_State.pagemap s pg); autosolve.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H18 in *; clear H18 pg0.
elimtype False.
apply H9 with (z * PGSIZE + Off 0)%Z (pg * PGSIZE + Off 0)%Z; try rewrite Pg_simpl in *; autosolve.
apply ValidVAComp; auto.
apply ValidPg_ValidVPg; auto.
apply ValidVAComp; auto.
unfold PMAP_State.trans', isPhys.
repeat rewrite Pg_simpl in *.
rewrite H10.
rewrite H0.
unfold PMAP_State.trans.
rewrite Pg_simpl.
rewrite Off_simpl.
rewrite H17; auto.
(***)
apply H6; autosolve.

intros.
red in H4; decompose record H4; clear H4.
red in H3; decompose record H3; clear H3.
rewrite H14 in H12.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
specialize (H7 pg0 H10 H11 H12).
makenew' (PMAP_State.pagemap s pg0); autosolve.
makenew' (PMAP_State.pagemap s pg); autosolve.
rewrite H17.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite <- H20 in *; clear z0 H20.
elimtype False.
apply H9 with (pg0 * PGSIZE + Off 0)%Z (pg * PGSIZE + Off 0)%Z; try rewrite Pg_simpl in *; autosolve.
apply ValidVAComp; autosolve.
apply ValidVAComp; autosolve.
unfold PMAP_State.trans', isPhys.
repeat rewrite Pg_simpl in *.
rewrite H11.
rewrite H0.
unfold PMAP_State.trans.
repeat rewrite Pg_simpl in *.
repeat rewrite Off_simpl in *.
rewrite H18; rewrite H19; auto.

intros.
red in H4; decompose record H4; clear H4.
red in H3; decompose record H3; clear H3.
rewrite H14 in H12.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H8; auto.

intros.
red in H4; decompose record H4; clear H4.
red in H3; decompose record H3; clear H3.
rewrite H16 in H13.
rewrite H16 in H14.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red; intros.
apply H9 with l l'; autosolve.
unfold PMAP_State.trans', isPhys in *.
makenew' (ValidPg (Pg l)); makenew' (ValidPg (Pg l')); autosolve.
unfold PMAP_State.trans in *.
rewrite H18 in H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold PMAP_State.trans in *.
rewrite H18 in H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
unfold PMAP_State.trans in *.
rewrite H18 in H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma vfree_repr_preserve : forall pg bs bs' s s',
ValidVPg pg = true ->
ValidPg pg = false ->
AS_State.allocdata (C_AS.State.memory bs) pg = true ->
C_AS_PMAP.CRepr bs s -> 
as_sig.vfree pg s s' ->
AS_Lib.freepg pg bs bs' ->
C_AS_PMAP.CRepr bs' s'.
intros.
red in H2; repeat autoclean.
red in H3; repeat autoclean.
red in H4; repeat autoclean.
red; split.
rewrite H7.
rewrite H6.
rewrite H2; autosolve.
eapply ReprMem_vfree_preserve; eauto.
Qed.


Lemma release_weak1 : ActionGTE (C_AS_PMAP.RF.reprAction C_AS_PMAP.CRepr AS_Lib.as_release_spec)
                                  as_sig.as_release_spec.
split.

intros.
unfold AS_Lib.as_release_spec, as_sig.as_release_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_AS.State.StackCheck, C_PMAP.State.StackCheck in *; simpl in *.
unfold C_AS.State.StackGet, C_PMAP.State.StackGet in *; simpl in *.
rewrite H in *.
repeat split; autosolve.
red in H4; repeat autoclean.
makenew ((nth 0 (C_PMAP.State.Stack.getArgs (C_PMAP.State.stack s)) 0%Z)).
specialize (H7 H10 H1 H2 H3).
makenew' (PMAP_State.pagemap (C_PMAP.State.memory s) H10); autosolve.
repeat autoclean.

(* rel *)
intros.
unfold AS_Lib.mem_free_spec, PMAP_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
destruct (exists_free (nth 0 (C_AS.State.StackGet C_AS.State.Stack.getArgs bs) 0%Z) bs) as [bs'].
exists bs'; split; auto.

apply vfree_repr_preserve with (nth 0 (C_AS.State.StackGet C_AS.State.Stack.getArgs bs) 0%Z) bs s; autosolve.
unfold C_AS.State.StackGet, C_PMAP.State.StackGet in *; simpl in *.
red in H5; repeat autoclean.
rewrite H5; autosolve.
Qed.

(********************************************************)
Require Import pmap_sig.

Lemma PMAP_AS_mem_exists : forall s, 
(forall pg, ValidVPg pg = true -> ValidPg pg = false -> PMAP_State.pagemap s pg = None) ->
exists bs, AS_PMAP.Repr_Mem bs s.
intros.
exists {| 
   AS_State.data := (fun l => 0%Z); 
   AS_State.allocdata := fun pg => false
|}; red; simpl.
repeat split; autosolve.
intros; apply H; auto.
Qed.

Lemma kernel_init_weak : ActionGTE as_sig.PMAP_kernel_init_spec
            (C_AS_PMAP.RF.reprAction C_AS_PMAP.CRepr as_sig.AS_kernel_init_spec).
split.

intros.
unfold as_sig.PMAP_kernel_init_spec, as_sig.AS_kernel_init_spec in *; simpl in *; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_AS.State.StackCheck in *.
destruct s; simpl in *.
destruct stack; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean.
destruct (PMAP_AS_mem_exists memory) as [bmem]; auto.
exists {| C_AS.State.stack := C_Stack.CallFrame nil :: stack; 
              C_AS.State.memory := bmem |}; simpl.
split; auto.
red; simpl.
split; auto.

intros.
unfold as_sig.PMAP_kernel_init_spec, as_sig.AS_kernel_init_spec in *; simpl in *; repeat autoclean.
destruct s; simpl in *.
destruct (PMAP_AS_mem_exists memory) as [bmem]; auto.
destruct H0 with {| C_AS.State.stack := stack; 
              C_AS.State.memory := bmem |}; simpl in *; autosolve.
red; simpl.
split; auto.
Qed.


(*****************************************************************************)
(* This is the assumption of a certified kernel *)

Require Import label.

    Definition ref := C_AS_PMAP.EqRef.

Lemma WF_kernel_pmap : forall kernel_code kernel_spec as_lib_code,
         C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint
                 (C_PMAP.SCAP.SpecHeap.SpecHeap_join PMAP_Lib.pmap_lib_spec as_sig.as_spec as_sig.lib_as_disj)
                    (C_AS_PMAP.PsiMake kernel_spec) ->
         C_AS.SCAP.WFCode C_AS.State.expr_to_Prop C_AS.Mc_Machine AS_Lib.as_lib_spec kernel_code kernel_spec ->
         
         C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine PMAP_Lib.pmap_lib_spec as_lib_code as_sig.as_spec ->
         C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine 
                 (C_PMAP.SCAP.SpecHeap.SpecHeap_join PMAP_Lib.pmap_lib_spec as_sig.as_spec as_sig.lib_as_disj)
                                        (C_AS_PMAP.CMake kernel_code) 
                                        (C_AS_PMAP.PsiMake kernel_spec).
intros.
set (x := ref AS_Lib.as_lib_spec kernel_code kernel_spec).
eapply C_PMAP.SCAP.WFCode_stronghypo; eauto.
red; intros.
destruct l; autosolve; simpl.
apply reserve_weak1.
apply release_weak1.
apply alloc_weak1.
apply free_weak1.
Qed.

Lemma WF_kernel_pmap' : forall kernel_code kernel_spec as_lib_code
         (disj : C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint (C_AS_PMAP.PsiMake kernel_spec) as_sig.as_spec)
         (disj' : C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint (C_AS_PMAP.PsiMake kernel_spec) PMAP_Lib.pmap_lib_spec),
         C_PMAP.SCAP.CH.CodeHeap_disjoint as_lib_code (C_AS_PMAP.CMake kernel_code) ->
         C_AS.SCAP.WFCode C_AS.State.expr_to_Prop C_AS.Mc_Machine AS_Lib.as_lib_spec kernel_code kernel_spec ->
         C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine PMAP_Lib.pmap_lib_spec as_lib_code as_sig.as_spec ->

      C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine 
                PMAP_Lib.pmap_lib_spec 
                (C_PMAP.SCAP.CH.CodeHeap_join (C_AS_PMAP.CMake kernel_code) as_lib_code) 
                (C_PMAP.SCAP.SpecHeap.SpecHeap_join (C_AS_PMAP.PsiMake kernel_spec) as_sig.as_spec disj).

intros.
apply C_PMAP.SCAP.WFCode_linking_smp with as_sig.lib_as_disj; auto.
eapply WF_kernel_pmap; eauto.
apply C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint_joinleft; auto.
apply C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
apply C_PMAP.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
Qed.

(****************************************************************)
