(* mem_repr.v *)
(* Defines the REPR-refinement between C machines based on the relation between the states of the models *)

Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.
Require Import c.
Require Import reprfinement.


(* The template for relation between two memory models *)
Module Type Mem_Repr_T (Mem1 Mem2 : Mem_T).

Parameter Repr_Mem : Mem1.t ->Mem2.t -> Prop.
Parameter Repr_getMemCheck : forall (m1 : Mem1.t) (m2 : Mem2.t), 
      Repr_Mem m1 m2 ->  
      forall l, Mem1.getMemCheck l m1 = true -> Mem2.getMemCheck l m2 = true.
Parameter Repr_getMem : forall (m1 : Mem1.t) (m2 : Mem2.t), 
      Repr_Mem m1 m2 ->  
      forall l v, Mem1.getMemCheck l m1 = true -> Mem1.getMem l m1 = v -> Mem2.getMem l m2 = v.
Parameter Repr_setMemCheck : forall (m1 : Mem1.t) (m2 : Mem2.t), 
      Repr_Mem m1 m2 ->  
      forall l, Mem1.setMemCheck l m1 = true -> Mem2.setMemCheck l m2 = true.
Parameter Repr_setMem : forall (m1 : Mem1.t) (m2 : Mem2.t), 
      Repr_Mem m1 m2 ->
      forall l v , Mem1.setMemCheck l m1 = true -> Repr_Mem (Mem1.setMem l v m1) (Mem2.setMem l v m2).
End Mem_Repr_T.


(***********************************************************)
(* Definition of the C Refinement (using memory model repr as a parameter *)

Module Type C_Refine_T (Mem1 Mem2 : Mem_T)  (MemRepr : Mem_Repr_T Mem1 Mem2) 
 (C1 : C_T Mem1) (C2 : C_T Mem2).
End C_Refine_T.


Module C_Refine (Mem1 Mem2 : Mem_T) (MemRepr : Mem_Repr_T Mem1 Mem2)
                            (C1 : C_T Mem1) (C2 : C_T Mem2)
                           <: C_Refine_T Mem1 Mem2 MemRepr  C1 C2.

Import reprfinement.
Export reprfinement.
Import MemRepr.

(* The REPR between two machines *)
Definition CRepr : C1.State.t -> C2.State.t -> Prop :=
  fun astate cstate => 
         (C1.State.stack astate = C2.State.stack cstate) /\
         Repr_Mem (C1.State.memory astate) (C2.State.memory cstate).

(* The Refinement generated between two machines *)
Module RF := Refinement C1.State C1.Ops Expr C1.SCAP C2.State C2.Ops Expr C2.SCAP.
Import RF.

Definition CMake := RF.Eq_codemake (fun x => x) (fun x => x).
Definition PsiMake := RF.Eq_specmake CRepr.

Require Import sact.
Require Import actsmp.
Import C1.State.

Lemma CRepr_StackCheck : 
forall e (x : C1.State.t) (s : C2.State.t), 
CRepr x s -> StackCheck e x = true -> C2.State.StackCheck e s = true.
intros.
unfold StackCheck, C2.State.StackCheck in *.
red in H; repeat autoclean.
rewrite H in H0; autosolve.
Qed.

Lemma CRepr_StackCheck' : 
forall e (x : C1.State.t) (s : C2.State.t), 
CRepr x s -> C2.State.StackCheck e s = true -> StackCheck e x = true.
intros.
unfold StackCheck, C2.State.StackCheck in *.
red in H; repeat autoclean.
rewrite H; autosolve.
Qed.


Lemma CRepr_StackGet : 
forall (A : Type) e (x : C1.State.t) (s : C2.State.t), 
CRepr x s -> StackGet (A := A) e x = C2.State.StackGet e s.
intros.
unfold StackGet, C2.State.StackGet in *.
red in H; repeat autoclean.
rewrite H; autosolve.
Qed.

Lemma CRepr_eval : forall e AS CS, CRepr AS CS -> evalCheck e AS = true ->
  eval e AS = C2.State.eval e CS.
induction e; intros; unfold CRepr in *; simpl in *; repeat autoclean; autosolve.
unfold StackGet, C2.State.StackGet in *.
f_equal; autosolve.

unfold MemGet, C2.State.MemGet.

sact_auf; repeat autoclean.
unfold MemCheck in H0.
rewrite Repr_getMem with (memory AS) (C2.State.memory CS) 
            (C2.State.eval e CS) (getMem (eval e AS) (memory AS)); autosolve.
rewrite <- IHe with AS _; try split; autosolve.
rewrite <- IHe with AS _; try split; autosolve.

sact_auf; repeat autoclean.
f_equal.
apply IHe1; try split; autosolve.
apply IHe2; try split; autosolve.
f_equal.
apply IHe; try split; autosolve.
Qed.


Lemma CRepr_evalCheck : 
forall e (x : C1.State.t) (s : C2.State.t), 
CRepr x s -> evalCheck e x = true -> C2.State.evalCheck e s = true.
induction e; autosolve.
intros.
unfold evalCheck, C2.State.evalCheck in *.
apply CRepr_StackCheck with x; autosolve.

intros.
sact_auf.
unfold CRepr in *; simpl in *; repeat autoclean.

unfold MemCheck, C2.State.MemCheck in *; simpl.
rewrite Repr_getMemCheck with (memory x) (C2.State.memory s) _; simpl; autosolve.
rewrite IHe with x s; try split; autosolve.
rewrite <- CRepr_eval with e x _; autosolve.
red; split; autosolve.

intros; sact_auf; simpl in *.
repeat autoclean.
rewrite IHe1 with x s; simpl; autosolve.
rewrite IHe2 with x s; simpl; autosolve.
Qed.

Lemma CRepr_evallistCheck : 
forall el (x : C1.State.t) (s : C2.State.t), 
CRepr x s -> evallistCheck el x = true -> C2.State.evallistCheck el s = true.
induction el; simpl in *; intros; autosolve.
sact_auf; repeat autoclean.
rewrite CRepr_evalCheck with a x s; simpl; autosolve.
apply IHel with x; autosolve.
Qed.


Lemma CRepr_evallist : forall e AS CS, CRepr AS CS -> evallistCheck e AS = true ->
  evallist e AS = C2.State.evallist e CS.
induction e; autosolve.
intros; simpl in *.
sact_auf; repeat autoclean.
rewrite CRepr_eval with a AS CS; simpl; autosolve.
f_equal.
apply IHe; autosolve.
Qed.

Lemma CRepr_evalCheck' : 
forall (e : Expr.expr_nomem) (x : C1.State.t) (s : C2.State.t),
CRepr x s -> C2.State.evalCheck (Expr.expr_nomem_embed e) s = true -> 
                   evalCheck (Expr.expr_nomem_embed e) x = true.
induction e; simpl in *; autosolve.
intros.
unfold evalCheck, C2.State.evalCheck in *.
apply CRepr_StackCheck' with s; autosolve.

intros; sact_auf; simpl in *.
repeat autoclean.
rewrite IHe1 with x s; simpl; autosolve.
rewrite IHe2 with x s; simpl; autosolve.
Qed.

Lemma CRepr_StackAct :
forall AS CS act act', act = act' -> CRepr AS CS ->
CRepr (StackAct act AS) (C2.State.StackAct act' CS).
intros; subst.
unfold CRepr in *.
repeat autoclean.
split; autosolve.
unfold StackAct, C2.State.StackAct; simpl.
f_equal; auto.
Qed.

(* Proof of existance of the REPR refinement between the C machines *)

Lemma refine : ReprRefinement CRepr (fun x => x) (fun x => x) C1.Mc_Machine C2.Mc_Machine.
unfold ReprRefinement.

intros.
case_eq (C2.Mc_Machine Ai); intros.
exists a; split; auto.

unfold C1.Mc_Machine, C2.Mc_Machine in *.
inversion_clear H0.
inversion_clear H1.

destruct Ai; simpl in *.

unfold reprAction.
sact_structured.
sact_auf.
split; intros; simpl in *.
repeat autoclean.
rewrite CRepr_evalCheck with e x s; autosolve; simpl.
rewrite CRepr_StackCheck with (setVarCheck v) x s; autosolve.
intros.
repeat autoclean.
eexists; split; autosolve.
apply CRepr_StackAct; autosolve.
f_equal.
apply CRepr_eval; autosolve.


unfold reprAction.
sact_structured.
sact_auf.
split; intros; simpl in *.
repeat autoclean.
apply CRepr_evallistCheck with x; autosolve.
intros.
eexists; split; autosolve.
repeat autoclean; subst.
apply CRepr_StackAct; autosolve.
f_equal.
apply CRepr_evallist; autosolve.

unfold reprAction.
sact_structured.
sact_auf.
split; intros; simpl in *.
repeat autoclean.
rewrite CRepr_StackCheck with _ x _; simpl; autosolve.
rewrite CRepr_evalCheck with e x s; simpl; autosolve.
intros.
eexists; split; autosolve.
repeat autoclean; subst.
apply CRepr_StackAct; autosolve.
apply functional_extensionality; intros.
f_equal.
apply CRepr_eval; autosolve.

unfold reprAction.
sact_auf.
split; intros; simpl in *.
repeat autoclean.
apply CRepr_StackCheck with x; autosolve.
intros.
eexists; split; autosolve.
repeat autoclean; subst.
apply CRepr_StackAct; autosolve.

unfold reprAction.
sact_structured.
sact_auf.
split; intros; simpl in *.
repeat autoclean.
apply CRepr_StackCheck with x; autosolve.
intros.
eexists; split; autosolve.
repeat autoclean; subst.
apply CRepr_StackAct; autosolve.
apply functional_extensionality; intros.
f_equal.
apply CRepr_StackGet; autosolve.

unfold reprAction.
sact_structured.
sact_auf.
split; intros; simpl in *.
repeat autoclean.
rewrite CRepr_StackCheck with (isDataFrame) x s; simpl; autosolve.
rewrite <- CRepr_eval with e x s; autosolve.
rewrite CRepr_evalCheck with e x s; simpl; autosolve.
rewrite CRepr_evalCheck with e0 x s; simpl; autosolve.
unfold MemCheck, C2.State.MemCheck in *; simpl in *.
rewrite Repr_setMemCheck with (memory x) _ _; simpl; autosolve.
unfold CRepr in H0; repeat autoclean; autosolve.

intros.
repeat autoclean.
eexists; split; autosolve.
(* complex stuff here *)
rewrite <- (CRepr_eval e bs s); autosolve.
rewrite <- (CRepr_eval e0 bs s); autosolve.
clear H0; unfold CRepr in *; simpl; repeat autoclean.
split; autosolve.
apply Repr_setMem; autosolve.

(* failing case *)
destruct Ai; unfold C2.Mc_Machine in H1; inversion H1.
Qed.



(*******************************)
(* Proof that any C program will not break branching condition *)

Lemma Rrel_eval : 
forall sa sa' expr1,
Rrel CRepr sa sa' -> 
eval (Expr.expr_nomem_embed expr1) sa = eval (Expr.expr_nomem_embed expr1) sa'.
induction expr1; repeat autoclean; simpl in *; repeat autoclean; autosolve.

destruct H; repeat autoclean.
destruct H; destruct H0; repeat autoclean.
unfold StackGet.
rewrite H; rewrite H0; auto.

rewrite IHexpr1_1; autosolve.
rewrite IHexpr1_2; autosolve.

rewrite IHexpr1; autosolve.
Qed.

Lemma Rrel_evalCheck: 
forall sa sa' expr1,
Rrel CRepr sa sa' -> 
evalCheck (Expr.expr_nomem_embed expr1) sa = true -> 
evalCheck (Expr.expr_nomem_embed expr1) sa' = true.
induction expr1; intros; simpl in *; repeat autoclean; autosolve.

destruct H; repeat autoclean.
destruct H; destruct H1; repeat autoclean.
unfold StackCheck in *.
rewrite H in *; rewrite H1; auto.

unfold SCheck_and in *; repeat autoclean.
rewrite IHexpr1_1; autosolve.
rewrite IHexpr1_2; autosolve.

rewrite IHexpr1; autosolve.
Qed.


Lemma C_excl_expr : forall expr, excl CRepr C1.State.expr_to_Prop expr.
induction expr0.
(* expr_z *)
unfold expr_to_Prop; simpl; red; intros.
red; intros.
apply H1; clear H1; autosolve.

(* expr_var *)
unfold expr_to_Prop; simpl; red; intros.
red; intros.
apply H1; clear H1; repeat autoclean; autosolve.
red in H; repeat autoclean.
red in H; repeat autoclean.
red in H2; repeat autoclean.
simpl in *.
unfold evalCheck, eval, StackCheck, StackGet in *.
rewrite H in *; rewrite H2 in *; autosolve.
split; autosolve.

(* expr_binop *)
unfold expr_to_Prop; simpl; red; intros.
red; intros.
apply H1; clear H1; repeat autoclean; autosolve.
simpl in *.
unfold SCheck_and in *.
repeat autoclean.
rewrite Rrel_evalCheck with sa _ expr0_1; autosolve.
rewrite Rrel_evalCheck with sa _ expr0_2; autosolve.
split; autosolve.
rewrite <- Rrel_eval with sa _ expr0_1; autosolve.
rewrite <- Rrel_eval with sa _ expr0_2; autosolve.

(* expr_unop *)
unfold expr_to_Prop; simpl; red; intros.
red; intros.
apply H1; clear H1; repeat autoclean; autosolve.
simpl in *.
unfold SCheck_and in *; repeat autoclean.
rewrite Rrel_evalCheck with sa _ expr0; autosolve.
split; autosolve.
rewrite <- Rrel_eval with sa _ expr0; autosolve.
Qed.


Lemma C_excl_proc : forall I, RF.excl_Proc CRepr (C1.State.expr_to_Prop) I.
induction I; intros; constructor.
apply IHI1; auto.
apply IHI2; auto.
apply IHI1; autosolve.
apply IHI2; autosolve.
apply C_excl_expr; auto.
Qed.

Lemma C_excl : forall C, RF.excl_Code CRepr (C1.State.expr_to_Prop) C.
intros.
red; intros.
apply C_excl_proc; auto.
Qed.

(***********************************)

Lemma d1 : 
forall (d : C1.SCAP.Process.Decider) (sa : AState) (sc : CState),
CRepr sa sc -> expr_to_Prop d sa -> C2.State.expr_to_Prop d sc.
intros; simpl in *.
unfold expr_to_Prop, C2.State.expr_to_Prop in *; repeat autoclean.
rewrite <- CRepr_eval with _ sa _; autosolve.
rewrite CRepr_evalCheck with _ sa sc; autosolve.
split; auto.
Qed.

Lemma d2 :
forall (d : C1.SCAP.Process.Decider) (sa : AState) (sc : CState),
CRepr sa sc -> C2.State.expr_to_Prop d sc -> expr_to_Prop d sa.
intros; simpl in *.
unfold expr_to_Prop, C2.State.expr_to_Prop in *; repeat autoclean.
rewrite <- CRepr_eval with _ sa sc in H1; autosolve.
split; auto.
apply CRepr_evalCheck' with sc; autosolve.
apply CRepr_evalCheck' with sc; autosolve.
Qed.

(* The useful refinement rule that allows us to refine a particular module *)
Definition EqRef : 
   forall AHypo AC APsi,
   C1.SCAP.WFCode (C1.State.expr_to_Prop) C1.Mc_Machine AHypo AC APsi -> 
   C2.SCAP.WFCode (C2.State.expr_to_Prop) C2.Mc_Machine
       (Eq_specmake CRepr AHypo) 
       (Eq_codemake (fun x => x) (fun x => x) AC)
       (Eq_specmake CRepr APsi).
intros.
apply EqAutoReprRefinement with C1.Mc_Machine (C1.State.expr_to_Prop); autosolve.
exact d1.
exact d2.
apply refine.
apply C_excl.
Qed.


End C_Refine.
