(* pe_hw.v *)
(* Definition of memory relation between the PE and HW memory models *)
(* Also includes compatibility proof that the stub of kernel_init in the PE model
    is compatible with the stub of kernel_init in the HW model *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.
Require Import mem.

Require Import mem_repr.
Require Import pe.
Require Import hw.

Opaque PTROOT PE_State.PMM_data_loc.

Module PE_HW <: Mem_Repr_T PE_State HW_State.

Definition Repr_Mem (pe : PE_State.t) (hw : HW_State.t) :=
   (forall l, (PE_State.data pe) l = HW_State.data hw l) /\ (HW_State.ptroot hw = PE_State.PTROOT) /\
   (HW_State.pe hw = true) /\ ValidPA (HW_State.ptroot hw) = true /\ (HW_State.tlb hw = PE_State.tlb pe).

Lemma Repr_getMemCheck : forall (m1 : PE_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PE_State.getMemCheck l m1 = true -> HW_State.getMemCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold PE_State.getMemCheck, HW_State.getMemCheck in *.
unfold PE_State.ValidVA, HW_State.ValidVA in *; simpl; repeat autoclean.
rewrite H2.
rewrite H3; simpl.
rewrite H1; simpl.
rewrite <- H; autosolve.
Qed.

Lemma Repr_getMem : forall (m1 : PE_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v, PE_State.getMemCheck l m1 = true -> PE_State.getMem l m1 = v -> HW_State.getMem l m2 = v.
intros.
red in H; repeat autoclean.
unfold PE_State.getMem, HW_State.getMem in *.
rewrite H3; simpl.
rewrite H2; simpl; autosolve.
repeat rewrite <- H; autosolve.
Qed.

Lemma Repr_setMemCheck : forall (m1 : PE_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PE_State.setMemCheck l m1 = true -> HW_State.setMemCheck l m2 = true.
intros.
red in H; repeat autoclean.
unfold PE_State.setMemCheck, HW_State.setMemCheck in *.
unfold PE_State.ValidVA, HW_State.ValidVA in *; repeat autoclean; simpl.
rewrite H2; simpl.
rewrite H3; simpl.
rewrite H1.
repeat rewrite <- H; autosolve.
Qed.

Lemma Repr_setMem : forall (m1 : PE_State.t) (m2 : HW_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v , PE_State.setMemCheck l m1 = true -> Repr_Mem (PE_State.setMem l v m1) (HW_State.setMem l v m2).
intros.
red in H; repeat autoclean.
unfold PE_State.setMem, HW_State.setMem, Repr_Mem in *; simpl.
repeat rewrite H2; simpl.
repeat split; autosolve.
rewrite <- H1.
rewrite H.
intros; Zeq_pull'; Zeq_norm; simpl; autosolve.
Qed.

End PE_HW.

Require Import c.

Module C_PE_HW := C_Refine PE_State HW_State PE_HW C_PE C_HW.

Import PE_HW.
Require Import pe_lib.
Require Import hw_lib.

Opaque PTROOT PE_State.PTROOT.

Lemma tlb_reset_weak1 : ActionGTE (C_PE_HW.RF.reprAction C_PE_HW.CRepr PE_Lib.tlb_reset_spec)
  HW_Lib.pe_enable_spec.
split.

intros.
unfold PE_Lib.tlb_reset_spec, HW_Lib.pe_enable_spec in *; simpl in *; repeat autoclean.
unfold C_PE.State.StackCheck, C_HW.State.StackCheck in *.
red in H; repeat autoclean.
rewrite H in *; autosolve.

intros.
unfold PE_Lib.tlb_reset_spec, HW_Lib.pe_enable_spec in *; simpl in *; repeat autoclean.
eexists; split; eauto.
clear H H1.
red in H2; repeat autoclean.
unfold C_PE.State.StackCheck, C_HW.State.StackCheck in *.
unfold C_PE.State.StackAct, C_HW.State.StackAct in *.
unfold C_PE.State.MemAct, C_HW.State.MemAct in *.
split; simpl.
rewrite H; autosolve.
red in H0; repeat autoclean.
red; repeat split; autosolve.
Qed.

(********************************************************)

Require Import ale.

Lemma PE_HW_mem_exists : forall (s : HW_State.t), 
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                      s.(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) ->
     (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> 
                      s.(HW_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) ->
     (forall pg, ValidPg pg = true -> s.(HW_State.data) (PTROOT + pg * 8) = pg) ->
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s.(HW_State.data) (PTROOT + pg * 8)= 0%Z) ->
      s.(HW_State.pe) = true ->
      s.(HW_State.ptroot) = PTROOT ->
      s.(HW_State.tlb) = true ->

(exists bs, PE_HW.Repr_Mem bs s /\ 
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                      bs.(PE_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) /\
     (forall l, ValidPA l = true -> PE_State.PMMDom l = true -> 
                      bs.(PE_State.data) (PE_State.PMM_data_loc + Pg l * 8) = 1%Z) /\
     (forall pg, ValidPg pg = true -> bs.(PE_State.data) (ALE_State.PTROOT + pg * 8) = pg) /\
     (forall pg, ValidVPg pg = true -> ValidPg pg = false ->
                            bs.(PE_State.data) (ALE_State.PTROOT + pg * 8)= 0%Z) /\
      bs.(PE_State.tlb) = true
).

intros.
exists {| 
   PE_State.data := (fun l => s.(HW_State.data) l);
   PE_State.tlb := true
|}; repeat split; simpl; autosolve.
rewrite H4.
replace (PTROOT) with (PTROOT + 0 * 8)%Z; try omega.
apply ALE_State.PtDom_Valid.
apply ValidPg_ValidVPg.
unfold ValidPg.
unfold NPAGES.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.


Require Import mem_sig.
Require Import hw_lib.

Lemma kernel_init_weak : ActionGTE HW_Lib.HW_kernel_init_spec
            (C_PE_HW.RF.reprAction C_PE_HW.CRepr mem_sig.PE_kernel_init_spec).
split.

intros.
unfold mem_sig.PE_kernel_init_spec, HW_Lib.HW_kernel_init_spec in *; simpl in *; repeat autoclean.
unfold C_PE.State.StackCheck, C_HW.State.StackCheck in *.
destruct s; simpl in *.
destruct stack; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean.
destruct (PE_HW_mem_exists memory) as [bmem]; auto.
repeat autoclean.
exists {| C_PE.State.stack := C_Stack.CallFrame nil :: stack; 
              C_PE.State.memory := bmem |}; simpl.
split; auto.
red; simpl.
split; auto.
split; auto.

intros.
unfold mem_sig.PE_kernel_init_spec, HW_Lib.HW_kernel_init_spec in *; simpl in *; repeat autoclean.
destruct s; simpl in *.
destruct (PE_HW_mem_exists memory) as [bmem]; auto.
repeat autoclean.
destruct H0 with {| C_PE.State.stack := stack; 
              C_PE.State.memory := bmem |}; simpl in *; autosolve.
red; simpl.
split; auto.
split; auto.
Qed.


(******************************************************************)

Require Import label.

Definition ref := C_PE_HW.EqRef.


Lemma WF_kernel_pe' : forall kernel_code kernel_spec
         (disj' : C_HW.SCAP.SpecHeap.SpecHeap_disjoint (C_PE_HW.PsiMake kernel_spec) HW_Lib.hw_lib_spec),
         C_PE.SCAP.WFCode C_PE.State.expr_to_Prop C_PE.Mc_Machine PE_Lib.pe_lib_spec kernel_code kernel_spec ->

      C_HW.SCAP.WFCode C_HW.State.expr_to_Prop C_HW.Mc_Machine 
                HW_Lib.hw_lib_spec 
                (C_PE_HW.CMake kernel_code) 
                (C_PE_HW.PsiMake kernel_spec).
intros.
set (x := ref PE_Lib.pe_lib_spec kernel_code kernel_spec).
eapply C_HW.SCAP.WFCode_stronghypo; eauto.
apply C_HW.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
red; intros.
destruct l; try solve [simpl; auto].
simpl.
apply tlb_reset_weak1.
Qed.


