(* pmap.v *)
(* Definition of the PMAP memory model *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.
Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem.

Module PMAP_State <: Mem_T.

Record State := { data : Z -> Z; pagemap : Z -> option Z; allocdata : Z -> bool}.
Definition t := State.

Definition transCheck (l : Z) : SCheck (State := State) :=
fun s => ValidVA l && 
 	      match s.(pagemap) (Pg l) with
              | Some ppg => ValidPA (ppg * PGSIZE + Off l)
              | None => false
              end.

Definition trans (l : Z) : SGet Z (State := State) :=
fun s => match s.(pagemap) (Pg l) with
              | Some ppg => (ppg * PGSIZE + Off l)%Z
              | None => (-1)%Z
              end.

Definition paCheck (l : Z) : SCheck (State := State) :=
fun s => ValidPA l && s.(allocdata) (Pg l).

Definition vaCheck (l :Z) : SCheck (State := State) := 
fun s => 
      if isPhys l then paCheck l s
      else transCheck l s && paCheck (trans l s) s.

Definition getMemCheck (l:Z) : SCheck (State := State) := 
fun s => vaCheck l s.
 
Definition getMem (l:Z) : SGet (State := State) Z :=
fun s => if isPhys l  then s.(data) l else s.(data) (trans l s).

Definition setMemCheck (l : Z) : SCheck (State := State) := 
fun s => vaCheck l s.

Definition updateMem (l z : Z) : SAct (State := State) :=
fun s => {| data := fun l' => if Zeq_bool l l' then z else s.(data) l'; 
                                      allocdata := s.(allocdata); pagemap := s.(pagemap) |}.
  
Definition setMem (l z : Z) :  SAct (State := State) :=
fun s => if isPhys l then updateMem l z s
              else updateMem (trans l s) z s.

Definition ValidPM : State -> Prop := fun s =>
   forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
	match s.(pagemap) pg with
	| Some ppg => ValidPg ppg = true
	| None => True
	end.

(* getMem simplifications *)
Lemma setMem_unneeded : forall l z s, getMemCheck l s = true -> (getMem l s = z) -> (setMem l z s = s).
intros.
unfold getMem, setMem, updateMem, ValidVA, ValidPA in *; simpl; subst.
destruct s; simpl in *.
unfold getMemCheck, vaCheck, paCheck, transCheck, trans in *; simpl in *.
repeat autoclean.
makenew' (isPhys l); autosolve; repeat autoclean.
f_equal.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.

makenew' (pagemap0 (Pg l)); autosolve.
f_equal; simpl.
apply functional_extensionality; intros.
Zeq_pull'; Zeq_norm; subst; simpl in *; autosolve.
Qed.

Definition trans' (l : Z) : SGet Z (State := State) :=
  fun s => if isPhys l then l else trans l s.

Definition Collision (l l' : Z) : SCheck (State := State) := 
fun s => Zeq_bool (trans' l s) (trans' l' s).

Lemma trans_update_unch : forall l l' z s, trans l (updateMem l' z s) = trans l s.
intros.
unfold updateMem, trans; simpl.
makenew' (pagemap s (Pg l)); autosolve.
Qed.

Lemma trans'_update_unch : forall l l' z s, trans' l (updateMem l' z s) = trans' l s.
intros.
unfold trans'.
makenew' (isPhys l); autosolve.
Qed.

Lemma trans_setMem_unch : forall l l' z s, trans l (setMem l' z s) = trans l s.
intros.
unfold setMem.
makenew' (isPhys l'); autosolve.
Qed.

Lemma transCheck_setMem_unch : forall l l' z s, transCheck l (setMem l' z s) = transCheck l s.
intros.
unfold setMem.
makenew' (isPhys l'); autosolve.
Qed.

Lemma Collision_setMem_unch : forall l0 l' l v m2, 
PMAP_State.Collision l0 l' (PMAP_State.setMem l v m2) = PMAP_State.Collision l0 l' m2.
intros.
unfold Collision; simpl.
unfold trans'; simpl.
unfold setMem; simpl.
makenew' (isPhys l0); makenew' (isPhys l'); try makenew' (isPhys l); simpl in *; auto.
Qed.

Lemma getMemCheck_setMem_unch : 
forall l0 l v m1, PMAP_State.getMemCheck l0 (PMAP_State.setMem l v m1) = PMAP_State.getMemCheck l0 m1.
intros.
unfold getMemCheck, vaCheck.
makenew' (isPhys l0).
unfold paCheck, setMem, updateMem; simpl.
makenew' (isPhys l); simpl; autosolve.

unfold paCheck, setMem, updateMem; simpl.
makenew' (isPhys l); simpl; autosolve.
Qed.

Lemma getMem_setMem : 
  forall l l' z s, 
         getMem l (setMem l' z s) = if Collision l l' s then z else getMem l s.
intros.
unfold Collision, getMemCheck, setMemCheck, getMem, setMem in *; simpl in *.
unfold trans' in *; simpl in *.
makenew' (isPhys l); makenew' (isPhys l'); simpl in *; repeat autoclean; autosolve.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
unfold updateMem, trans; simpl in *.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
unfold updateMem, trans; simpl in *.
Zeq_pull'; Zeq_norm; autosolve; repeat subst; simpl; autosolve.
Qed.

End PMAP_State. 
