(* pmap_ale.v *)
(* Definition of the memory relation between PMAP and ALE memory models *)
(* Also includes the proofs of compatibility for stubs of mem_alloc and mem_free *)
(* Also includes proofs of compatibility for pt_set and pt_lookup between PMAP primitives
    and specs of ALE implementation defined in pmap_sig.v *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import mem_repr.
Require Import pmap.
Require Import ale.

Module PMAP_ALE <: Mem_Repr_T PMAP_State ALE_State. 

Opaque PTROOT ALE_State.PTROOT.

Definition PT_notpresent_in_PMAP (pmap : PMAP_State.t) := 
   (forall l, ALE_State.PtDom l = true -> PMAP_State.allocdata pmap (Pg l) = false).

Definition INVALID_ENTRY : Z := 0%Z.

Definition Repr_Mem (pmap : PMAP_State.t) (ale : ALE_State.t) :=
   (forall l, ValidPA l = true ->
                PMAP_State.allocdata pmap (Pg l) = true -> PMAP_State.data pmap l = ALE_State.data ale l) /\
   (forall l, ValidPA l = true ->
                ALE_State.PtDom l = false -> PMAP_State.allocdata pmap (Pg l) = ALE_State.allocdata ale (Pg l)) /\
   (forall l, ValidPA l = true ->
                ALE_State.PtDom l = true -> PMAP_State.allocdata pmap (Pg l) = false) /\
    (forall l, ValidPA l = true ->
                ALE_State.PtDom l = true -> ALE_State.allocdata ale (Pg l) = true) /\
   (forall vpg, ValidVPg vpg = true ->
                     match PMAP_State.pagemap pmap vpg with
                     | Some ppg => ALE_State.data ale (ALE_State.PTROOT + vpg * 8)%Z = ppg
                     | None => ALE_State.data ale (ALE_State.PTROOT + vpg * 8)%Z = INVALID_ENTRY
                     end) /\
   ALE_State.DM ale /\
   ALE_State.tlb ale = true.

Lemma Repr_vaCheck : forall (m1 : PMAP_State.t) (m2 : ALE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PMAP_State.vaCheck l m1 = true -> ALE_State.vaCheck l m2 = true.
intros.
unfold PMAP_State.vaCheck, ALE_State.vaCheck in *.
unfold PMAP_State.paCheck, ALE_State.paCheck in *.
unfold PMAP_State.transCheck, ALE_State.transCheck in *.
unfold PMAP_State.trans, ALE_State.trans in *.
unfold isPhys in *.
(* phys *)
makenew' (ValidPg (Pg l)); simpl in *; repeat autoclean.
rewrite ValidPA_ValidVA; simpl; autosolve.
rewrite Pg_simpl.
red in H; repeat autoclean.
rewrite H8; simpl.
assert (ALE_State.PtDom l = false).
makenew' (ALE_State.PtDom l); autosolve.
rewrite H4 in H1; autosolve.
red in H7.
specialize (H7 l H0).
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H11.
rewrite <- Pg_simpl with ((ALE_State.data m2 (ALE_State.PTROOT + Pg l * 8))) l.
rewrite H11.
rewrite <- H3; autosolve.
rewrite H1; autosolve.
(* virtual *)
rewrite H0; simpl.
red in H; repeat autoclean.
rewrite H10; simpl.
makenew' (PMAP_State.pagemap m1 (Pg l)); autosolve.
specialize (H8 (Pg l)).
rewrite H12 in H8.
rewrite H8 in *.
unfold negb.
rewrite Pg_simpl in *.
assert (ALE_State.PtDom (z * 4096 + Off l) = false).
makenew' (ALE_State.PtDom (z * 4096 + Off l)); autosolve.
specialize (H6 (z * 4096 + Off l)%Z H1 H13).
rewrite Pg_simpl in *.
rewrite H6 in H3; autosolve.
rewrite H11; simpl.
rewrite <- Pg_simpl with z l.
rewrite <- H5; autosolve.
rewrite Pg_simpl in *; autosolve.
rewrite H3; autosolve.
apply ValidVA_ValidVPg; autosolve.
Qed.

Lemma Repr_getMemCheck : forall (m1 : PMAP_State.t) (m2 : ALE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PMAP_State.getMemCheck l m1 = true -> ALE_State.getMemCheck l m2 = true.
apply Repr_vaCheck.
Qed.

Lemma Repr_getMem : forall (m1 : PMAP_State.t) (m2 : ALE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v, PMAP_State.getMemCheck l m1 = true -> PMAP_State.getMem l m1 = v -> ALE_State.getMem l m2 = v.
intros.
unfold PMAP_State.getMemCheck, ALE_State.getMemCheck in *.
unfold PMAP_State.vaCheck, ALE_State.vaCheck in *.
unfold PMAP_State.paCheck, ALE_State.paCheck in *.
unfold PMAP_State.transCheck, ALE_State.transCheck in *.
unfold PMAP_State.trans, ALE_State.trans in *.
unfold isPhys in *.

red in H; repeat autoclean.
unfold ALE_State.getMem, PMAP_State.getMem.
unfold isPhys in *.
makenew' (ValidPg (Pg l)).
repeat autoclean.
red in H6.
rewrite H6; autosolve.
rewrite H; autosolve.
(* virt *)
repeat autoclean.
makenew' (PMAP_State.pagemap m1 (Pg l)); autosolve.
specialize (H5 (Pg l)).
rewrite H12 in H5.
unfold ALE_State.trans, PMAP_State.trans.
rewrite H5 in *.
rewrite H12.
rewrite H; autosolve.
apply ValidVA_ValidVPg; autosolve.
Qed.


Lemma Repr_setMemCheck : forall (m1 : PMAP_State.t) (m2 : ALE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l, PMAP_State.setMemCheck l m1 = true -> ALE_State.setMemCheck l m2 = true.
intros.
unfold PMAP_State.setMemCheck, ALE_State.setMemCheck in *.
rewrite Repr_vaCheck with m1 m2 l; auto; simpl.
Qed.


Lemma Repr_trans : forall l m1 m2,
ValidVPg (Pg l) = true -> 
Repr_Mem m1 m2 ->
ValidPA (PMAP_State.trans l m1) = true ->
PMAP_State.trans l m1 = ALE_State.trans l m2.
intros.
red in H0; repeat autoclean.
unfold PMAP_State.trans, ALE_State.trans; simpl in *.
specialize (H5 (Pg l) H).
unfold PMAP_State.trans in *.
makenew' (PMAP_State.pagemap m1 (Pg l)); autosolve.
Qed.


Lemma Repr_setMem : forall (m1 : PMAP_State.t) (m2 : ALE_State.t), 
      Repr_Mem m1 m2 ->  
      forall l v , PMAP_State.setMemCheck l m1 = true -> 
         Repr_Mem (PMAP_State.setMem l v m1) (ALE_State.setMem l v m2).
intros.
unfold PMAP_State.setMemCheck in *.
unfold PMAP_State.setMem, ALE_State.setMem in *.
unfold PMAP_State.updateMem in *.
unfold PMAP_State.vaCheck, ALE_State.vaCheck in *.
unfold PMAP_State.paCheck, ALE_State.paCheck in *.
unfold PMAP_State.transCheck, ALE_State.transCheck in *.

unfold isPhys in *.
makenew' (ValidPg (Pg l)); simpl in *; autosolve.
(* phys *)
repeat autoclean.
red in H; repeat autoclean.
red; simpl.
repeat split; intros; autosolve.
rewrite H7; autosolve.
rewrite H; autosolve.
rewrite H3; autosolve.
rewrite H4; autosolve.
rewrite H5; autosolve.
specialize (H6 vpg).
makenew' (PMAP_State.pagemap m1 vpg); autosolve.
rewrite H7; autosolve.
rewrite H6; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H12 in *; simpl.
rewrite H4 in H1; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
(**********)
rewrite H7; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H4 in H1; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
apply H6; autosolve.

(**********)
red; intros; simpl.
rewrite H7; autosolve.
specialize (H7 l0 H9).
unfold ALE_State.trans in *; simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H4 in H1; autosolve.
rewrite H11.
rewrite ALE_State.PtDom_PTROOT; autosolve.
apply ValidVA_ValidVPg; autosolve.
apply ValidPA_ValidVA; autosolve.

(* virt *)
repeat autoclean.
red; simpl.
repeat split; intros; autosolve.
rewrite <- Repr_trans with _ m1 _; autosolve.
red in H; repeat autoclean.
rewrite H; autosolve.
apply ValidVA_ValidVPg; autosolve.

red in H; repeat autoclean.
apply H7; autosolve.

red in H; repeat autoclean.
apply H8; autosolve.

red in H; repeat autoclean.
apply H9; autosolve.

rewrite <- Repr_trans with _ m1 _; autosolve.
red in H; repeat autoclean.
specialize (H9 vpg H5).
makenew' (PMAP_State.pagemap m1 vpg); autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H7 in H3; autosolve.
rewrite H14.
rewrite ALE_State.PtDom_PTROOT; autosolve.
(********)
rewrite H9.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H14 in *.
rewrite H7 in H3; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
(********)
apply ValidVA_ValidVPg; autosolve.

rewrite <- Repr_trans with _ m1 _; autosolve.
red in H; repeat autoclean.
red; intros; simpl in *.
specialize (H9 l0 H11).
unfold ALE_State.trans in *; simpl in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H6 in H3; autosolve.
rewrite H13.
rewrite ALE_State.PtDom_PTROOT; autosolve.
apply ValidVA_ValidVPg; autosolve.
apply ValidPA_ValidVA; autosolve.
apply ValidVA_ValidVPg; autosolve.

red in H; repeat autoclean; autosolve.
Qed.

End PMAP_ALE.

(*******************************************************)
(*******************************************************)
(*******************************************************)
(*******************************************************)
(*******************************************************)

Require Import c.

Module C_PMAP_ALE := C_Refine PMAP_State ALE_State PMAP_ALE C_PMAP C_ALE.

Require Import pmap_lib.
Require Import ale_lib.
Require Import pmap_sig.

Opaque PTROOT ALE_State.PTROOT.

Lemma exists_alloc : forall pg newpagesrc s, 
   s.(C_PMAP.State.memory).(PMAP_State.allocdata) pg = false ->
   exists s', 
   PMAP_Lib.allocpg pg s s' /\  (forall l, Pg l = pg -> (C_PMAP.State.memory s').(PMAP_State.data) l = newpagesrc l).
intros.
unfold PMAP_Lib.allocpg.
unfold PMAP_Lib.PMAP_alloc.
unfold C_PMAP.State.Stack.StackRets.
destruct s; simpl in *.
destruct memory; simpl in *.
exists 
{| C_PMAP.State.stack := (C_ALE.State.Stack.pop +++ C_ALE.State.Stack.setRet pg) stack;
    C_PMAP.State.memory := 
         {| PMAP_State.data := (fun l => if Zeq_bool (Pg l) pg then newpagesrc l else data l);
             PMAP_State.pagemap := pagemap;
             PMAP_State.allocdata := (fun x : Z => if Zeq_bool x pg then true else allocdata x)
          |}
|}; simpl.
repeat split; auto.
intros; auto.
Zeq_pull'; Zeq_norm; repeat autoclean; simpl in *; autosolve.
rewrite H0 in H; autosolve.
intros.
rewrite H0; rewrite Zeq_bool_refl; auto.
Qed.

Lemma PMAP_ALE_allocdata : forall x0 bs s,
ValidPg x0 = true ->
PMAP_ALE.Repr_Mem bs s ->
ALE_State.allocdata s x0 = false ->
PMAP_State.allocdata bs x0 = false.
intros.
red in H0; repeat autoclean.
makenew' (PMAP_State.allocdata bs x0); autosolve.
rewrite <- (Pg_simpl x0 0) in *.
specialize (H2 (x0 * PGSIZE + Off 0)%Z).
specialize (H3 (x0 * PGSIZE + Off 0)%Z).
makenew' (ALE_State.PtDom (x0 * PGSIZE + Off 0)%Z).
rewrite H3 in H9; autosolve.
apply ValidPg_ValidPA_zero; auto.
rewrite Pg_simpl in *; auto.
rewrite H2 in H9; autosolve.
rewrite H1 in H9; autosolve.
apply ValidPg_ValidPA_zero; auto.
rewrite Pg_simpl in *; auto.
Qed.

Lemma vvv : forall x0 bs s bs' s',
  (1 <= x0 < NPAGES)%Z ->
  s.(ALE_State.allocdata) x0 = false ->
  (forall l, (Pg l = x0) -> bs'.(PMAP_State.data) l = 
                                s'.(ALE_State.data) l ) ->
  PMAP_ALE.Repr_Mem bs s ->
  PMAP_Lib.PMAP_alloc x0 bs bs' ->
  ALE_Lib.ALE_alloc x0 s s' ->
  PMAP_ALE.Repr_Mem bs' s'.
intros.
assert (ValidPg x0 = true).
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.
red in H2; repeat autoclean.
red in H3; decompose record H3; clear H3.
red in H4; decompose record H4; clear H4.

red; repeat split.
intros.
destruct (Zeq_dec (Pg l) x0).
apply H1; autosolve.
rewrite H15 in H14; simpl in *.
Zeq_pull'; Zeq_norm; autosolve.
rewrite H13; autosolve.
rewrite H3; autosolve.
apply H2; autosolve.
rewrite <- H6; autosolve.
makenew' (ALE_State.PtDom l); autosolve.
rewrite H7 in H14; autosolve.

intros.
rewrite H15.
rewrite H17.
rewrite H6; autosolve.

intros.
rewrite H15.
rewrite H7; autosolve.
Zeq_pull'; Zeq_norm; autosolve.
subst.
rewrite H8 in H0; autosolve.

intros.
rewrite H17.
rewrite H8; autosolve.
Zeq_pull'; Zeq_norm; autosolve.

intros.
rewrite H16.
specialize (H9 vpg H4).
makenew' (PMAP_State.pagemap bs vpg); autosolve.
rewrite H3; autosolve.
destruct (Zeq_dec (Pg (ALE_State.PTROOT + vpg * 8)) x0); autosolve.
rewrite <- e in *; clear e.
rewrite H8 in H0; autosolve.
apply ALE_State.PtDom_Valid; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
apply H8.
apply ALE_State.PtDom_Valid; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
(************)
rewrite H3; autosolve.
destruct (Zeq_dec (Pg (ALE_State.PTROOT + vpg * 8)) x0); autosolve.
rewrite <- e in *; clear e.
rewrite H8 in H0; autosolve.
apply ALE_State.PtDom_Valid; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
apply H8.
apply ALE_State.PtDom_Valid; autosolve.
apply ALE_State.PtDom_PTROOT; autosolve.
(************)


red; intros.
red in H10.
unfold ALE_State.trans in *; simpl.
rewrite H3; autosolve.
apply H10; autosolve.
destruct (Zeq_dec (Pg (ALE_State.PTROOT + (Pg l) * 8)) x0); autosolve.
rewrite <- e in *; clear e.
rewrite H8 in H0; autosolve.
apply ALE_State.PtDom_Valid; autosolve.
apply ValidPg_ValidVPg.
apply ValidPA_ValidPg; auto.
apply ALE_State.PtDom_PTROOT; autosolve.
apply ValidVA_ValidVPg.
apply ValidPA_ValidVA; autosolve.
apply H8.
apply ALE_State.PtDom_Valid; autosolve.
apply ValidPg_ValidVPg.
apply ValidPA_ValidPg; auto.
apply ALE_State.PtDom_PTROOT; autosolve.
apply ValidVA_ValidVPg.
apply ValidPA_ValidVA; autosolve.
rewrite H18; autosolve.
Qed.


Lemma ttt : forall x0 bs s bs' s',
  (1 <= x0 < NPAGES)%Z ->
  s.(C_ALE.State.memory).(ALE_State.allocdata) x0 = false ->
  (forall l, (Pg l = x0) -> bs'.(C_PMAP.State.memory).(PMAP_State.data) l = 
                                s'.(C_ALE.State.memory).(ALE_State.data) l ) ->
  C_PMAP_ALE.CRepr bs s ->
  PMAP_Lib.allocpg x0 bs bs' ->
  ALE_Lib.allocpg x0 s s' ->
  C_PMAP_ALE.CRepr bs' s'.
intros.
red in H2; decompose record H2; clear H2.
red in H3; decompose record H3; clear H3.
red in H4; decompose record H4; clear H4.
red; split.
rewrite H5 in H7.
red in H7, H8.
rewrite H7; rewrite H8; autosolve.

(****)
clear H7 H8 H5.
eapply vvv; eauto.
Qed.

Lemma alloc_weak1 : ActionGTE (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr PMAP_Lib.mem_alloc_spec)
                                  ALE_Lib.mem_alloc_spec.
split.
(* pre *)
intros.
unfold PMAP_Lib.mem_alloc_spec, ALE_Lib.mem_alloc_spec in *; simpl in *; repeat autoclean.
red in H; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_ALE.State.StackCheck in *; simpl in *.
rewrite H in *.
red in H2; repeat autoclean.
repeat split; autosolve.

(* rel *)
intros.
unfold PMAP_Lib.mem_alloc_spec, ALE_Lib.mem_alloc_spec in *; simpl in *; repeat autoclean.
inversion_clear H0; repeat autoclean. 

(* positive case - allocated x0 *)
destruct (exists_alloc x0 s'.(C_ALE.State.memory).(ALE_State.data) bs) as [bs'].
(* exists_alloc: need to show that x0 is currently unallocated *)
destruct H3.
red in H9; decompose record H9; clear H9.
rewrite <- (Pg_simpl x0 0%Z) in *.
destruct (bool_dec (ALE_State.PtDom (x0 * PGSIZE + Off 0%Z)) true).
apply H11; auto.
apply ValidPg_ValidPA_zero.
unfold ValidPg.
rewrite Pg_simpl in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
rewrite H12; autosolve.
apply ValidPg_ValidPA_zero.
unfold ValidPg.
rewrite Pg_simpl in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (ALE_State.PtDom (x0 * 4096 + Off 0)); autosolve.
elimtype False; apply n; auto.
(*****)

exists bs'; repeat autoclean.
split.
eapply ttt; eauto.
left.
exists x0.
repeat (split; autosolve).
red in H3; repeat autoclean.
apply PMAP_ALE_allocdata with (C_ALE.State.memory s); autosolve.
unfold ValidPg; Zeq_pull'; Zeq_norm; autosolve.

(* negative case *)
exists (PMAP_Lib.allocfail bs).
split.
red in H3; repeat autoclean.
unfold PMAP_Lib.allocfail, ALE_Lib.allocfail; simpl.
red; split; simpl; autosolve.
rewrite <- H0; autosolve.
right; autosolve.
Qed.


Lemma exists_free : forall pg s, exists s', PMAP_Lib.freepg pg s s'.
intros.
unfold PMAP_Lib.freepg.
unfold PMAP_Lib.PMAP_free; simpl.
unfold C_PMAP.State.Stack.StackRets; simpl.
destruct s; simpl.
destruct memory; simpl.
exists 
{| C_PMAP.State.stack := (C_PMAP.State.Stack.pop +++ C_PMAP.State.Stack.setRet 0%Z) stack;
    C_PMAP.State.memory := 
         {| PMAP_State.data := data;
             PMAP_State.pagemap := pagemap;
             PMAP_State.allocdata := (fun x : Z => if Zeq_bool x pg then false else allocdata x)
          |}
|}; simpl.
repeat split; auto.
Qed.

Lemma ReprMem_free_preserve : forall pg bs bs' s s',
(0 < pg < NPAGES)%Z ->
PMAP_State.allocdata bs pg = true ->
PMAP_ALE.Repr_Mem bs s -> 
ALE_Lib.ALE_free pg s s' ->
PMAP_Lib.PMAP_free pg bs bs' ->
PMAP_ALE.Repr_Mem bs' s'.
intros.
red in H1; repeat autoclean.
repeat split; autosolve.

intros.
rewrite PMAP_Lib.PMAP_free_data_unch with _ pg bs bs'; autosolve.
rewrite ALE_Lib.ALE_free_data_unch with _ pg s s'; autosolve.
apply H1; autosolve.
apply PMAP_Lib.PMAP_free_allocdata_pullback with pg bs'; autosolve.
assert (PMAP_State.allocdata bs (Pg l) = true).
apply PMAP_Lib.PMAP_free_allocdata_pullback with pg bs'; autosolve.
red in H2; decompose record H2; clear H2.
rewrite H16.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
red in H3; decompose record H3; clear H3.
rewrite H18 in H12.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct (bool_dec (ALE_State.PtDom l) false).
rewrite <- H4; autosolve.
red in n.
apply H6; autosolve.
makenew' (ALE_State.PtDom l); autosolve.
elimtype False; apply n; auto.

intros.
red in H2; decompose record H2; clear H2.
red in H3; decompose record H3; clear H3.
rewrite H15; rewrite H17.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H4; autosolve.

intros.
red in H3; decompose record H3; clear H3.
rewrite H15.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply H5; autosolve.

intros.
red in H2; decompose record H2; clear H2.
rewrite H15.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H5 in H0; autosolve.
apply H6; autosolve.

intros.
specialize (H7 vpg H11).
red in H3; decompose record H3; clear H3.
rewrite H15.
makenew' (PMAP_State.pagemap bs vpg); autosolve.
red in H2; decompose record H2; clear H2.
rewrite H3; autosolve.
rewrite H17; simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H5 in H0; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ALE_State.PtDom_PTROOT; auto.
apply H6; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ALE_State.PtDom_PTROOT; auto.
red in H2; decompose record H2; clear H2.
rewrite H3; autosolve.
rewrite H17; simpl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H5 in H0; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ALE_State.PtDom_PTROOT; auto.
apply H6; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ALE_State.PtDom_PTROOT; auto.

red; intros.
unfold ALE_State.trans.
red in H2; decompose record H2; clear H2.
rewrite H12; autosolve.
apply H8; autosolve.
rewrite H14.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
subst.
rewrite H5 in H0; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ValidPg_ValidVPg.
apply ValidPA_ValidPg; auto.
apply ALE_State.PtDom_PTROOT; auto.
apply ValidPg_ValidVPg.
apply ValidPA_ValidPg; autosolve.
apply H6; autosolve.
apply ALE_State.PtDom_Valid; auto.
apply ValidPg_ValidVPg.
apply ValidPA_ValidPg; auto.
apply ALE_State.PtDom_PTROOT; auto.
apply ValidPg_ValidVPg.
apply ValidPA_ValidPg; autosolve.

red in H2; decompose record H2; clear H2.
rewrite H14; auto.
Qed.

Lemma freepg_repr_preserve : forall pg bs bs' s s',
(0 < pg < NPAGES)%Z ->
PMAP_State.allocdata (C_PMAP.State.memory bs) pg = true ->
C_PMAP_ALE.CRepr bs s -> 
ALE_Lib.freepg pg s s' ->
PMAP_Lib.freepg pg bs bs' ->
C_PMAP_ALE.CRepr bs' s'.
intros.
red in H1; repeat autoclean.
red in H2; repeat autoclean.
red in H3; repeat autoclean.
red; split.
rewrite H6.
rewrite H7.
rewrite H1; autosolve.
eapply ReprMem_free_preserve; eauto.
Qed.

Lemma free_weak1 : ActionGTE (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr PMAP_Lib.mem_free_spec)
                                  ALE_Lib.mem_free_spec.
split.

(* pre *)
intros.
unfold PMAP_Lib.mem_free_spec, ALE_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
red in H; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_ALE.State.StackCheck in *; simpl in *.
unfold C_PMAP.State.StackGet, C_ALE.State.StackGet in *; simpl in *.
rewrite H in *.
red in H4; repeat autoclean.
repeat split; autosolve.

makenew ((nth 0 (C_ALE.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z)).
rewrite <- (Pg_simpl H11 0%Z).
rewrite <- (Pg_simpl H11 0%Z) in H2.
rewrite <- H5; autosolve.
apply ValidPg_ValidPA_zero.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (ALE_State.PtDom (H11 * PGSIZE + Off 0%Z)); autosolve.
rewrite H6 in H2; autosolve.
apply ValidPg_ValidPA_zero.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.


makenew (nth 0 (C_ALE.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z).
makenew' (ALE_State.PtDom (H11 * PGSIZE + Off 0)); autosolve.
destruct (ValidVPg_existsVA H11).
apply ValidPg_ValidVPg.
unfold ValidPg.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
destruct H13.
rewrite <- H15 in *; clear H11 H15.
rewrite <- ALE_State.PtDom_zero in H14.
rewrite H6 in H2; autosolve.
unfold ValidPA, ValidVA in *.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.

(* rel *)
intros.
unfold PMAP_Lib.mem_free_spec, ALE_Lib.mem_free_spec in *; simpl in *; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_ALE.State.StackCheck in *; simpl in *.
unfold C_PMAP.State.StackGet, C_ALE.State.StackGet in *; simpl in *.
unfold C_ALE.State.StackAct, C_ALE.State.MemAct in *; simpl in *.
destruct (exists_free (nth 0 (C_PMAP.State.Stack.getArgs (C_PMAP.State.stack bs)) 0%Z) bs).
exists x0; split; autosolve.
eapply freepg_repr_preserve; eauto.
red in H5; repeat autoclean.
rewrite H5; autosolve.
Qed.

Lemma pmap_set_weak1 : ActionGTE (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr PMAP_Lib.pmap_set_spec)
                                  pmap_sig.pmap_set_spec.
split.

intros.
unfold PMAP_Lib.pmap_set_spec, pmap_sig.pmap_set_spec in *; simpl in *.
repeat autoclean.
red in H; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_PMAP.State.StackGet in *.
unfold C_ALE.State.StackCheck, C_ALE.State.StackGet in *.
rewrite H in *.
red in H5; repeat autoclean.
repeat split; autosolve.
makenew' (ValidPg (nth 0 (C_ALE.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z)); autosolve.

(* rel *)
intros.
unfold PMAP_Lib.pmap_set_spec, pmap_sig.pmap_set_spec in *; simpl in *.
repeat autoclean.
eexists; split; autosolve.
red in H; repeat autoclean.
sact_auf; simpl in *.
unfold C_PMAP.State.StackAct, C_ALE.State.StackAct in *; simpl in *.
unfold C_PMAP.State.MemAct, C_ALE.State.MemAct in *; simpl in *.
unfold C_PMAP.State.StackCheck, C_PMAP.State.StackGet in *.
unfold C_ALE.State.StackCheck, C_ALE.State.StackGet in *.
rewrite H in *; simpl in *.
red in H2; repeat autoclean.
rewrite H2 in *; simpl in *.
red in H12; repeat autoclean.
makenew (nth 0 (C_PMAP.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z).
makenew (nth 1 (C_PMAP.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z).
makenew' (ValidPg H19); autosolve.
clear H11 H7.
red; repeat split; simpl in *; autosolve.

intros.
makenew'  (Zeq_bool
      (ALE_State.trans (ALE_State.PTROOT + H19 * 8) (C_ALE.State.memory s)) l); Zeq_norm; simpl in *; autosolve.
rewrite H14 in H11; autosolve.
rewrite <- H25.
rewrite H17.
apply ALE_State.PtDom_PTROOT; autosolve.
apply ALE_State.PtDom_Valid; auto.
rewrite H12; autosolve.

intros.
rewrite H17; autosolve.
makenew' (Zeq_bool vpg H19); Zeq_norm.
rewrite <- H23 in *; simpl in *.
rewrite Zeq_bool_refl.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
makenew' (Zeq_bool (ALE_State.PTROOT + H19 * 8) (ALE_State.PTROOT + vpg * 8));
Zeq_norm; simpl in *; autosolve.
specialize (H16 vpg H7).
makenew' (PMAP_State.pagemap (C_PMAP.State.memory bs) vpg); autosolve.
apply ALE_State.PtDom_Valid; auto.

red; intros.
(* can not use trans_setMem_unch - we are actually updating PTABLES *)
unfold ALE_State.trans.
unfold ALE_State.setMem; simpl.
destruct (Zeq_dec H19 (Pg l)).
rewrite e in *.
rewrite ValidPA_ValidPg in H24; autosolve.
rewrite H17.
replace (Zeq_bool (ALE_State.PTROOT + H19 * 8) (ALE_State.PTROOT + Pg l * 8)) with false.
apply H17; autosolve.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
apply ALE_State.PtDom_Valid; autosolve.
Qed.

Lemma pmap_lookup_weak1 : ActionGTE (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr PMAP_Lib.pmap_lookup_spec)
                                  pmap_sig.pmap_lookup_spec.
split.

intros.
unfold PMAP_Lib.pmap_lookup_spec, pmap_sig.pmap_lookup_spec in *; simpl in *.
repeat autoclean.
red in H; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_PMAP.State.StackGet in *.
unfold C_ALE.State.StackCheck, C_ALE.State.StackGet in *.
rewrite H in *.
red in H3; repeat autoclean.
repeat split; autosolve.
makenew' (ValidPg (nth 0 (C_ALE.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z)); autosolve.

(* rel *)
intros.
unfold PMAP_Lib.pmap_set_spec, pmap_sig.pmap_set_spec in *; simpl in *.
repeat autoclean.
eexists; split; autosolve.
sact_auf; simpl in *.
unfold C_PMAP.State.StackAct, C_ALE.State.StackAct in *; simpl in *.
unfold C_PMAP.State.MemAct, C_ALE.State.MemAct in *; simpl in *.
unfold C_PMAP.State.StackCheck, C_PMAP.State.StackGet in *.
unfold C_ALE.State.StackCheck, C_ALE.State.StackGet in *.
unfold C_PMAP.State.MemGet, C_ALE.State.MemGet in *; simpl in *.
red in H; red in H2; repeat autoclean.
rewrite H in *; rewrite H0 in *; simpl in *.
set (repr_mem_bs_s := H2); generalize repr_mem_bs_s; clear repr_mem_bs_s; intros repr_mem_bs_s.
red in H2; repeat autoclean.
specialize (H12 (nth 0 (C_PMAP.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z) H4).

makenew' (PMAP_State.pagemap (C_PMAP.State.memory bs)
        (nth 0 (C_PMAP.State.Stack.getArgs (C_ALE.State.stack s)) 0%Z)); autosolve.
red; simpl.
unfold ALE_State.getMem.
rewrite H13; autosolve.
rewrite H12.
split; auto.
apply ALE_State.PtDom_Valid; autosolve.

red; simpl.
unfold ALE_State.getMem.
rewrite H13; autosolve.
rewrite H12.
split; auto.
apply ALE_State.PtDom_Valid; autosolve.
Qed.



(**************************************************************)

Lemma PMAP_ALE_mem_exists : forall (s : ALE_State.t), 
     (forall l, ValidPA l = true -> ALE_State.PtDom l = true -> 
                       s.(ALE_State.allocdata) (Pg l) = true) ->
     (forall pg, ValidPg pg = true -> s.(ALE_State.data) (ALE_State.PTROOT + pg * 8) = pg) ->
     (forall pg, ValidVPg pg = true -> ValidPg pg = false -> 
                            s.(ALE_State.data) (ALE_State.PTROOT + pg * 8)= 0%Z) ->
      s.(ALE_State.tlb) = true ->
exists bs, (PMAP_ALE.Repr_Mem bs s /\ 
forall pg : Z, ValidVPg pg = true -> ValidPg pg = false -> PMAP_State.pagemap bs pg = None).
intros.
exists {| 
   PMAP_State.data := (fun l => if s.(ALE_State.allocdata) (Pg l) then s.(ALE_State.data) l else 0%Z);
   PMAP_State.pagemap := (fun pg => if ValidPg pg then Some pg else None); 
   PMAP_State.allocdata := (fun pg => 
             if ALE_State.PtDom (pg * PGSIZE +0%Z) then false else s.(ALE_State.allocdata) pg)
|}; split; simpl.

red; simpl.
repeat split; autosolve.

intros.
makenew' (ALE_State.PtDom (Pg l * PGSIZE + 0%Z)); autosolve.
rewrite H4; auto.

intros.
rewrite <- ALE_State.PtDom_zero.
rewrite H4; auto.

intros.
rewrite <- ALE_State.PtDom_zero.
rewrite H4; auto.

intros.
makenew' (ValidPg vpg); autosolve.
apply H0; auto.
apply H1; auto.

red; intros.
unfold ALE_State.trans.
rewrite H0.
rewrite Addr_parts; auto.
apply ValidPA_ValidPg; auto.

intros.
rewrite H4; auto.
Qed.

Require Import as_sig.

Lemma kernel_init_weak : ActionGTE pmap_sig.ALE_kernel_init_spec
            (C_PMAP_ALE.RF.reprAction C_PMAP_ALE.CRepr as_sig.PMAP_kernel_init_spec).
split.

intros.
unfold as_sig.PMAP_kernel_init_spec, pmap_sig.ALE_kernel_init_spec in *; simpl in *; repeat autoclean.
unfold C_PMAP.State.StackCheck, C_ALE.State.StackCheck in *.
destruct s; simpl in *.
destruct stack; simpl in *; autosolve.
destruct f; simpl in *; autosolve.
destruct l; simpl in *; autosolve.
repeat autoclean.
destruct (PMAP_ALE_mem_exists memory) as [bmem]; auto.
repeat autoclean.
exists {| C_PMAP.State.stack := C_Stack.CallFrame nil :: stack; 
              C_PMAP.State.memory := bmem |}; simpl.
split; auto.
red; simpl.
split; auto.

intros.
unfold as_sig.PMAP_kernel_init_spec, as_sig.AS_kernel_init_spec in *; simpl in *; repeat autoclean.
destruct s; simpl in *.
destruct (PMAP_ALE_mem_exists memory) as [bmem]; auto.
repeat autoclean.
destruct H0 with {| C_PMAP.State.stack := stack; 
              C_PMAP.State.memory := bmem |}; simpl in *; autosolve.
red; simpl.
split; auto.
split; auto.
Qed.


(******************************************************************)
Require Import label.

Definition ref := C_PMAP_ALE.EqRef.

Lemma WF_kernel_pmap : forall kernel_code kernel_spec as_lib_code,
         C_ALE.SCAP.SpecHeap.SpecHeap_disjoint
                 (C_ALE.SCAP.SpecHeap.SpecHeap_join ALE_Lib.ale_lib_spec pmap_sig.pmap_spec pmap_sig.lib_pmap_disj)
                    (C_PMAP_ALE.PsiMake kernel_spec) ->
         C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine 
                                PMAP_Lib.pmap_lib_spec kernel_code kernel_spec ->
         
         C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine 
                              ALE_Lib.ale_lib_spec as_lib_code pmap_sig.pmap_spec ->
         C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine 
                 (C_ALE.SCAP.SpecHeap.SpecHeap_join ALE_Lib.ale_lib_spec pmap_sig.pmap_spec pmap_sig.lib_pmap_disj)
                                        (C_PMAP_ALE.CMake kernel_code) 
                                        (C_PMAP_ALE.PsiMake kernel_spec).
intros.
set (x := ref PMAP_Lib.pmap_lib_spec kernel_code kernel_spec).
eapply C_ALE.SCAP.WFCode_stronghypo; eauto.
red; intros.
destruct l; autosolve; simpl.
apply pmap_set_weak1.
apply pmap_lookup_weak1.
apply alloc_weak1.
apply free_weak1.
Qed.

Lemma WF_kernel_pmap' : forall kernel_code kernel_spec as_lib_code
         (disj : C_ALE.SCAP.SpecHeap.SpecHeap_disjoint (C_PMAP_ALE.PsiMake kernel_spec) pmap_sig.pmap_spec)
         (disj' : C_ALE.SCAP.SpecHeap.SpecHeap_disjoint (C_PMAP_ALE.PsiMake kernel_spec) ALE_Lib.ale_lib_spec),
         C_ALE.SCAP.CH.CodeHeap_disjoint as_lib_code (C_PMAP_ALE.CMake kernel_code) ->
         C_PMAP.SCAP.WFCode C_PMAP.State.expr_to_Prop C_PMAP.Mc_Machine 
                       PMAP_Lib.pmap_lib_spec kernel_code kernel_spec ->
         C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine 
                       ALE_Lib.ale_lib_spec as_lib_code pmap_sig.pmap_spec ->

      C_ALE.SCAP.WFCode C_ALE.State.expr_to_Prop C_ALE.Mc_Machine 
                ALE_Lib.ale_lib_spec 
                (C_ALE.SCAP.CH.CodeHeap_join (C_PMAP_ALE.CMake kernel_code) as_lib_code) 
                (C_ALE.SCAP.SpecHeap.SpecHeap_join (C_PMAP_ALE.PsiMake kernel_spec) pmap_sig.pmap_spec disj).

intros.
apply C_ALE.SCAP.WFCode_linking_smp with pmap_sig.lib_pmap_disj; auto.
eapply WF_kernel_pmap; eauto.
apply C_ALE.SCAP.SpecHeap.SpecHeap_disjoint_joinleft; auto.
apply C_ALE.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
apply C_ALE.SCAP.SpecHeap.SpecHeap_disjoint_sym; auto.
Qed.
