(* pmap_lib.v *)
(* Specifications of the stubs of the PMAP memory model *)
(* mem_alloc, mem_free, pt_set, pt_lookup *)

Require Import DecidableType.
Require Import ZArith.
Require Import NArith.
Require Import Bool.

Require Import label.
Require Import spec.
Require Import scap.
Require Import zbool_assist.

Require Import Logic.FunctionalExtensionality.

Require Export common.
Require Import sact.
Require Import actsmp.

Require Import c.
Require Import pmap.

Module PMAP_Lib.

Import C_PMAP.

Definition allocfail : SAct (State := State) :=
  StackAct (pop +++ setRet 0%Z).


Definition PMAP_alloc (pg : Z) : PMAP_State.t -> PMAP_State.t -> Prop :=
  fun s s' => 
	 (forall l, s.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\ 
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then true else s.(allocdata) x)) /\
	(s'.(pagemap) = s.(pagemap)).

Definition allocpg (pg : Z) : State -> State -> Prop :=
  fun s s' => PMAP_alloc pg s.(memory) s'.(memory) /\ StackRets pg s.(stack) s'.(stack).


Lemma PMAP_alloc_trans_unch :  forall x0 l s s',
PMAP_Lib.PMAP_alloc x0 s s' ->
PMAP_State.trans l s = PMAP_State.trans l s'.
intros.
red in H; decompose record H; clear H.
unfold PMAP_State.trans.
rewrite H3.
makenew' (PMAP_State.pagemap s (Pg l)); autosolve.
Qed.

Lemma PMAP_alloc_trans'_unch :  forall x0 l s s',
PMAP_Lib.PMAP_alloc x0 s s' ->
PMAP_State.trans' l s = PMAP_State.trans' l s'.
intros.
unfold PMAP_State.trans'.
makenew' (isPhys l); autosolve.
apply PMAP_alloc_trans_unch with x0; auto.
Qed.

Lemma PMAP_alloc_allocdata_unch : forall pg x0 s s',
pg <> x0 ->
PMAP_Lib.PMAP_alloc x0 s s' ->
PMAP_State.allocdata s' pg = PMAP_State.allocdata s pg.
intros.
red in H0; decompose record H0; clear H0.
rewrite H3; simpl.
Zeq_pull'; Zeq_norm; simpl; autosolve.
Qed.

Lemma PMAP_alloc_pagemap_unch : forall x0 s s',
PMAP_Lib.PMAP_alloc x0 s s' ->
PMAP_State.pagemap s' = PMAP_State.pagemap s.
intros.
red in H; decompose record H; clear H; autosolve.
Qed.

Lemma PMAP_alloc_data_unch : forall l x0 s s',
PMAP_State.allocdata s (Pg l) = true ->
PMAP_Lib.PMAP_alloc x0 s s' ->
PMAP_State.data s' l = PMAP_State.data s l.
intros.
red in H0; decompose record H0; clear H0.
rewrite H1; autosolve.
Qed.



Definition mem_alloc_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 0) s = true /\ StackCheck isFrame s = true
     );

rel := (fun s s' =>
    (exists pg, (1 <= pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = false /\ allocpg pg s s')
    \/
    s' = allocfail s)
|}.


Definition PMAP_free (pg : Z) : PMAP_State.t -> PMAP_State.t -> Prop :=
  fun s s' => 
	 (forall l, s'.(allocdata) (Pg l) = true -> s'.(data) l = s.(data) l) /\ 
	 (s'.(allocdata) = (fun x => if Zeq_bool x pg then false else s.(allocdata) x)) /\
	(s'.(pagemap) = s.(pagemap)).


Definition freepg (pg : Z) : State -> State -> Prop :=
   fun s s' => PMAP_free pg s.(memory) s'.(memory) /\ StackRets 0%Z s.(stack) s'.(stack).


Lemma PMAP_free_data_unch : forall l x0 s s',
PMAP_State.allocdata s' (Pg l) = true ->
PMAP_Lib.PMAP_free x0 s s' ->
PMAP_State.data s' l = PMAP_State.data s l.
intros.
red in H0; decompose record H0; clear H0.
rewrite H1; autosolve.
Qed.

Lemma PMAP_free_allocdata_unch : forall pg x0 s s',
pg <> x0 ->
PMAP_Lib.PMAP_free x0 s s' ->
PMAP_State.allocdata s' pg = PMAP_State.allocdata s pg.
intros.
red in H0; decompose record H0; clear H0.
rewrite H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.

Lemma PMAP_free_allocdata_pullback : forall pg x0 s s',
PMAP_Lib.PMAP_free x0 s s' ->
PMAP_State.allocdata s' pg = true ->
PMAP_State.allocdata s pg = true.
intros.
red in H; decompose record H; clear H.
rewrite H3 in *; clear H3.
Zeq_pull'; Zeq_norm; simpl in *; autosolve.
Qed.



Definition mem_free_spec : Action State :=
{|
pre := (fun s => StackCheck (getArgsCheck 1) s = true /\
                                 let pg := nth 0 (StackGet getArgs s) 0%Z in
                                 (0 < pg < NPAGES)%Z /\ s.(memory).(allocdata) pg = true);

rel := (fun s s' => freepg (nth 0 (StackGet getArgs s) 0%Z) s s')
|}.

Definition pmap_set_spec : Action State := 
SAct_Action (fun s => StackCheck (getArgsCheck 2) s &&
                                  let (vpage, entry) := (nth 0 (StackGet getArgs s) 0%Z, nth 1 (StackGet getArgs s) 0%Z) in
	 			  ValidVPg vpage && negb (ValidPg vpage) && ValidPg entry && 
                                  (if Zeq_bool entry 0%Z then true else s.(memory).(allocdata) entry)
                   )
	           (fun s => let (vpage, entry) := (nth 0 (StackGet getArgs s) 0%Z, nth 1 (StackGet getArgs s) 0%Z) in
				  SAct_chain 
                                     (StackAct (pop +++ setRet 0%Z)) 
	                             (MemAct (fun m => {| data := m.(data) ; allocdata := m.(allocdata); 
                                          pagemap := 
                                                 fun pg' => if Zeq_bool pg' vpage then 
							 (if Zeq_bool entry 0%Z then None else Some entry) 
                                                         else m.(pagemap) pg' |})) s).

Definition pmap_lookup_spec : Action State :=
SAct_Action (fun s => StackCheck (getArgsCheck 1) s &&
                                  let vpage := nth 0 (StackGet getArgs s) 0%Z in
                                  ValidVPg vpage && negb (ValidPg vpage))
                   (fun s => match (s.(memory).(pagemap) (nth 0 (StackGet getArgs s) 0%Z)) with
                           | Some entry => StackAct (pop +++ setRet entry) s
                           | None => StackAct (pop +++ setRet 0%Z) s
                           end).

Definition pmap_lib_spec : SpecHeap := fun l =>
  match l with
  | mem_free_label => Some mem_free_spec
  | mem_alloc_label => Some mem_alloc_spec
  | pmap_set_label => Some pmap_set_spec
  | pmap_lookup_label => Some pmap_lookup_spec
  | _ => None
  end.

End PMAP_Lib.
