(* process.v *)
(* Defined the notion of process as used in our verification framework *)
(* Requires the type of "decider", which is a type of conditionals used in the machine *)

Require Import Equalities.

Require Import spec.
Require Import label.
Require Import zbool_assist.
Require Import specheap.

Module Process (MachineState : Typ) (MachineOp : Typ) (Decider : Typ) <: Typ.
Module S := SpecHeap MachineState.
Import S.

Definition Op := MachineOp.t.
Definition Label := Label.

Definition Machine := Op -> option (Action MachineState.t) .

Definition getOpAction (mach : Machine) (op : Op) (a : (Action MachineState.t)) := mach op = Some a.
Definition getLabelAction (psi : SpecHeap) (l : Label) (a : Action MachineState.t) := psi l = Some a.
Hint Unfold getOpAction.
Hint Unfold getLabelAction.

Section ProcessSection.

Definition Decider := Decider.t.
Definition Decider_to_Prop := Decider -> (MachineState.t -> Prop).


Inductive Process 
: Type :=
| Inil : Process
| Iperf : Op -> Process
| Icall : Label -> Process
| Iseq : Process -> Process -> Process
| Iplus : Decider-> Process -> Process -> Process
.
End ProcessSection.

Definition t := Process.

Section WFP.
Variable dp : Decider_to_Prop.
Inductive WFProc : Machine -> SpecHeap -> Action MachineState.t -> Process -> Prop :=
| WFProc_nil : forall Mach Psi, WFProc Mach Psi (ActionID MachineState.t) Inil
| WFProc_perf: forall Mach Psi i g, getOpAction Mach i g -> WFProc Mach Psi g (Iperf i)
| WFProc_call: forall Mach Psi l g, getLabelAction Psi l g -> WFProc Mach Psi g (Icall l)
| WFProc_weaken: forall Mach Psi g g' I, ActionGTE g g' -> WFProc Mach Psi g' I -> WFProc Mach Psi g I
| WFProc_seq : forall Mach Psi g' g'' I' I'', 
  WFProc Mach Psi g' I' -> WFProc Mach Psi g'' I'' -> WFProc Mach Psi (ActionChain g' g'') (Iseq I' I'')
| WFProc_plus : forall Mach Psi p g' g'' I' I'',
  WFProc Mach Psi g' I' -> WFProc Mach Psi g'' I'' -> 
  WFProc Mach Psi (ActionPlus (dp p) g' g'') (Iplus p I' I'').

(* We can always strenthen the hypothetical *)

Lemma WFProc_strengthen : forall Mach psi psi' g I, 
  SpecHeap_stronger psi psi' ->
  WFProc Mach psi' g I ->
  WFProc Mach psi g I.
intros.
induction H0; try solve [constructor; auto].

red in H0.
red in H.
specialize (H l).
rewrite H0 in H.
makenew' (psi l); try contradiction.
apply WFProc_weaken with a; auto.
constructor; auto.

apply WFProc_weaken with g'; auto.
Qed.

Lemma WFProc_bigger_hypo : forall Mach psi psi' g I, 
  SpecHeap_bigger psi psi' ->
  WFProc Mach psi' g I ->
  WFProc Mach psi g I.
intros.
induction H0; try solve [constructor; auto].

red in H0.
red in H.
specialize (H l).
rewrite H0 in H.
makenew' (psi l); try contradiction.
subst; auto.
constructor; auto.

apply WFProc_weaken with g'; auto.
Qed.

Fixpoint genSpec (M : Machine) (I : Process) : option (Action MachineState.t) :=
  match I with 
  | Iperf i => M i
  | Inil => Some (ActionID MachineState.t)
  | Iseq I' I'' => 
    match (genSpec M I', genSpec M I'') with 
    | (Some a', Some a'') => Some (ActionChain a' a'')
    | _ => None
    end
  | _ => None
  end.


Lemma WFgenSpec : forall Mach Psi I (g : Action MachineState.t) , genSpec Mach I = Some g -> WFProc Mach Psi g I.
induction I;intros; inversion H.
constructor; auto.
constructor; auto.
assert (exists g1, genSpec Mach I1 = Some g1).
destruct (genSpec Mach I1).
exists a; auto.
inversion_clear H1.
inversion_clear H0 as [g1].
rewrite H2 in H1.
assert (exists g2, genSpec Mach I2 = Some g2).
destruct (genSpec Mach I2).
exists a; auto.
inversion_clear H1.
inversion_clear H0 as [g2].
rewrite H3 in H1.
inversion_clear H1.
constructor; auto.
Qed.

Lemma WFensure : forall M psi spec trace P, 
  WFProc M psi spec trace -> WFProc M psi (ActionCheck P spec) trace.
intros.
apply WFProc_weaken with spec; auto.
split; auto.
intros.
inversion_clear H0; auto.
Qed.

End WFP.

Hint Resolve WFensure.

Hint Constructors WFProc.

End Process.
