(* scap.v *)
(* Alex Vaynberg *)

(* Contains the definition of SCAP logic that is parametric over the machine *)
(* Also includes correctness lemmas *)

Require Import Equalities.
Require Import Logic.FunctionalExtensionality.
Require Import spec.
Require Import specheap.
Require Import codeheap.
Require Import process.
Require Import Omega.
Require Import zbool_assist.


Module Type SCAPt (MachineState : Typ) (MachineOp : Typ) (Decider : Typ).

Module Process := (Process MachineState MachineOp Decider).
Export Process.
Module CH := CodeHeap Process.
Export CH.
Definition CodeHeap := CodeHeap.
Module SpecHeap := SpecHeap MachineState.
Export SpecHeap.
Hint Resolve ActionGTE_refl.

Section DP.
Variable dp : Decider_to_Prop.

(* Machine, Hypothetical, Code, Specs *)
Inductive WFCode : Machine -> SpecHeap -> CodeHeap -> SpecHeap -> Prop :=
  | WFCode_1 : forall Mach Hypo (C : CodeHeap) Psi (disj : SpecHeap_disjoint Hypo Psi), 
      (forall f g, getLabelAction Psi f g ->
      (exists I, C f = Some I /\ WFProc dp Mach (SpecHeap_join Hypo Psi disj) g I)) ->
    WFCode Mach Hypo C Psi.

Parameter WFCode_stronghypo : forall Mach Hypo SHypo C Psi,
SpecHeap_disjoint SHypo Psi ->
SpecHeap_stronger SHypo Hypo ->
WFCode Mach Hypo C Psi ->
WFCode Mach SHypo C Psi.

Parameter WFCode_biggerhypo : forall Mach Hypo BHypo C Psi,
SpecHeap_disjoint BHypo Psi ->
SpecHeap_bigger BHypo Hypo ->
WFCode Mach Hypo C Psi ->
WFCode Mach BHypo C Psi.


Definition Program := (CodeHeap * MachineState.t * Process)%type.

(* Well-formed Program *)
Inductive WFProgram : Machine -> SpecHeap -> Action MachineState.t -> Program -> Prop :=
  | WFProgram_1 : forall Mach Psi C S I g,
  WFCode Mach SpecHeap_empty C Psi ->
  WFProc dp Mach Psi g I -> pre g S ->
  WFProgram Mach Psi g (C,S,I).

Definition loop : Action MachineState.t :=
  mkAct (fun s => True) (fun s s' => False).

Fixpoint mean (n : nat) (Mach : Machine) (C : CodeHeap) (I : Process)
     : option (Action MachineState.t) :=
match n with
  | 0 => Some loop
  | S n' => match I with
      | Inil => Some (ActionID MachineState.t)
      | Iperf i => Mach i
      | Icall l => match C l with
                   | Some I' => mean n' Mach C I'
                   | None => None
                   end
      | Iseq I1 I2 => match (mean n' Mach C I1, mean n' Mach C I2) with
                      | (Some g1, Some g2) => Some (ActionChain g1 g2)
                      | _ => None
                      end
      | Iplus p I1 I2 => match (mean n' Mach C I1, mean n' Mach C I2) with
                      | (Some g1, Some g2) => Some (ActionPlus (dp p) g1 g2)
                      | _ => None
                      end
      end
  end.



(* Psi holds valid specifications for indexes less than n *)

(* g >= [[C,I]]^n *)
Definition ProcCorrect Mach C I g n := 
  match mean n Mach C I with
  | Some g' => ActionGTE g g'
  | None => False
  end.

(* Psi(l) is weaker than the meaning of C(l) with index less than n *)
Definition CodeCorrect Mach C Psi n := 
forall l cn, cn < n -> 
match (C l, Psi l) with
   | (Some cI, Some cg) => ProcCorrect Mach C cI cg cn
   | (Some _, None) => True
   | (None, Some _) => False
   | _ => True
end.

Parameter CodeCorrect_red : forall Mach C Psi n n', 
  n' < n -> 
  CodeCorrect Mach C Psi n ->
  CodeCorrect Mach C Psi n'.

Parameter Correctness : forall Mach Psi C n I g, 
WFProc dp Mach Psi g I -> 
WFCode Mach (SpecHeap_empty) C Psi ->
CodeCorrect Mach C Psi n ->
  match mean n Mach C I with
  | Some g' => ActionGTE g g'
  | None => False
  end.

End DP.

End SCAPt.











Module SCAP (MachineState : Typ) (MachineOp : Typ) (Decider : Typ)
   <: SCAPt MachineState MachineOp Decider.

Module Process := Process MachineState MachineOp Decider.
Export Process.
Module SpecHeap := SpecHeap MachineState.
Export SpecHeap.
Module CH := CodeHeap Process.
Export CH.
Definition CodeHeap := CodeHeap.

Section DP.
Variable dp : Decider_to_Prop.


(* Machine, Hypothetical, Code, Specs *)
Inductive WFCode : Machine -> SpecHeap -> CodeHeap -> SpecHeap -> Prop :=
  | WFCode_1 : forall Mach Hypo C Psi (disj : SpecHeap_disjoint Hypo Psi), 
      (forall f g, getLabelAction Psi f g ->
      (exists I, C f = Some I /\ WFProc dp Mach (SpecHeap_join Hypo Psi disj) g I)) ->
    WFCode Mach Hypo C Psi.

(* Non-term *)


Lemma WFCode_stronghypo : forall Mach Hypo SHypo C Psi,
SpecHeap_disjoint SHypo Psi ->
SpecHeap_stronger SHypo Hypo ->
WFCode Mach Hypo C Psi ->
WFCode Mach SHypo C Psi.
intros.
apply WFCode_1 with H; intros.
inversion_clear H1.
specialize (H3 f g H2).
destruct H3.
destruct H1.
exists x; split; auto.
apply WFProc_strengthen with (SpecHeap_join Hypo Psi disj); auto.
apply SpecHeap_stronger_join; auto.
Qed.

Lemma WFCode_biggerhypo : forall Mach Hypo BHypo C Psi,
SpecHeap_disjoint BHypo Psi ->
SpecHeap_bigger BHypo Hypo ->
WFCode Mach Hypo C Psi ->
WFCode Mach BHypo C Psi.
intros.
apply WFCode_1 with H; intros.
inversion_clear H1.
specialize (H3 f g H2).
destruct H3.
destruct H1.
exists x; split; auto.
apply WFProc_bigger_hypo with (SpecHeap_join Hypo Psi disj); auto.
apply SpecHeap_bigger_join; auto.
Qed.



(* Program and safety *)


Definition Program := (CodeHeap * MachineState.t * Process)%type.

Lemma WFCode_linking {Mach Hypo1 C1 Psi1 Hypo2 C2 Psi2}
 (H1H2_compat : SpecHeap_compat Hypo1 Hypo2)
 (P1P2_disj : SpecHeap_disjoint Psi1 Psi2)
 (C1C2_disj : CodeHeap_disjoint C1 C2)
 (H1P2_compat : SpecHeap_compat Hypo1 Psi2)
 (H2P1_compat : SpecHeap_compat Hypo2 Psi1) :
 WFCode Mach Hypo1 C1 Psi1 ->
 WFCode Mach Hypo2 C2 Psi2 ->
 WFCode Mach 
   (SpecHeap_add (SpecHeap_sub Hypo1 Psi2) (SpecHeap_sub Hypo2 Psi1))
   (CodeHeap_join C1 C2) (SpecHeap_join Psi1 Psi2 P1P2_disj).

intros.

assert (SpecHeap_disjoint
     (SpecHeap_add (SpecHeap_sub Hypo1 Psi2) (SpecHeap_sub Hypo2 Psi1))
     (SpecHeap_join Psi1 Psi2 P1P2_disj)).
inversion_clear H.
clear H1.
inversion_clear H0.
clear H.
red; intros.
unfold SpecHeap_add, SpecHeap_disjoint, SpecHeap_join, SpecHeap_sub in *.
specialize (H1H2_compat l).
specialize (P1P2_disj l).
specialize (H1P2_compat l).
specialize (H2P1_compat l).
specialize (disj l).
specialize (disj0 l).
simpl in *.
makenew' (Psi1 l); subst; auto.
makenew' (Psi2 l); subst; auto.
makenew' (Hypo1 l); subst; auto.
makenew' (Hypo2 l); subst; auto.
makenew' (Psi2 l); subst; auto.
makenew' (Hypo1 l); subst; auto.

rename H1 into disj.
apply WFCode_1 with disj; intros.
inversion_clear H.
inversion_clear H0.
red in H1.
unfold SpecHeap_join in H1.
makenew' (Psi1 f).
inversion H1; clear H1; subst.
specialize (H2 f g H3).
destruct H2.
destruct H0.
exists x.
split.
unfold CodeHeap_join.
rewrite H0; auto.
apply WFProc_strengthen with (SpecHeap_join Hypo1 Psi1 disj0); auto.
red in H1H2_compat.
red in P1P2_disj.
red in H1P2_compat.
red in H2P1_compat.
red in disj0.
red in disj1.
unfold SpecHeap_add, SpecHeap_sub, SpecHeap_join.
red; intros.
clear disj H H1.
specialize (H1H2_compat l).
specialize (P1P2_disj l).
specialize (H1P2_compat l).
specialize (H2P1_compat l).
specialize (disj0 l).
specialize (disj1 l).
makenew' (Hypo1 l); 
makenew' (Psi1 l); 
makenew' (Psi2 l); 
makenew' (Hypo2 l);
subst;
try match goal with 
| h1 : Some _ = None |- _ => inversion h1
| h1 : None = Some _ |- _ => inversion h1
end; try apply ActionGTE_refl; auto.

specialize (H f g H1).
destruct H.
destruct H.
exists x.
split.
unfold CodeHeap_join.
specialize (C1C2_disj f).
makenew' (C1 f); auto.
rewrite C1C2_disj in H; inversion H.
apply WFProc_strengthen with (SpecHeap_join Hypo2 Psi2 disj1); auto.
red in H1H2_compat.
red in P1P2_disj.
red in H1P2_compat.
red in H2P1_compat.
red in disj0.
red in disj1.
unfold SpecHeap_add, SpecHeap_sub, SpecHeap_join.
red; intros.
clear disj H H0 H2.
specialize (H1H2_compat l).
specialize (P1P2_disj l).
specialize (H1P2_compat l).
specialize (H2P1_compat l).
specialize (disj0 l).
specialize (disj1 l).
makenew' (Hypo1 l); 
makenew' (Psi1 l); 
makenew' (Psi2 l); 
makenew' (Hypo2 l);
subst;
try match goal with 
| h1 : Some _ = None |- _ => inversion h1
| h1 : None = Some _ |- _ => inversion h1
end; try apply ActionGTE_refl; auto.
Qed.

Lemma WFCode_linking_smp  : 
forall Mach hypo_spec lib_code lib_spec main_code main_spec disj
(disj' : SpecHeap_disjoint main_spec lib_spec) (C_disj : CodeHeap_disjoint lib_code main_code),
WFCode Mach hypo_spec lib_code lib_spec ->
WFCode Mach 
                 (SpecHeap_join hypo_spec lib_spec disj)
                                        main_code main_spec ->
WFCode Mach hypo_spec
  (CodeHeap_join main_code lib_code)
  (SpecHeap_join main_spec lib_spec disj').

intros.
inversion_clear H.
inversion_clear H0.
assert (SpecHeap_disjoint hypo_spec (SpecHeap_join main_spec lib_spec disj')).
red; red in disj, disj1.
intros.
clear H H1.
unfold SpecHeap_join in disj1.
specialize (disj l).
specialize (disj1 l).
destruct (hypo_spec l); auto.
unfold SpecHeap_join.
rewrite disj1; auto.

apply WFCode_1 with H0.
intros.
red in H2.
unfold SpecHeap_join in H2.
specialize (H f g).
specialize (H1 f g).
unfold getLabelAction in *.
unfold CodeHeap_join.
makenew' (main_spec f).
inversion H2; subst; clear H2.
destruct H; auto.
destruct H.
rewrite H.
exists x; split; auto.
eapply WFProc_strengthen; eauto.
clear H2.
red; intros.
unfold SpecHeap_join.
red in disj1, H0.
specialize (disj1 l).
specialize (H0 l).
unfold SpecHeap_join in *.
specialize (disj l).
specialize (disj' l).
specialize (disj0 l).
destruct (hypo_spec l).
apply ActionGTE_refl.
destruct (lib_spec l); auto.
rewrite disj1.
apply ActionGTE_refl.
destruct (main_spec l); auto.
apply ActionGTE_refl.

clear H.
destruct H1; auto.
destruct H.
exists x.
red in C_disj.
specialize (C_disj f).
rewrite H in C_disj.
rewrite C_disj.
split; auto. 
eapply WFProc_strengthen; eauto.
clear H1.
red; intros.
unfold SpecHeap_join in *.
specialize (disj l).
specialize (disj' l).
specialize (disj0 l).
destruct (hypo_spec l).
apply ActionGTE_refl.
destruct (lib_spec l); auto.
destruct (main_spec l).
inversion disj'.
apply ActionGTE_refl.
Qed.

Lemma WFCode_simple_linking {Mach CBase PsiBase Hypo C Psi}
 {Psi_disj : SpecHeap_disjoint PsiBase Psi} :
 CodeHeap_disjoint CBase C ->
 SpecHeap_stronger PsiBase Hypo ->
 WFCode Mach Hypo C Psi ->
 WFCode Mach SpecHeap_empty CBase PsiBase ->
 WFCode Mach SpecHeap_empty (CodeHeap_join CBase C) (SpecHeap_join PsiBase Psi Psi_disj).

intros Cdisj str wfc wfbase.
apply WFCode_1 with (SpecHeap_empty_disjoint (SpecHeap_join PsiBase Psi Psi_disj)).
intros.
unfold SpecHeap_join in H; red in H.
makenew' (PsiBase f).
inversion H; clear H; subst.
inversion_clear wfbase.
specialize (H f g H1).
destruct H.
destruct H.
exists x; split.
unfold CodeHeap_join.
rewrite H; auto.
rewrite SpecHeap_join_empty in *.
apply WFProc_strengthen with PsiBase; auto.
unfold SpecHeap_stronger, S.SpecHeap_stronger, SpecHeap_join, SpecHeap_compat, SpecHeap_empty in *.
intros.
specialize (str l).
makenew' (PsiBase l); auto.
apply ActionGTE_refl.

inversion_clear wfc.
specialize (H0 f g H).
destruct H0.
destruct H0.
exists x; split.
unfold CodeHeap_join.
unfold CodeHeap_disjoint in Cdisj.
specialize (Cdisj f).
makenew' (CBase f); auto.
rewrite Cdisj in H0; inversion H0.
rewrite SpecHeap_join_empty in *.
apply WFProc_strengthen with (SpecHeap_join Hypo Psi disj); auto.
clear H2.
unfold SpecHeap_stronger, S.SpecHeap_stronger, SpecHeap_join, SpecHeap_compat, SpecHeap_empty in *.
intros.
specialize (str l).
specialize (Psi_disj l).
specialize (disj l).
makenew' (Hypo l); auto.
makenew' (PsiBase l); auto.
contradiction.
makenew' (Psi l); auto.
makenew' (PsiBase l); subst; auto.
inversion Psi_disj.
apply ActionGTE_refl.
Qed.

Require Import zbool_assist.

Lemma WFCode_join {Mach Hypo C Psi C' Psi'}
 {Psi_disj : SpecHeap_disjoint Psi Psi'} :
 CodeHeap_disjoint C C' ->
 SpecHeap_disjoint Hypo (SpecHeap_join Psi Psi' Psi_disj) ->
 WFCode Mach Hypo C Psi ->
 WFCode Mach Hypo C' Psi' ->
 WFCode Mach Hypo (CodeHeap_join C C') (SpecHeap_join Psi Psi' Psi_disj).
intros.
eapply WFCode_1 with H0.
intros.
red in H3.
unfold SpecHeap_join in H3.
makenew' (Psi f).
(* Psi contains f *)
inversion H3; subst; clear H3.
inversion_clear H1.
specialize (H3 f g H5).
destruct H3.
destruct H1.
exists x.
split.
unfold CodeHeap_join.
rewrite H1; auto.
eapply WFProc_strengthen; eauto.
red; intros.
unfold SpecHeap_join.
destruct (Hypo l); auto.
apply ActionGTE_refl.
destruct (Psi l); auto.
apply ActionGTE_refl.

(* Psi' contains f *)
inversion_clear H2.
specialize (H4 f g H3).
destruct H4.
destruct H2.
exists x.
split.
unfold CodeHeap_join.
specialize (H f).
destruct (C f); auto.
rewrite H in H2; inversion H2.
eapply WFProc_strengthen; eauto.
clear H4.
red; intros.
unfold SpecHeap_join.
clear disj H0.
destruct (Hypo l); auto.
apply ActionGTE_refl.
specialize (Psi_disj l).
destruct (Psi' l); auto.
destruct (Psi l); auto.
inversion Psi_disj.
apply ActionGTE_refl.
Qed.

(* Well-formed Program *)
Inductive WFProgram : Machine -> SpecHeap -> Action MachineState.t -> Program -> Prop :=
  | WFProgram_1 : forall Mach Psi C S I g,
  WFCode Mach SpecHeap_empty C Psi ->
  WFProc dp Mach Psi g I -> pre g S ->
  WFProgram Mach Psi g (C,S,I).

Definition loop : Action MachineState.t :=
  mkAct (fun s => True) (fun s s' => False).

Fixpoint mean (n : nat) (Mach : Machine) (C : CodeHeap) (I : Process)
     : option (Action MachineState.t) :=
match n with
  | 0 => Some loop
  | S n' => match I with
      | Inil => Some (ActionID MachineState.t)
      | Iperf i => Mach i
      | Icall l => match C l with
                   | Some I' => mean n' Mach C I'
                   | None => None
                   end
      | Iseq I1 I2 => match (mean n' Mach C I1, mean n' Mach C I2) with
                      | (Some g1, Some g2) => Some (ActionChain g1 g2)
                      | _ => None
                      end
      | Iplus p I1 I2 => match (mean n' Mach C I1, mean n' Mach C I2) with
                      | (Some g1, Some g2) => Some (ActionPlus (dp p) g1 g2)
                      | _ => None
                      end
      end
  end.


(* g >= [[C,I]]^n *)
Definition ProcCorrect Mach C I g n := 
  match mean n Mach C I with
  | Some g' => ActionGTE g g'
  | None => False
  end.

(* Psi(l) is weaker than the meaning of C(l) with index less than n *)
Definition CodeCorrect Mach C Psi n := 
forall l cn, cn < n -> 
match (C l, Psi l) with
   | (Some cI, Some cg) => ProcCorrect Mach C cI cg cn
   | (Some _, None) => True
   | (None, Some _) => False
   | _ => True
end.


Lemma CodeCorrect_red : forall Mach C Psi n n', 
  n' < n -> 
  CodeCorrect Mach C Psi n ->
  CodeCorrect Mach C Psi n'.
unfold CodeCorrect; intros.
assert (cn < n).
omega.
apply H0; auto.
Qed.

Definition gt_dec : forall a b, a > b \/ b >= a.
intros.
omega.
Qed.


Hint Resolve ActionGTE_refl.

(* correctness for procedures *)
(* if the codeheap is correct with index n, then the procedure is correct with index n *)
Lemma Correctness : forall Mach Psi C n I g, 
WFProc dp Mach Psi g I -> 
WFCode Mach (SpecHeap_empty) C Psi ->
CodeCorrect Mach C Psi n ->
ProcCorrect Mach C I g n.

induction n; firstorder.
(* unfold ProcCorrect. *)
induction H; firstorder.

unfold ProcCorrect; unfold getOpAction in *; simpl; rewrite H; simpl; auto.

unfold ProcCorrect in *.
simpl.
inversion_clear H0.
destruct (H2 l g H); clear H2.
destruct H0.
assert (n < S n); [omega |].
generalize (H1 l n H3); clear H1; intros.
rewrite H0 in * |- *; clear H0.
red in H.
rewrite H in * |- *.
unfold ProcCorrect in H1.
makenew' (mean n Mach C x); firstorder.

unfold ProcCorrect in *.
makenew' (mean (S n) Mach C I); firstorder.

unfold ProcCorrect in *.
clear H3 H4; simpl.
assert (n < S n); [omega |].
generalize (CodeCorrect_red Mach C Psi (S n) n H3 H1); clear H1; intros H1.

generalize (IHn I' g' H H0 H1); intros.
generalize (IHn I'' g'' H2 H0 H1); intros.
unfold ProcCorrect in *.
makenew' (mean n Mach C I'); 
makenew' (mean n Mach C I''); 
firstorder.

unfold ProcCorrect in *.
clear H3 H4; simpl.
assert (n < S n); [omega |].
generalize (CodeCorrect_red Mach C Psi (S n) n H3 H1); clear H1; intros H1.
generalize (IHn I' g' H H0 H1); intros.
generalize (IHn I'' g'' H2 H0 H1); intros.
unfold ProcCorrect in *.
makenew' (mean n Mach C I'); 
makenew' (mean n Mach C I''); 
firstorder.
Qed.

Lemma Corr2 : forall Mach Psi C n,
WFCode Mach (SpecHeap_empty) C Psi ->
CodeCorrect Mach C Psi n.

induction n.
intros; red; intros; inversion H0.

intros H.
generalize (IHn H); clear IHn; intros H0.

set (wfcode_copy := H); generalize wfcode_copy; clear wfcode_copy; intros wfcode_copy.
inversion_clear H.
rewrite SpecHeap_join_empty in *; auto.
red; intros.
makenew' (Psi l); makenew' (C l); auto.
unfold getLabelAction in H1.
generalize (H1 l a H3); clear H1; intros H1.
destruct H1.
rewrite H4 in H1.
destruct H1.
inversion H1; clear H1.
rewrite H6 in *; clear p H6.

apply Correctness with Psi; auto.
assert (cn < n \/ cn = n).
omega.
destruct H1.
apply CodeCorrect_red with n; try omega; auto.
rewrite H1; auto.

(* false cases *)
specialize (H1 l a).
unfold getLabelAction in *.
specialize (H1 H3).
destruct H1.
destruct H1.
rewrite H4 in H1; inversion H1.
Qed.

End DP.

End SCAP.
