(* spec.v *)
(* Alex Vaynberg *)
(* This module defines the concept of actions and specifications in the SCAP system *)

Require Import DecidableType.
(* Require Import OptionFacts. *)

Set Implicit Arguments.

Section Specification.

Variable State : Type.
(* Define Assert, Action, and Branch Predicate *)

Record Action := mkAct { pre: State -> Prop ; rel : State -> State -> Prop }.


(*** Core Combinators *)

Definition ActionID : Action := mkAct (fun s => True) (fun s s' => s' = s).
Definition ActionFail : Action := mkAct (fun s => False) (fun s s' => False).
Definition ActionLoop : Action := mkAct (fun s => True) (fun s s' => False).

Definition ActionChain (g g' : Action) : Action := 
mkAct 
  (fun s => g.(pre) s /\ forall s', g.(rel) s s' -> g'.(pre) s')
  (fun s s'' => (* (forall s', g.(rel) s s' -> g'.(pre) s') /\  *)
                (* (exists s', g.(pre) s /\ g.(rel) s s' /\ g'.(pre) s' /\ g'.(rel) s' s'')). *)
                (exists s', g.(rel) s s' /\ g'.(rel) s' s'')).

Definition ActionPlus (p : State -> Prop) (g g' : Action) : Action :=
mkAct
  (fun s => (g.(pre) s /\ p s) \/ (g'.(pre) s /\ ~(p s)))
  (fun s s' => (p s -> g.(rel) s s') /\ (~(p s) -> g'.(rel) s s')).


Definition dom (r : State -> State -> Prop) := fun s => exists s', r s s'.

Definition ActionGTE (g g' : Action) : Prop :=
  (forall s, g.(pre) s -> g'.(pre) s) /\ 
  (forall s s', g.(pre) s -> g'.(rel) s s' -> g.(rel) s s').


Definition ActionAnd (g g' : Action) :=
mkAct (fun s => g.(pre) s /\ g'.(pre) s)
      (fun s s' => g.(rel) s s' /\ g'.(rel) s s').

Definition ActionOr (g g' : Action) :=
mkAct (fun s => g.(pre) s \/ g'.(pre) s)
      (fun s s' => g.(rel) s s' \/ g'.(rel) s s').

Definition ActionAll {A} (g : A -> Action) :=
mkAct (fun s => exists a, (g a).(pre) s)
      (fun s s' => exists a, (g a).(pre) s /\ (g a).(rel) s s').



(* Non-core Useful Combinators *)

Definition ActionLet (A : Type) (v : State -> A) (f : A -> Action) : Action :=
  mkAct (fun s => (f (v s)).(pre) s) (fun s s' => (f (v s)).(rel) s s').
Implicit Arguments ActionLet [A].

Definition ActionLetO (A : Type) (v : State -> option A) (f : A -> Action) : Action :=
  mkAct (fun s => match v s with | Some x => (f x).(pre) s | None => False end)
  (fun s s' => match v s with | Some x => (f x).(rel) s s' | None => False end).
Implicit Arguments ActionLetO [A].

Definition ActionCheck (p : State -> Prop) (a : Action) : Action := 
  mkAct (fun s => p s /\ a.(pre) s) a.(rel).



(* Properties of Weaker-than relation *)

Definition ActionGTE_refl : forall g, ActionGTE g g.
firstorder.
Qed.

Lemma ActionGTE_trans : forall g1 g2 g3, ActionGTE g1 g2 -> ActionGTE g2 g3 -> ActionGTE g1 g3.
firstorder.
Qed.

Lemma ActionGTE_Plus : forall p g1 g1' g2 g2', 
 ActionGTE g1 g1' -> ActionGTE g2 g2' -> ActionGTE (ActionPlus p g1 g2) (ActionPlus p g1' g2').
firstorder.
Qed.

Lemma ActionGTE_Chain : forall g1 g1' g2 g2',
 ActionGTE g1 g1' -> ActionGTE g2 g2' -> ActionGTE (ActionChain g1 g2) (ActionChain g1' g2').
firstorder.
Qed.

Lemma ActionChain_assoc : forall g1 g2 g3, 
ActionGTE (ActionChain g1 (ActionChain g2 g3)) (ActionChain (ActionChain g1 g2) g3).
firstorder.
Qed.

Lemma ActionChain_assoc' : forall g1 g2 g3, 
ActionGTE (ActionChain (ActionChain g1 g2) g3) (ActionChain g1 (ActionChain g2 g3)).
firstorder.
Qed.

(*
Ltac rewriter := 
match goal with
 | h1 : Some ?a = Some ?b |- _ => inversion h1; clear h1; subst
 | h1 : ?a = Some ?b |- _ => 
    match type of a with
    | option State => symmetry in h1
    | _ => fail
    end
 | h1 : Some ?a = ?b |- _ => rewrite h1 in * |- *; try clear a h1
 | h1 : ?a = ?b |- _ => rewrite h1 in * |- *
end.
*)

End Specification.
