(* specheap.v *)
(* Holds the type of the specification heap *)
(* as well as operation over specification heaps *)

Require Import Logic.FunctionalExtensionality.
Require Import Equalities.

Require Import spec.
Require Import label.
Require Import zbool_assist.


Module SpecHeap (State : Typ).


(* The definition of the Specification Heap *)
Definition SpecHeap := Label -> option (Action State.t).

(* Empty Heap *)
Definition SpecHeap_empty : SpecHeap := fun l => None.


(* DEPRECATED?: Compatible Specification Heaps *)
Definition SpecHeap_compat (psi psi' : SpecHeap) : Prop :=
  forall l,
  match (psi l, psi' l) with
     | (Some g, Some g') => g = g'
     | _ => True
  end.

Lemma SpecHeap_compat_sym {psi psi'} : 
   SpecHeap_compat psi psi' -> SpecHeap_compat psi' psi.
intros.
unfold SpecHeap_compat in *; intros.
specialize (H l).
makenew' (psi' l); auto.
makenew' (psi l); auto.
Qed.

Lemma SpecHeap_empty_compat : forall psi, SpecHeap_compat SpecHeap_empty psi.
firstorder.
Qed.


(* Disjoint Specification Heaps *)

Definition SpecHeap_disjoint (psi psi' : SpecHeap) : Prop :=
 forall l, 
  match psi l with
  | Some _ => psi' l = None
  | _ => True
  end.

Lemma SpecHeap_disjoint_sym {psi psi'} : 
   SpecHeap_disjoint psi psi' -> SpecHeap_disjoint psi' psi.
unfold SpecHeap_disjoint; intros.
specialize (H l).
makenew' (psi' l); auto.
makenew' (psi l); auto.
inversion H.
Qed.

Lemma SpecHeap_disjoint_compat {psi psi'} :
   SpecHeap_disjoint psi psi' -> SpecHeap_compat psi psi'.
unfold SpecHeap_disjoint, SpecHeap_compat; intros.
specialize (H l).
makenew' (psi l); auto.
rewrite H; auto.
Qed.

Lemma SpecHeap_empty_disjoint : forall psi, SpecHeap_disjoint SpecHeap_empty psi.
firstorder.
Qed.


(* Addition of Heaps *)

Definition SpecHeap_join (Psi Psi' : SpecHeap) (compat : SpecHeap_disjoint Psi Psi') :=
  fun l => match Psi l with
	   | Some g => Some g
	   | None => Psi' l
	   end.

Definition SpecHeap_join_sym : forall psi psi' disj,
   SpecHeap_join psi psi' disj = SpecHeap_join psi' psi (SpecHeap_disjoint_sym disj).
intros.
unfold SpecHeap_join, SpecHeap_compat in *.
apply functional_extensionality; intros.
specialize (disj x).
makenew' (psi x); auto.
rewrite disj; auto.
makenew' (psi' x); subst; auto.
Qed.


Lemma SpecHeap_join_empty : forall Psi disj, SpecHeap_join SpecHeap_empty Psi disj = Psi.
intros.
apply functional_extensionality; firstorder.
Qed.

Lemma SpecHeap_disjoint_joinleft : forall psi psi' psi''
(disj : SpecHeap_disjoint psi psi'),
SpecHeap_disjoint psi psi'' -> 
SpecHeap_disjoint psi' psi'' ->
SpecHeap_disjoint
  (SpecHeap_join psi psi' disj) psi''.
unfold SpecHeap_disjoint, SpecHeap_join in *.
intros.
specialize (disj l).
specialize (H l).
specialize (H0 l).
destruct (psi l); auto.
Qed.



(* Bigger Heaps *)
(* A bigger heap is one that that contains more labels, but all the matching labels are equivalent *)

Definition SpecHeap_bigger (psi psi' : SpecHeap) : Prop :=
forall l, 
   match psi' l with 
   | Some g' => match psi l with
               | Some g => g = g'
               | None => False
               end
   | None => True
   end.

Lemma SpeHeap_bigger_refl : forall psi, SpecHeap_bigger psi psi.
intros; red; intros; simpl.
destruct (psi l); auto.
Qed.

(* If a bigger heap is disjoint with some other heap, than a smaller heap is also disjoint*)
Lemma SpecHeap_disjoint_smaller : forall psi bpsi psi',
  SpecHeap_bigger bpsi psi ->
  SpecHeap_disjoint bpsi psi' -> SpecHeap_disjoint psi psi'.
unfold SpecHeap_bigger, SpecHeap_disjoint; intros.
specialize (H l).
specialize (H0 l).
destruct (psi l); auto.
destruct (bpsi l); auto.
inversion H.
Qed.

Lemma SpecHeap_join_bigger {psi psi'} {disj:(SpecHeap_disjoint psi psi')} : 
  SpecHeap_bigger (SpecHeap_join psi psi' disj) psi.
red; intros.
unfold SpecHeap_join.
specialize (disj l).
destruct (psi l); auto.
Qed.

Lemma SpecHeap_bigger_join {psi psi' psi''} 
      {disj : SpecHeap_disjoint psi psi''} {disj' : SpecHeap_disjoint psi' psi''}:
  SpecHeap_bigger psi psi' ->
  SpecHeap_bigger (SpecHeap_join psi psi'' disj) (SpecHeap_join psi' psi'' disj').
unfold SpecHeap_bigger, SpecHeap_join, SpecHeap_disjoint in *.
intros.
specialize (disj l).
specialize (disj' l).
specialize (H l).
destruct (psi' l); auto.
destruct (psi l); auto.
inversion H.
destruct (psi l); auto.
rewrite disj; auto.
destruct (psi'' l); auto.
Qed.

Lemma SpecHeap_bigger_join' : forall psi psi' psi''
      {disj : SpecHeap_disjoint psi'' psi} {disj' : SpecHeap_disjoint psi'' psi'},
  SpecHeap_bigger psi psi' ->
  SpecHeap_bigger (SpecHeap_join psi'' psi disj) (SpecHeap_join psi'' psi' disj').
unfold SpecHeap_bigger, SpecHeap_join, SpecHeap_disjoint in *.
intros.
specialize (disj l).
specialize (disj' l).
specialize (H l).
destruct (psi'' l); auto.
Qed.


(* Stronger Heaps *)

Definition SpecHeap_stronger (psi psi' : SpecHeap) : Prop :=
forall l, 
   match psi' l with 
   | Some g' => match psi l with
               | Some g => ActionGTE g' g
               | None => False
               end
   | None => True
   end.

Lemma SpecHeap_stronger_refl : forall psi, SpecHeap_stronger psi psi.
intros; red; intros; simpl.
destruct (psi l); auto.
apply ActionGTE_refl.
Qed.


Lemma SpecHeap_disjoint_stronger {psi spsi psi'} :
  SpecHeap_stronger spsi psi ->
  SpecHeap_disjoint spsi psi' -> SpecHeap_disjoint psi psi'.
unfold SpecHeap_disjoint, SpecHeap_stronger; intros.
specialize (H l).
specialize (H0 l).
makenew' (psi l); auto.
makenew' (spsi l); subst; auto.
contradiction.
Qed.

Lemma SpecHeap_stronger_add : forall psi spsi (disj : SpecHeap_disjoint psi spsi),
SpecHeap_stronger (SpecHeap_join psi spsi disj) psi. 
intros.
red.
unfold SpecHeap_join.
intros.
destruct (psi l); auto.
apply ActionGTE_refl.
Qed.

Lemma SpecHeap_stronger_join : forall psi spsi j (disj : SpecHeap_disjoint spsi j)
(str : SpecHeap_stronger spsi psi) spsi_j_disj psi_j_disj,
SpecHeap_stronger 
      (SpecHeap_join spsi j spsi_j_disj) 
      (SpecHeap_join psi j psi_j_disj). 

unfold SpecHeap_stronger, SpecHeap_join, SpecHeap_disjoint.
intros; subst.
specialize (disj l).
specialize (str l).
makenew' (psi l); auto.
makenew' (spsi l); subst; auto.
contradiction.
makenew' (spsi l); subst; auto.
rewrite disj; auto.
makenew' (j l); auto.
apply ActionGTE_refl.
Qed.

Lemma SpecHeap_stronger_join' : forall psi spsi j (disj : SpecHeap_disjoint j spsi)
(str : SpecHeap_stronger spsi psi) j_spsi_disj j_psi_disj,
SpecHeap_stronger 
      (SpecHeap_join j spsi j_spsi_disj) 
      (SpecHeap_join j psi j_psi_disj). 

unfold SpecHeap_stronger, SpecHeap_join, SpecHeap_disjoint.
intros; subst.
specialize (disj l).
specialize (str l).
makenew' (j l); auto.
apply ActionGTE_refl.
Qed.

(* SpecHeap subtractions *)
Definition SpecHeap_sub (psi psi' : SpecHeap) : SpecHeap :=
fun l => match psi' l with
         | Some _ => None
         | None => psi l
         end.

Lemma SpecHeap_sub_compat : forall psi psi' s, 
  SpecHeap_compat psi psi' -> SpecHeap_compat (SpecHeap_sub psi s) psi'.
unfold SpecHeap_compat, SpecHeap_sub; intros.
specialize (H l).
makenew' (s l); auto.
Qed.

Lemma SpecHeap_sub_disj : forall psi psi', SpecHeap_disjoint (SpecHeap_sub psi psi') psi'.
intros.
unfold SpecHeap_disjoint, SpecHeap_sub; intros.
makenew' (psi' l); auto.
makenew' (psi l); auto.
Qed.

Definition SpecHeap_add psi psi' := 
  SpecHeap_join (SpecHeap_sub psi psi') psi' (SpecHeap_sub_disj psi psi').

End SpecHeap.