(* base_simp.v *)
(* General simplifications of terms, mostly based on proof irrelevance *)
(* And basic mathematical facts *)

Require Import ZArith.
Require Import NArith.
Require Import Bool.

Definition nat_of_Z (z : Z) := nat_of_N (N_of_Z z).
Definition bool_of_Z (z : Z) : bool := negb (Zeq_bool z 0).
Definition Zdiv_up (a b :Z) := if Zeq_bool (Zmod a b) 0%Z then Zdiv a b else ((Zdiv a b) + 1)%Z.

Lemma Zeq_dec : forall (a b : Z), (a = b)%Z \/ (a <> b)%Z.
intros.
destruct (Z_eq_dec a b); firstorder.
Qed.

Lemma Zlt_dec : forall (a b : Z), (a < b)%Z \/ ~(a < b)%Z.
intros.
destruct (Z_lt_dec a b); firstorder.
Qed.

Lemma andb_true1 : forall b, true && b = b.
auto.
Qed.

Lemma andb_true2 : forall b, b && true = b.
intros.
destruct b; auto.
Qed.

Lemma andb_false1 : forall b, false && b = false.
intros; auto.
Qed.

Lemma andb_false2 : forall b, b && false = false.
intros.
destruct b; auto.
Qed.

Hint Rewrite andb_true1 andb_true2 andb_false1 andb_false2 : base.

Lemma if_irrelev : forall (A:Type) (b:bool) (x:A), (if b then x else x) = x.
intros.
destruct b; auto.
Qed.
Hint Rewrite if_irrelev : base.

Lemma match_option_irrel : forall (A B: Type) (b : option A) (x: B),
  (match b with | Some _ => x | None => x end) = x.
intros.
destruct b; auto.
Qed.
Hint Rewrite match_option_irrel : base.

Lemma match_option_bool_irrel : forall (A: Type) (b : option bool) (x: A),
  (match b with | Some true => x | Some false => x | None => x end) = x.
intros.
destruct b; auto.
destruct b; auto.
Qed.
Hint Rewrite match_option_bool_irrel : base.

Require Import ProofIrrelevance.
Require Import Classical.
Require Import ClassicalFacts.

Axiom prop_ext : prop_extensionality.

Lemma neg_simp : forall b, (negb b = true) = (b = false).
intros.
apply prop_ext.
destruct b; firstorder.
Qed.

Lemma neg_simp' : forall b, (negb b = false) = (b = true).
intros.
apply prop_ext.
destruct b; firstorder.
Qed.
Hint Rewrite neg_simp neg_simp': base.

Lemma nottrue_false : not True = False.
apply prop_ext.
firstorder.
Qed.
Hint Rewrite nottrue_false : base.

Lemma conj_simpl : forall (x: Prop), (True /\ x) = x.
intros.
apply prop_ext; firstorder.
Qed.
Hint Rewrite conj_simpl : base.

Lemma refl_elim : forall (A:Type) (x:A), (x = x) = True.
intros.
apply prop_ext; firstorder.
Qed.
Hint Rewrite refl_elim : base.

Lemma notFalse : (~ False) = True.
apply prop_ext; firstorder.
Qed.

Lemma notTrue : (~True) = False.
apply prop_ext; firstorder.
Qed.

Hint Rewrite notFalse notTrue : base.


Lemma Telim_left : forall P, (True /\ P) = P.
intros.
apply prop_ext; firstorder.
Qed.
Lemma Telim_right : forall P, (P /\ True) = P.
intros.
apply prop_ext; firstorder. 
Qed.
Hint Rewrite Telim_left Telim_right : base.

Lemma CFalse_left : forall P, (False /\ P) = False.
intros.
apply prop_ext; firstorder.
Qed.

Lemma CFalse_right : forall P, (P /\ False) = False.
intros.
apply prop_ext; firstorder.
Qed.
Hint Rewrite CFalse_left CFalse_right : base.

Lemma Delim_left : forall P, (False \/ P) = P.
intros.
apply prop_ext; firstorder.
Qed.
Lemma Delim_right : forall P, (P \/ False) = P.
intros.
apply prop_ext; firstorder.
Qed.
Lemma DTelim1 : forall P, (True \/ P) = True.
intros.
apply prop_ext; firstorder.
Qed.
Lemma DTelim2 : forall P, (P \/ True) = True.
intros.
apply prop_ext; firstorder.
Qed.
Hint Rewrite Delim_left Delim_right DTelim1 DTelim2 : base.

Lemma Some_remove : forall (A : Type) (a b : A), (Some a = Some b) = (a = b).
intros.
apply prop_ext; firstorder.
inversion H; auto.
subst; auto.
Qed.

Lemma Some_None_left : forall (A : Type) (a : A), (None = Some a) = False.
intros.
apply prop_ext; firstorder.
inversion H; auto.
Qed.

Lemma Some_None_right : forall (A : Type) (a : A), (Some a = None) = False.
intros.
apply prop_ext; firstorder.
inversion H; auto.
Qed.
Hint Rewrite Some_remove Some_None_left Some_None_right : base.

Lemma true_false : (true = false) = False.
apply prop_ext; firstorder.
Qed.

Lemma false_true : (false = true) = False.
apply prop_ext; firstorder.
Qed.
Lemma true_true : (true = true) = True.
apply prop_ext; firstorder.
Qed.

Lemma false_false : (false = false) = True.
apply prop_ext; firstorder.
Qed.
Hint Rewrite true_false false_true true_true false_false : base.

Lemma plus_minus : forall (z l: Z), (z + l - l)%Z = z.
intros.
omega.
Qed.

Lemma minus_plus : forall (z l : Z), (z - l + l)%Z = z.
intros.
omega.
Qed.

Hint Rewrite plus_minus minus_plus : base.

Lemma zero_one : (0%Z = 1%Z) = False.
apply prop_ext; firstorder.
Qed.

Lemma one_zero : (1%Z = 0%Z) = False.
apply prop_ext; firstorder.
Qed.
Hint Rewrite zero_one one_zero : base.

Lemma if_remove : forall (A : Type) (b:bool) (x:A), (if b then x else x) = x.
intros.
destruct b; auto.
Qed.
Hint Rewrite if_remove : base.

Lemma andb_split : forall a b, (andb a b = true) = (a = true /\ b = true).
intros.
apply prop_ext; firstorder.
destruct a; destruct b; auto.
destruct a; destruct b; auto.
Qed.

Lemma andb_or : forall a b, (andb a b = false) = (a = false \/ b = false).
intros.
apply prop_ext; firstorder.
destruct a; destruct b; auto.
destruct a; destruct b; auto.
destruct a; destruct b; auto.
Qed.
Hint Rewrite andb_split andb_or: base.

Lemma geq_remove : forall (a b : Z), (a >= b)%Z = (b <= a)%Z.
intros.
apply prop_ext; firstorder.
Qed.
Hint Rewrite geq_remove : base.

Lemma Zeq_bool_plus : forall a b c, Zeq_bool (a + b) (a + c) = Zeq_bool b c.
intros.
unfold Zeq_bool.
rewrite Zcompare_plus_compat; auto.
Qed.
Hint Rewrite Zeq_bool_plus : base.

Lemma time_div_PGSIZE: forall z, (z * 4096 / 4096)%Z = z.
intros.
rewrite Z_div_mult; auto.
omega.
Qed.
Hint Rewrite time_div_PGSIZE : base.

Lemma not01 : (0%Z = 1%Z) = False.
apply prop_ext; firstorder.
Qed.

Lemma not10 : (1%Z = 0%Z) = False.
apply prop_ext; firstorder.
Qed.

Lemma not02 : (0%Z = 2%Z) = False.
apply prop_ext; firstorder.
Qed.

Lemma not20 : (2%Z = 0%Z) = False.
apply prop_ext; firstorder.
Qed.

Hint Rewrite not01 not10 : base.
